Require Export tcb.

(* Semantics invariants *)
Inductive wf_val (h:heap) : val -> Prop :=
| wf_val_num : forall n, wf_val h (Num n)
| wf_val_ref : forall a loc, h loc = Array a -> wf_val h (Ref loc).

Definition wf_state (st:state) : Prop :=
  match st with
    | St pc s l h => 
      (forall x v, l x = Def v -> wf_val h v)
      /\ (forall v, In v s -> wf_val h v)
    | Error => True
  end.

Lemma wf_state_invariant_step : forall p st1 st2,
  st1 -[p]-> st2 ->
  wf_state st1 -> wf_state st2.
Proof.
  intros.
  inv H; simpl in *; destruct H0; split; 
    intuition (subst; eauto with datatypes);
      try constructor.
  inv H2.
  destruct (eq_word_dec x0 x); subst.
  replace v with v0 in * by congruence; auto.
  apply H with x0; rewrite H5; auto.
  inv H2.
  destruct v; try constructor.
  destruct (eq_word_dec l0 loc); subst.
  econstructor; eauto.
  assert (H':=H _ _ H3).
  inv H'.
  rewrite H5 in H7; auto.
  econstructor; eauto.
  inv H2.
  econstructor; eauto.
  inv H2.
  destruct v; try constructor.
  destruct (eq_word_dec l0 loc); subst.
  econstructor; eauto.
  assert (wf_val h (Ref l0)) by auto  with datatypes.
  inv H2.
  rewrite H5 in H9; auto.
  econstructor; eauto.
  inv H2.
  destruct v; try constructor.
  inv H6.
  destruct (eq_word_dec l0 loc); subst.
  econstructor; eauto.
  assert (H':=H _ _ H3).
  inv H'.
  rewrite <- H7 in H8; auto.
  econstructor; eauto.
  destruct v; try constructor.
  inv H2; inv H6.
  destruct (eq_word_dec l0 loc); subst.
  econstructor; eauto.
  assert (wf_val h (Ref l0)) by auto  with datatypes.
  inv H6.
  rewrite <- H7 in H9; auto.
  econstructor; eauto.
Qed.

(* Semantic proof about error hecking *)
Inductive pre_state_error (p:program) : state -> Prop :=
| no_pre_state_error_newarray : forall pc n s l h,
  instr_at p pc = Some Newarray ->
  n<0 ->
  pre_state_error p (St pc (Num n :: s) l h)
| no_pre_state_error_iaload : forall pc loc s l h i a,
  instr_at p pc = Some Arrayload ->
  h loc = Array a ->
  ~ 0 <= i < array_length a ->
  pre_state_error p (St pc (Num i::Ref loc :: s) l h)
| pre_state_error_iastore : forall pc loc s l h i n a,
  instr_at p pc = Some Arraystore ->
  h loc = Array a ->
  ~ 0 <= i < array_length a ->
  pre_state_error p (St pc (Num n :: Num i :: Ref loc :: s) l h).

Lemma no_pre_state_error_implies_safe : forall p,
  forall st1 st2, ~ pre_state_error p st1 -> st1 -[p]-> st2 -> Safe st2.
Proof.
  unfold Safe; intros.
  inv H0; try congruence; elim H; clear H. 
  econstructor 1; eauto.
  econstructor 2; eauto.
  econstructor 3; eauto.
Qed.

Lemma step_instr_at : forall p pc s l h st2,
  St pc s l h -[ p ]-> st2 ->
  exists ins, instr_at p pc = Some ins.
Proof.
  intros.
  case_eq (instr_at p pc); eauto.
  intros.
  apply False_ind.
  inv H; congruence.
Qed.

Lemma no_pre_state_error_implies_safe_reach : forall p,
  (forall pc s l h ins, 
    ReachableStates p (St pc s l h) -> 
    instr_at p pc = Some ins ->
    ~ pre_state_error p (St pc s l h)) ->
  forall st, ReachableStates p st -> Safe st.
Proof.
  induction 2.
  simpl; congruence.
  destruct st1.
  eapply no_pre_state_error_implies_safe; eauto.
  elim step_instr_at with (1:=H1); eauto.
  unfold Safe in *; congruence.
Qed.

Definition PreSafe (p:program) (st:state) : Prop := ~ pre_state_error p st.

(* This is all we need about the numeric abstraction *)
Module Type AB_NUM.

  Declare Module Lat : Lattice.

  Parameter gamma : Lat.t -> integer -> Prop.
  Parameter gamma_monotone : forall N1 N2 n,
    Lat.order N1 N2 -> gamma N1 n -> gamma N2 n.
  Parameter gamma_meet_morph : forall N1 N2 n,
    gamma N1 n -> gamma N2 n -> gamma (Lat.meet N1 N2) n.

  Parameter bottom_empty : forall n, ~ gamma Lat.bottom n.

  Parameter const : integer -> Lat.t.
  Parameter const_correct : forall n, gamma (const n) n.

  Parameter sem_add : Lat.t -> Lat.t -> Lat.t.
  Parameter sem_add_correct : forall N1 N2 n1 n2 n,
    gamma N1 n1 -> gamma N2 n2 -> 
    n = n1 + n2 ->
    gamma (sem_add N1 N2) n.

  Parameter sem_sub : Lat.t -> Lat.t -> Lat.t.
  Parameter sem_sub_correct : forall N1 N2 n1 n2 n,
    gamma N1 n1 -> gamma N2 n2 -> 
    n = n1 - n2 ->
    gamma (sem_sub N1 N2) n.

  Parameter sem_mult : Lat.t -> Lat.t -> Lat.t.
  Parameter sem_mult_correct : forall N1 N2 n1 n2 n,
    gamma N1 n1 -> gamma N2 n2 -> 
    n = n1 * n2 ->
    gamma (sem_mult N1 N2) n.

  Parameter backsem_add : Lat.t -> Lat.t -> Lat.t -> Lat.t*Lat.t.
  Parameter backsem_add_correct : forall N N1 N2 n1 n2 n N1' N2',
    gamma N1 n1 -> gamma N2 n2 ->
    n = n1 + n2 -> gamma N n ->
    backsem_add N N1 N2 = (N1',N2') ->
    gamma N1' n1 /\ gamma N2' n2.

  Parameter backsem_sub : Lat.t -> Lat.t -> Lat.t -> Lat.t*Lat.t.
  Parameter backsem_sub_correct : forall N N1 N2 n1 n2 n N1' N2',
    gamma N1 n1 -> gamma N2 n2 ->
    n = n1 - n2 -> gamma N n ->
    backsem_sub N N1 N2 = (N1',N2') ->
    gamma N1' n1 /\ gamma N2' n2.

  Parameter backsem_mult : Lat.t -> Lat.t -> Lat.t -> Lat.t*Lat.t.
  Parameter backsem_mult_correct : forall N N1 N2 n1 n2 n N1' N2',
    gamma N1 n1 -> gamma N2 n2 ->
    n = n1 * n2 -> gamma N n ->
    backsem_mult N N1 N2 = (N1',N2') ->
    gamma N1' n1 /\ gamma N2' n2.

  Parameter backtest_eq : Lat.t -> Lat.t -> Lat.t*Lat.t.
  Parameter backtest_eq_correct : forall N1 N2 N1' N2' n1 n2,
    n1 = n2 ->
    gamma N1 n1 -> gamma N2 n2 ->
    backtest_eq N1 N2 = (N1',N2') ->
    gamma N1' n1 /\ gamma N2' n2.

  Parameter backtest_lt : Lat.t -> Lat.t -> Lat.t*Lat.t.
  Parameter backtest_lt_correct : forall N1 N2 N1' N2' n1 n2,
    n1 < n2 ->
    gamma N1 n1 -> gamma N2 n2 ->
    backtest_lt N1 N2 = (N1',N2') ->
    gamma N1' n1 /\ gamma N2' n2.

  Parameter top : Lat.t.
  Parameter top_correct : forall n, gamma top n.

  Parameter check_positive : Lat.t -> bool.
  Parameter check_positive_correct : forall N n,
    check_positive N = true ->
    gamma N n -> n>=0.

  Parameter check_in_range : Lat.t -> integer -> integer -> bool.
  Parameter check_in_range_correct : forall N z1 z2 n,
    check_in_range N z1 z2 = true ->
    gamma N n -> z1 < n < z2.

  Parameter is_constant : Lat.t -> option integer.  
  Parameter check_is_constant_correct : forall N z n,
    is_constant N = Some z ->
    gamma N n -> n = z.

End AB_NUM.



(* Symbolic expressions for op. stack *)
Inductive expr := 
   Const (n:integer)
 | Var (x:var)
 | Bin (bin:binop) (e1 e2:expr).


Module Make (AbNum:AB_NUM).

(* Abstract domain construction using the lattice library *)
Module Expr <: Eq.

  Definition t := expr.
  Definition eq (x y : t) : Prop := x=y.
  Lemma eq_refl : forall x : t, eq x x.
  Proof. unfold eq; auto. Qed.

  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof. unfold eq; auto. Qed.

  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof. unfold eq; intros; subst; auto. Qed.

  Lemma eq_dec : forall x y : t, {eq x y}+{~ eq x y}.
  Proof.
    unfold eq; repeat decide equality.
    apply eq_word_dec.
  Qed.

End Expr.

Module ExprLattice := MakeFlatLattice Expr.
Module LocalVarLattice := ArrayBinLattice AbNum.Lat.
Module StackLattice := ListLattice ExprLattice.
Module MemLattice := ProdLattice StackLattice LocalVarLattice.
Module GlobalLattice := ArrayBinLattice MemLattice.


(* Concretization construction *)

Inductive sem_expr (h:heap) (l:locvar) : expr -> val -> Prop :=
| sem_expr_const : forall n, sem_expr h l (Const n) (Num n)
| sem_expr_var : forall x v, l x = Def v -> sem_expr h l (Var x) v
| sem_expr_bin : forall op e1 e2 n1 n2, 
  sem_expr h l e1 (Num n1) ->
  sem_expr h l e2 (Num n2) ->
  sem_expr h l (Bin op e1 e2) (Num (binop_sem op n1 n2))
| sem_expr_ref : forall e a loc,
  h loc = Array a ->
  sem_expr h l e (Num a.(array_length)) ->
  sem_expr h l e (Ref loc)
| sem_expr_length : forall e a loc,
  h loc = Array a ->
  sem_expr h l e (Ref loc) ->
  sem_expr h l e (Num a.(array_length)).

Definition gamma_expr (e:ExprLattice.t) (h:heap) (l:locvar) (v:val) : Prop :=
  match e with
    | Top => True
    | Base e => sem_expr h l e v
    | Bot => False
  end.

Lemma gamma_expr_monotone : forall e1 e2 h l v,
  gamma_expr e1 h l v -> ExprLattice.order e1 e2 -> gamma_expr e2 h l v.
Proof.
  intros.
  inv H0; simpl in H; simpl; auto.
  inv H1; auto.
  elim H.
Qed.

Inductive gamma_list (h:heap) (l:locvar) : list ExprLattice.t -> list val -> Prop :=
 | gamma_list_nil : gamma_list h l nil nil
 | gamma_list_cons : forall e L v q,
    gamma_expr e h l v -> gamma_list h l L q -> gamma_list h l (e::L) (v::q).

Lemma gamma_list_monotone : forall L1 h lv l,
  gamma_list h lv L1 l -> 
  forall L2, list.order _ ExprLattice.order L1 L2 -> gamma_list h lv L2 l.
Proof.
  induction 1; intros L2 H2; inv H2; constructor; eauto using gamma_expr_monotone.
Qed.

Definition gamma_opstack (S:StackLattice.t) (h:heap) (l:locvar) (s:opstack) : Prop :=
  match S with
    | Top => True
    | Base ab_s => gamma_list h l ab_s s
    | Bot => False
  end.

Lemma gamma_opstack_monotone : forall S1 S2 h l s,
  gamma_opstack S1 h l s ->
  StackLattice.order S1 S2 ->
  gamma_opstack S2 h l s.
Proof.
  intros.
  inv H0; simpl in *; intuition eauto using gamma_list_monotone.
Qed.

Inductive gamma_val (i:AbNum.Lat.t) (h:heap) : val -> Prop :=
| gamma_val_num : forall n,
  AbNum.gamma i n -> gamma_val i h (Num n)
| gamma_val_ref : forall a loc,
  h loc = Array a ->
  AbNum.gamma i a.(array_length) -> gamma_val i h (Ref loc).

Lemma gamma_val_monotone : forall V1 V2 h v,
  gamma_val V1 h v ->
  AbNum.Lat.order V1 V2 ->
  gamma_val V2 h v.
Proof.
  intros.
  inv H; econstructor; eauto using AbNum.gamma_monotone.
Qed.

Lemma gamma_val_meet_morph : forall V1 V2 h v,
  gamma_val V1 h v ->
  gamma_val V2 h v ->
  gamma_val (AbNum.Lat.meet V1 V2) h v.
Proof.
  intros.
  inv H; inv H0; econstructor; eauto.
  apply AbNum.gamma_meet_morph; auto.
  replace a with a0 in * by congruence.
  apply AbNum.gamma_meet_morph; auto.
Qed.

Lemma gamma_val_wf_val : forall int h v,
  gamma_val int h v -> wf_val h v.
Proof.
  induction 1; econstructor; eauto.
Qed.

Definition gamma_locvar (L:LocalVarLattice.t) (h:heap) (l:locvar) : Prop :=
  forall x v, l x = Def v -> gamma_val (LocalVarLattice.get L x) h v.

Lemma gamma_locvar_monotone : forall L1 L2 h l,
  gamma_locvar L1 h l ->
  LocalVarLattice.order L1 L2 ->
  gamma_locvar L2 h l.
Proof.
  unfold gamma_locvar; intros.
  eauto using gamma_val_monotone.
Qed.

Lemma gamma_locvar_meet_morph : forall L1 L2 h l,
  gamma_locvar L1 h l ->
  gamma_locvar L2 h l ->
  gamma_locvar (LocalVarLattice.meet L1 L2) h l.
Proof.
  unfold gamma_locvar; intros.
  apply gamma_val_monotone with 
    (AbNum.Lat.meet (LocalVarLattice.get L1 x) (LocalVarLattice.get L2 x)).
  eapply gamma_val_meet_morph; eauto.
  unfold LocalVarLattice.meet, LocalVarLattice.Pos.meet.
  apply AbNum.Lat.order_refl.
  apply AbNum.Lat.eq_sym.
  apply LocalVarLattice.Pos1.get_map2'; intros.
  apply AbNum.Lat.order_antisym.
  apply AbNum.Lat.bottom_is_bottom.
  apply AbNum.Lat.meet_bound1.
  apply AbNum.Lat.order_antisym.
  apply AbNum.Lat.bottom_is_bottom.
  apply AbNum.Lat.meet_bound2.
Qed.
  
Definition gamma_mem : MemLattice.t -> heap -> locvar -> opstack -> Prop :=
  fun M h l s =>
    let (S,L) := M in
      match S with
        | Top => True
        | Base _ => gamma_opstack S h l s /\ gamma_locvar L h l
        | Bot => False
      end.

Lemma gamma_mem_monotone : forall M1 M2 h l s,
  gamma_mem M1 h l s ->
  MemLattice.order M1 M2 ->
  gamma_mem M2 h l s.
Proof.
  unfold gamma_mem; destruct M1; destruct M2; intros.
  inv H0; simpl in *.
  inv H1; auto; inv H; split; eauto using gamma_locvar_monotone.
  apply gamma_opstack_monotone with (Base x); auto.
  constructor; auto. 
Qed.

Definition gamma (p:program) (F:GlobalLattice.t) : state -> Prop :=
  fun st => 
    match st with
      | St i s l h => gamma_mem (GlobalLattice.get F i) h l s
      | _ => True
    end.


(* Analysis specification *)

Record cstr : Set := C {
  source : pc;
  target : pc;
  constraint : MemLattice.t -> MemLattice.t
}.

Definition lift0
  (f:list ExprLattice.t->LocalVarLattice.t->list ExprLattice.t*LocalVarLattice.t)
  (m:MemLattice.t) : MemLattice.t :=
  let (s,l) := m in
    match s with
      | Top => m
      | Base s => let (s',l') := f s l in (Base s',l')
      | Bot => (Bot,LocalVarLattice.bottom)
    end.

Definition lift1
  (f:ExprLattice.t->list ExprLattice.t->LocalVarLattice.t->list ExprLattice.t*LocalVarLattice.t)
  (m:MemLattice.t) : MemLattice.t :=
  let (s,l) := m in
    match s with
      | Top => m
      | Base (e::s) => let (s',l') := f e s l  in (Base s',l')
      | _ => (Bot,LocalVarLattice.bottom)
    end.

Definition lift2
  (f:ExprLattice.t->ExprLattice.t->list ExprLattice.t->LocalVarLattice.t->list ExprLattice.t*LocalVarLattice.t)
  (m:MemLattice.t) : MemLattice.t :=
  let (s,l) := m in
    match s with
      | Top => m
      | Base (e1::e2::s) => let (s',l') := f e1 e2 s l in (Base s',l')
      | _ => (Bot,LocalVarLattice.bottom)
    end.

Definition lift3 
  (f:ExprLattice.t->ExprLattice.t->ExprLattice.t->list ExprLattice.t->LocalVarLattice.t->list ExprLattice.t*LocalVarLattice.t)
  (m:MemLattice.t) : MemLattice.t :=
  let (s,l) := m in
    match s with
      | Top => m
      | Base (e1::e2::e3::s) => let (s',l') := f e1 e2 e3 s l in (Base s',l')
      | _ => (Bot,LocalVarLattice.bottom)
    end.

Module AbEnv.
  Definition forget (x:var) (l:LocalVarLattice.t) : LocalVarLattice.t :=
    LocalVarLattice.modify l x (fun _ => AbNum.top).

  Lemma forget_correct : forall L h l x v l',
    wf_val h v ->
    subst l x v l' ->
    gamma_locvar L h l ->
    gamma_locvar (forget x L) h l'.
  Proof.
    unfold AbEnv.forget, gamma_locvar; intros.
    inv H0.
    destruct (eq_word_dec x0 x).
    assert (v=v0) by congruence.
    subst.
    rewrite LocalVarLattice.apply_modify1; auto.
    inv H;  econstructor; eauto using AbNum.top_correct.
    rewrite LocalVarLattice.apply_modify2; auto.
    apply H1.
    rewrite H4; auto.
  Qed.

  Fixpoint eval_expr (e:expr) (l:LocalVarLattice.t) : AbNum.Lat.t :=
    match e with
      | Const n => AbNum.const n
      | Var x => LocalVarLattice.get l x
      | Bin Add e1 e2 => AbNum.sem_add (eval_expr e1 l) (eval_expr e2 l)
      | Bin Sub e1 e2 => AbNum.sem_sub (eval_expr e1 l) (eval_expr e2 l)
      | Bin Mult e1 e2 => AbNum.sem_mult (eval_expr e1 l) (eval_expr e2 l)
    end.

  Lemma  eval_expr_correct : forall L h l e v,
    sem_expr h l e v ->
    gamma_locvar L h l ->
    gamma_val (eval_expr e L) h v.
  Proof.
    induction 1; intros; simpl.
    constructor.
    apply AbNum.const_correct.
    eauto.
    assert (IH1:=IHsem_expr1 H1);
      assert (IH2:=IHsem_expr2 H1); clear IHsem_expr1 IHsem_expr2.
    inv IH1; inv IH2.
    destruct op; constructor.
    eapply AbNum.sem_add_correct; eauto.
    eapply AbNum.sem_sub_correct; eauto.
    eapply AbNum.sem_mult_correct; eauto.
    econstructor; eauto.
    assert (IH:=IHsem_expr H1).
    inv IH; auto.
    assert (IH:=IHsem_expr H1).
    inv IH; econstructor; eauto.
    replace a with a0 in * by congruence; auto.
  Qed.

  Definition assign (x:var) (e:expr) (l:LocalVarLattice.t) : LocalVarLattice.t :=
    LocalVarLattice.modify l x (fun _ => eval_expr e l).

  Lemma assign_correct : forall L h l x v l' e,
    subst l x v l' ->
      gamma_locvar L h l ->
      sem_expr h l e v ->
      gamma_locvar (AbEnv.assign x e L) h l'.
  Proof.
    unfold AbEnv.assign, gamma_locvar; intros.
    inv H.
    destruct (eq_word_dec x0 x).
    assert (v=v0) by congruence.
    subst.
    rewrite LocalVarLattice.apply_modify1; eauto.
    eapply eval_expr_correct; eauto.
    rewrite LocalVarLattice.apply_modify2; auto.
    apply H0.
    rewrite H4; auto.
  Qed.

  Fixpoint back_eval_expr (l:LocalVarLattice.t) (e:expr) (int:AbNum.Lat.t) {struct e} : LocalVarLattice.t :=
    match e with
      | Const n => 
        if (AbNum.Lat.eq_dec 
             (AbNum.Lat.meet (AbNum.const n) int)
             AbNum.Lat.bottom) then LocalVarLattice.bottom
          else l
      | Var x => LocalVarLattice.modify l x (fun int' => AbNum.Lat.meet int int')
      | Bin Add e1 e2 =>
        let (int1,int2) := AbNum.backsem_add int (eval_expr e1 l) (eval_expr e2 l) in
          LocalVarLattice.meet (back_eval_expr l e1 int1) (back_eval_expr l e2 int2)
      | Bin Sub e1 e2 =>
        let (int1,int2) := AbNum.backsem_sub int (eval_expr e1 l) (eval_expr e2 l) in
          LocalVarLattice.meet (back_eval_expr l e1 int1) (back_eval_expr l e2 int2)
      | Bin Mult e1 e2 =>
        let (int1,int2) := AbNum.backsem_mult int (eval_expr e1 l) (eval_expr e2 l) in
          LocalVarLattice.meet (back_eval_expr l e1 int1) (back_eval_expr l e2 int2)
    end.

  Lemma back_eval_expr_correct : forall h l e v,
    sem_expr h l e v -> forall L V,
      gamma_locvar L h l ->
      gamma_val V h v ->
      gamma_locvar (AbEnv.back_eval_expr L e V) h l.
  Proof.
    induction 1; simpl; intros.
    destruct AbNum.Lat.eq_dec.
    elim AbNum.bottom_empty with n.
    apply AbNum.gamma_monotone with 
      (AbNum.Lat.meet (AbNum.const n) V).
    auto.
    apply AbNum.gamma_meet_morph.
    apply AbNum.const_correct.
    inv H0; auto.
    auto.
    repeat intro.
    destruct (eq_word_dec x0 x); subst.
    replace v with v0 in * by congruence.
    rewrite LocalVarLattice.apply_modify1; auto.
    apply gamma_val_meet_morph; auto.
    rewrite LocalVarLattice.apply_modify2; auto.
    inv H2.
    destruct op.
    case_eq (AbNum.backsem_add V 
      (AbEnv.eval_expr e1 L)
      (AbEnv.eval_expr e2 L)); intros int1 int2 T.
    elim AbNum.backsem_add_correct with (5:=T) (n1:=n1) (n2:=n2) (n:=(binop_sem Add n1 n2)); auto.
    intros.
    apply gamma_locvar_meet_morph; auto.
    apply IHsem_expr1; auto.
    constructor; auto.
    apply IHsem_expr2; auto.
    constructor; auto.
    generalize (eval_expr_correct _ _ _ _ _ H H1); intros.
    inv H2; auto.
    generalize (eval_expr_correct _ _ _ _ _ H0 H1); intros.
    inv H2; auto.
    case_eq (AbNum.backsem_sub V 
      (AbEnv.eval_expr e1 L)
      (AbEnv.eval_expr e2 L)); intros int1 int2 T.
    elim AbNum.backsem_sub_correct with (5:=T) (n1:=n1) (n2:=n2) (n:=(binop_sem Sub n1 n2)); auto.
    intros.
    apply gamma_locvar_meet_morph; auto.
    apply IHsem_expr1; auto.
    constructor; auto.
    apply IHsem_expr2; auto.
    constructor; auto.
    generalize (eval_expr_correct _ _ _ _ _ H H1); intros.
    inv H2; auto.
    generalize (eval_expr_correct _ _ _ _ _ H0 H1); intros.
    inv H2; auto.
    case_eq (AbNum.backsem_mult V 
      (AbEnv.eval_expr e1 L)
      (AbEnv.eval_expr e2 L)); intros int1 int2 T.
    elim AbNum.backsem_mult_correct with (5:=T) (n1:=n1) (n2:=n2) (n:=(binop_sem Mult n1 n2)); auto.
    intros.
    apply gamma_locvar_meet_morph; auto.
    apply IHsem_expr1; auto.
    constructor; auto.
    apply IHsem_expr2; auto.
    constructor; auto.
    generalize (eval_expr_correct _ _ _ _ _ H H1); intros.
    inv H2; auto.
    generalize (eval_expr_correct _ _ _ _ _ H0 H1); intros.
    inv H2; auto.
    apply IHsem_expr; auto.
    inv H2.
    econstructor; eauto.
    replace a with a0 in * by congruence; auto.
    apply IHsem_expr; auto.
    inv H2.
    econstructor; eauto.
  Qed.

  Definition assert_eq (e1 e2:expr) (l:LocalVarLattice.t) : LocalVarLattice.t :=
    let (int1,int2) := AbNum.backtest_eq (eval_expr e1 l) (eval_expr e2 l) in
      LocalVarLattice.meet (back_eval_expr l e1 int1) (back_eval_expr l e2 int2).

  Definition assert_lt (e1 e2:expr) (l:LocalVarLattice.t) : LocalVarLattice.t :=
    let (int1,int2) := AbNum.backtest_lt (eval_expr e1 l) (eval_expr e2 l) in
      LocalVarLattice.meet (back_eval_expr l e1 int1) (back_eval_expr l e2 int2).

  Definition assert (cmp:cmp) (e1 e2:expr) (l:LocalVarLattice.t) : LocalVarLattice.t :=
    match cmp with
      | Eq => assert_eq e1 e2 l
      | Ne => LocalVarLattice.join (assert_lt e1 e2 l) (assert_lt e2 e1 l)
      | Gt => assert_lt e2 e1 l
      | Ge => LocalVarLattice.join (assert_lt e2 e1 l) (assert_eq e1 e2 l)
    end.

  Lemma assert_eq_correct : forall e1 e2 L h l n,
    sem_expr h l e1 (Num n) ->
    sem_expr h l e2 (Num n) ->
    gamma_locvar L h l ->
    gamma_locvar (AbEnv.assert_eq e1 e2 L) h l.
  Proof.
    unfold AbEnv.assert_eq; intros.
    case_eq (AbNum.backtest_eq (AbEnv.eval_expr e1 L) (AbEnv.eval_expr e2 L)).
    intros int1 int2 T.
    assert (W1:=eval_expr_correct L _ _ _ _ H H1).
    assert (W2:=eval_expr_correct L _ _ _ _ H0 H1).
    inv W1; inv W2.
    elim AbNum.backtest_eq_correct with (4:=T) (n1:=n) (n2:=n); auto; intros.
    apply gamma_locvar_meet_morph.
    eapply back_eval_expr_correct; eauto.
    constructor; auto.
    eapply back_eval_expr_correct; eauto.
    constructor; auto.
  Qed.

  Lemma assert_lt_correct : forall e1 e2 L h l n1 n2,
    n1 < n2 ->
    sem_expr h l e1 (Num n1) ->
    sem_expr h l e2 (Num n2) ->
    gamma_locvar L h l ->
    gamma_locvar (AbEnv.assert_lt e1 e2 L) h l.
  Proof.
    unfold AbEnv.assert_lt; intros e1 e2 L h l n1 n2 W H H0 H1.
    case_eq (AbNum.backtest_lt (AbEnv.eval_expr e1 L) (AbEnv.eval_expr e2 L)).
    intros int1 int2 T.
    assert (W1:=eval_expr_correct L _ _ _ _ H H1).
    assert (W2:=eval_expr_correct L _ _ _ _ H0 H1).
    inv W1; inv W2.
    elim AbNum.backtest_lt_correct with (4:=T) (n1:=n1) (n2:=n2); auto; intros.
    apply gamma_locvar_meet_morph.
    eapply back_eval_expr_correct; eauto.
    constructor; auto.
    eapply back_eval_expr_correct; eauto.
    constructor; auto.
  Qed.

  
  Lemma assert_correct : forall e1 e2 cmp L h l n1 n2,
    cmp_sem cmp n1 n2 ->
    sem_expr h l e1 (Num n1) ->
    sem_expr h l e2 (Num n2) ->
    gamma_locvar L h l ->
    gamma_locvar (AbEnv.assert cmp e1 e2 L) h l.
  Proof.
    unfold AbEnv.assert; simpl; intros.
    destruct cmp; simpl in *.
    subst.
    eapply assert_eq_correct; eauto.
    assert (n1<n2 \/ n2<n1) by (unfold integer in *; omega).
    destruct H3.
    apply gamma_locvar_monotone with (AbEnv.assert_lt e1 e2 L).
    eapply assert_lt_correct; eauto.
    apply LocalVarLattice.join_bound1.
    apply gamma_locvar_monotone with (AbEnv.assert_lt e2 e1 L).
    eapply assert_lt_correct; eauto.
    apply LocalVarLattice.join_bound2.
    eapply assert_lt_correct; eauto.
    omega.
    assert (n1>n2 \/ n1=n2) by (unfold integer in *; omega).
    destruct H3; subst.
    apply gamma_locvar_monotone with (AbEnv.assert_lt e2 e1 L).
    eapply assert_lt_correct; eauto.
    omega.
    apply LocalVarLattice.join_bound1.
    apply gamma_locvar_monotone with (AbEnv.assert_eq e1 e2 L).
    eapply assert_eq_correct; eauto.
    apply LocalVarLattice.join_bound2.
  Qed.

End AbEnv.

(* Expression manipulation *)
Fixpoint var_in_expr (x:var) (e:expr) : bool :=
  match e with
    | Const _ => false
    | Var y => if eq_word_dec x y then true else false
    | Bin op e1 e2 =>
      if var_in_expr x e1 then true else var_in_expr x e2
  end.

Definition clear_var (x:var) (e:flat expr) : flat expr :=
  match e with
    | Base e0 => if var_in_expr x e0 then Top else e
    | _ => e
  end.

Lemma update_locvar_gamma_expr : forall e h l v,
  gamma_expr e h l v ->
  forall x v0 l',
    subst l x v0 l' ->
      gamma_expr (clear_var x e) h l' v.
Proof.
  destruct e; simpl; intuition.
  case_eq (var_in_expr x0 x); simpl; auto.
  intros.
  inv H0.
  induction H; simpl in *; try constructor.
  destruct eq_word_dec; try discriminate.
  rewrite <- H3; auto.
  generalize H1; clear H1.
  case_eq (var_in_expr x0 e1); intros; try discriminate; auto.
  generalize H1; clear H1.
  case_eq (var_in_expr x0 e1); intros; try discriminate; auto.
  econstructor; eauto.
  econstructor; eauto.
Qed.

Definition ab_store (x:var) (e:flat expr) (l:LocalVarLattice.t) : LocalVarLattice.t :=
  match e with
    | Bot => LocalVarLattice.bottom
    | Base e => AbEnv.assign x e l
    | Top => AbEnv.forget x l
  end.

Definition ab_binop (op:binop) (e1 e2:flat expr) : flat expr :=
  match e1, e2 with
    | Bot, _ | _, Bot => Bot
    | Top, _ | _, Top => Top
    | Base e1, Base e2 => Base (Bin op e1 e2)
  end.

Definition ab_cmp (cmp:cmp) (e1 e2:flat expr) (l:LocalVarLattice.t) : LocalVarLattice.t :=
  match e1, e2 with
    | Bot, _ | _, Bot => LocalVarLattice.bottom
    | Top, _ | _, Top => l
    | Base e1, Base e2 => AbEnv.assert cmp e1 e2 l
  end.

Definition neg_cmp (cmp:cmp) : tcb.cmp :=
  match cmp with
    | Eq => Ne 
    | Ne => Eq
    | Ge => Gt
    | Gt => Ge
  end.



(* The main function that specifies constraints *)
Definition gen_one_constraint (i:pc) (ins:instruction) : list cstr :=
  match ins with
    | Nop =>
      C i (next i)
      (fun x => x)
      ::nil
    | Push n => 
      C i (next i)
      (lift0 (fun s l => (Base (Const n)::s,l)))
      ::nil
    | Pop => 
      C i (next i)
      (lift1 (fun _ s l => (s,l)))
      ::nil
    | Dup => 
      C i (next i)
      (lift1 (fun e s l => (e::e::s,l)))
      ::nil
    | Load x =>
      C i (next i)
      (lift0 (fun s l => (Base (Var x)::s,l)))
      ::nil
    | Store x => 
      C i (next i)
      (lift1 (fun e s l => (map (clear_var x) s,ab_store x e l)))
      ::nil
    | Binop op =>
      C i (next i)
      (lift2 (fun e1 e2 s l => (ab_binop op e2 e1::s,l)))
      ::nil
    | Newarray => 
      C i (next i)
      (lift1 (fun e s l => (e::s,l)))
      ::nil
    | Arraylength => 
      C i (next i)
      (lift1 (fun e s l => (e::s,l)))
      ::nil
    | Arrayload => 
      C i (next i)
      (lift2 (fun _ _ s l => (Top::s,l)))
      ::nil
    | Arraystore => 
      C i (next i)
      (lift3 (fun _ _ _ s l => (s,l)))
      ::nil
    | Input => 
      C i (next i)
      (lift0 (fun s l => (Top::s,l)))
      ::nil
    | If cmp jump => 
      C i jump
        (lift2 (fun e2 e1 s l => (s,ab_cmp cmp e1 e2 l))) ::
      C i (next i)
      (lift2 (fun e2 e1 s l => (s,ab_cmp (neg_cmp cmp) e2 e1 l)))
      ::nil
    | Goto jump => C i jump (fun x => x) ::nil 
  end.

Definition cstr2prop (s:GlobalLattice.t) (c:cstr) : Prop :=
  MemLattice.order 
    (c.(constraint) (GlobalLattice.get s c.(source))) 
    (GlobalLattice.get s c.(target)).

Definition Spec (p:program) (s:GlobalLattice.t) : Prop :=
  MemLattice.order (Base nil,LocalVarLattice.bottom) (GlobalLattice.get s Word_1) /\
  forall pc instr,
    instr_at p pc = Some instr -> 
    forall c, In c (gen_one_constraint pc instr) -> cstr2prop s c.


(* Constraint generation *)
Fixpoint collect_all_cstr (f:pc->instruction->list cstr)
  (l:list (pc*instruction)) : list cstr :=
  match l with
    | nil => nil
    | (i,ins)::q => (f i ins)++(collect_all_cstr f q)
  end.

Lemma In_collect_all_cstr : forall f l pc instr c,
  instr_at l pc = Some instr ->
  In c (f pc instr) ->
  In c (collect_all_cstr f l).
Proof.
  induction l; simpl; intros.
  discriminate.
  destruct a.
  destruct eq_word_dec; subst.
  inv H; auto with datatypes.
  eauto with datatypes.
Qed.

Definition apply_all_cstr2 (l:list cstr) (s s0:GlobalLattice.t) : GlobalLattice.t :=
  fold_left
    (fun s' c => GlobalLattice.modify s' c.(target) (MemLattice.join (c.(constraint) (GlobalLattice.get s c.(source)))))
    l
    s0.

Lemma apply_all_cstr2_prop1 : forall l s s0,
  GlobalLattice.order s0 (apply_all_cstr2 l s s0).
Proof.
  unfold apply_all_cstr2.
  induction l; simpl; intuition.
  apply GlobalLattice.order_refl; 
    apply GlobalLattice.eq_refl.
  apply GlobalLattice.order_trans with (2:=IHl _ _).
  repeat intro.
  inv H.
  destruct (eq_word_dec a2 (target a)); subst.
  rewrite GlobalLattice.apply_modify1; auto.
  rewrite GlobalLattice.apply_modify2; auto.
Qed.


Lemma apply_all_cstr2_prop2 : forall l s s0 c,
  In c l ->
  MemLattice.order 
    (c.(constraint) (GlobalLattice.get s c.(source))) 
    (GlobalLattice.get (apply_all_cstr2 l s s0) c.(target)).
Proof.
  unfold apply_all_cstr2.
  induction l; simpl; intuition.
  subst.
  apply MemLattice.order_trans with
    (GlobalLattice.get
      (GlobalLattice.modify s0 (target c)
              (MemLattice.join
                 (constraint c (GlobalLattice.get s (source c)))))
        (target c)).
  rewrite GlobalLattice.apply_modify1; auto.
  apply (apply_all_cstr2_prop1 l s); auto.
Qed.

Definition apply_all_cstr3 (l:list cstr) (s:GlobalLattice.t) : GlobalLattice.t :=
  fold_left
    (fun s' c => GlobalLattice.modify s' c.(target) (MemLattice.join (c.(constraint) (GlobalLattice.get s c.(source)))))
    l
    (GlobalLattice.modify GlobalLattice.bottom Word_1
      (fun _ => (Base nil,LocalVarLattice.bottom))).

Lemma apply_all_cstr3_prop : forall l s,
  GlobalLattice.order (apply_all_cstr3 l s) s ->
  forall c, In c l -> cstr2prop s c.
Proof.
  unfold apply_all_cstr3, cstr2prop; intros.
  apply MemLattice.order_trans with 
    (GlobalLattice.get (apply_all_cstr2 l s
      (GlobalLattice.modify GlobalLattice.bottom Word_1
        (fun _ : MemLattice.t => (Base nil, LocalVarLattice.bottom)))) c.(target)); auto.
  apply apply_all_cstr2_prop2; auto.
Qed.

Definition Func p : GlobalLattice.t -> GlobalLattice.t := 
  apply_all_cstr3 (collect_all_cstr gen_one_constraint p).

Lemma Func_correct : forall p s,
  GlobalLattice.order (Func p s) s -> Spec p s.
Proof.
  split; intros.
  generalize (apply_all_cstr2_prop1 
    (collect_all_cstr gen_one_constraint p)
    s (GlobalLattice.modify GlobalLattice.bottom Word_1
      (fun _ => (Base nil,LocalVarLattice.bottom))) Word_1 Word_1).
  rewrite GlobalLattice.apply_modify1; auto; intros.
  apply MemLattice.order_trans with (1:=H0 (refl_equal _)); auto.
  eapply apply_all_cstr3_prop; eauto.
  eapply In_collect_all_cstr; eauto.
Qed.




(* op stack manipualtion *)

Lemma update_locvar_gamma_list : forall h l S s,
  gamma_list h l S s ->
  forall x v l',
    subst l x v l' ->
      gamma_list h l' (map (clear_var x) S) s.
Proof.
  induction 1; intros; simpl; constructor; eauto.
  eapply update_locvar_gamma_expr; eauto.
Qed.



Lemma ab_store_correct : forall l x v l' L h e,
  wf_val h v ->
  subst l x v l' ->
  gamma_locvar L h l ->
  gamma_expr e h l v ->
  gamma_locvar (ab_store x e L) h l'.
Proof.
  intros.
  unfold ab_store.
  destruct e; simpl in *.
  eapply AbEnv.forget_correct; eauto.
  eapply AbEnv.assign_correct; eauto.
  intuition.
Qed.


Ltac prop_on_list l P :=
  match  l with
    | nil => idtac
    | cons ?x ?q => assert (P x) by auto with datatypes; prop_on_list q P
  end.


Lemma create_new_array_gamma_expr : forall h n h' loc e v l,
  create_new_array h n h' loc ->
  gamma_expr e h l v ->
  gamma_expr e h' l v.
Proof.
  intros.
  destruct e; simpl in *; intuition.
  inv H; induction H0; try constructor; auto.
  destruct (eq_word_dec loc loc0); subst.
  replace a with a0 in * by congruence.
  econstructor; eauto.
  rewrite H2 in H; auto.
  econstructor; eauto.
  destruct (eq_word_dec loc loc0); subst.
  replace a with a0 in * by congruence.
  econstructor; eauto.
  rewrite H2 in H; auto.
  econstructor; eauto.
Qed.

Lemma create_new_array_gamma_val : forall h n h' loc V v,
  create_new_array h n h' loc ->
  gamma_val V h v ->
  gamma_val V h' v.
Proof.
  intros.
  inv H0; inv H; econstructor; eauto.
  destruct (eq_word_dec loc loc0); subst.
  replace a with a0 in * by congruence; auto.
  rewrite <- H3; auto.
Qed.


Lemma heap_update_array_gamma_expr : forall h n i0 h' loc e v l,
  heap_update_array h loc i0 n h' ->
  gamma_expr e h l v ->
  gamma_expr e h' l v.
Proof.
  intros.
  destruct e; simpl in *; intuition.
  inv H; inv H2; inv H3; induction H0; try constructor; auto.
  destruct (eq_word_dec loc loc0); subst.
  replace a with a0 in * by congruence.
  econstructor; eauto.
  rewrite H; auto.
  rewrite <- H4 in H0; auto.
  econstructor; eauto.
  destruct (eq_word_dec loc loc0); subst.
  replace a with a0 in * by congruence.
  rewrite <- H.
  econstructor; eauto.
  rewrite <- H4 in H0; auto.
  econstructor; eauto.
Qed.

Lemma heap_update_array_gamma_val : forall h n i0 h' loc V v,
  heap_update_array h loc i0 n h' ->
  gamma_val V h v ->
  gamma_val V h' v.
Proof.
  intros.
  inv H0; inv H.
  constructor; auto.
  inv H3; inv H4.
  destruct (eq_word_dec loc loc0); subst.
  replace a with a0 in * by congruence; auto.
  econstructor; eauto.
  rewrite H; auto.
  rewrite <- H5 in H1; auto.
  econstructor; eauto.
Qed.

Lemma neg_cmp_correct : forall cmp n1 n2,
  ~ cmp_sem cmp n1 n2 ->
  cmp_sem (neg_cmp cmp) n2 n1.
Proof.
  destruct cmp; simpl; unfold integer; intros; try omega.
Qed.

Lemma in1 : forall (c:cstr) l, In c (c::l).
Proof. auto with datatypes. Qed.

Lemma in2 : forall (c1 c2:cstr) l, In c2 (c1::c2::l).
Proof. auto with datatypes. Qed.

Lemma step_correct : forall p F i ins h l s s2,
  instr_at p i = Some ins ->
  (forall c, In c (gen_one_constraint i ins) -> cstr2prop F c) ->
  (St i s l h) -[p]-> s2 ->
  wf_state (St i s l h) ->
  gamma p F (St i s l h) ->
  gamma p F s2.
Proof.
  intros.
  inv H1;
  match goal with 
    [ h1:instr_at ?p ?i = Some ?ins1, h2:instr_at ?p ?i = Some ?ins2 |- _] =>
    rewrite h1 in h2; inv h2
  end;
  simpl gen_one_constraint in *;
  match goal with
    [ h: (forall c, In c ?l -> cstr2prop ?F c) |- _] =>
    prop_on_list l (cstr2prop F); clear h;
      unfold cstr2prop in *;
        unfold gamma in *; 
          simpl in *; auto;
            try (eapply gamma_mem_monotone; [idtac|eauto;fail];
              auto;
                repeat (match goal with [ h: MemLattice.order _ _ |- _] => clear h end);
                  destruct (GlobalLattice.get F i) as [S L];
                    match goal with [ h: gamma_mem _ _ _ _ |- _] => 
                      (destruct S; simpl in h; 
                        [simpl;auto|
                          destruct h; simpl|
                            elim h]) end)
  end.
  (* Ipush *)
  split; [repeat constructor;auto|auto].
  (* Pop *)
  inv H0; split; [repeat constructor;auto|auto].
  (* Pop *)
  inv H0; split; [repeat constructor;auto|auto].
  (* Load x *)
  split; [repeat constructor;auto|auto].
  (* Store x *)
  inv H0; split; [repeat constructor;auto|auto].
  simpl.
  eapply update_locvar_gamma_list; eauto.
  eapply ab_store_correct; eauto.
  destruct H2; auto with datatypes.
  (* Binop *)
  inv H0.
  inv H7.
  repeat constructor; auto.
  destruct e; destruct e0; simpl in *; try (intuition;fail).
  constructor; auto.
  (* Newarray *)
  inv H0; simpl.
  repeat constructor; auto.
  assert (gamma_expr e h' l (Num n)) by (eapply create_new_array_gamma_expr; eauto).
  destruct e; simpl in *; intuition.
  inv H10.
  econstructor; eauto.
  clear H2.
  induction H7; constructor; eauto using create_new_array_gamma_expr.
  unfold gamma_locvar in *; eauto using create_new_array_gamma_val.
  (* Newarray *)
  inv H0; simpl.
  repeat constructor; auto.
  destruct e; simpl in *; intuition.
  econstructor; eauto.
  (* Iaload *)
  inv H0.
  inv H7; simpl.
  repeat constructor; auto.
  (* Iastore *)
  inv H0.
  inv H7.
  inv H8; simpl.
  repeat constructor; auto.
  clear H2; induction H9; constructor; eauto using heap_update_array_gamma_expr.
  unfold gamma_locvar in *; eauto using heap_update_array_gamma_val.
  (* If *)
  inv H0.
  inv H7; simpl.
  repeat constructor; auto.
  destruct e0; destruct e; simpl in *; try (intuition; fail).
  eapply AbEnv.assert_correct; eauto.
  inv H0.
  inv H7; simpl.
  repeat constructor; auto.
  destruct e0; destruct e; simpl in *; try (intuition; fail).
  eapply AbEnv.assert_correct; eauto.
  apply neg_cmp_correct; auto.
  (* Input *)
  repeat constructor; auto.
Qed.


Lemma spec_correct : forall p F,
  Spec p F ->
  forall st, ReachableStates p st -> gamma p F st.
Proof.
  destruct 1.
  assert (forall st, ReachableStates p st -> gamma p F st /\ wf_state st).
  induction 1.
  simpl.
  destruct (GlobalLattice.get F Word_1).
  inv H; simpl in *.
  inv H1.
  repeat split; auto.
  repeat intro; discriminate.
  intuition.  
  inv H3.
  simpl; split.
  repeat constructor; auto.
  repeat intro; discriminate.
  split; repeat intro; try discriminate; intuition.
  (**)
  destruct st1 as [i h s l|].
  assert (exists ins, instr_at p i = Some ins) by (inv H2;eauto).
  destruct H3 as [ins H3].
  split.
  eapply step_correct; intuition eauto.
  eapply wf_state_invariant_step; intuition eauto.
  inv H2.
  intros.
  elim H1 with (1:=H2); auto.
Qed.

(* abstract error check *)
Definition check_instr (F:GlobalLattice.t) (i:pc) (ins:instruction) : bool :=
  match ins with
    | Newarray =>
      match GlobalLattice.get F i with
        | (Top,_) => false
        | (Base (Top::_),L) => false
        | (Base (Base e::_),L) => AbNum.check_positive (AbEnv.eval_expr e L)
        | _ => true
      end
    | Arrayload =>
      match GlobalLattice.get F i with
        | (Top,_) 
        | (Base (Top::_::_),_)  
        | (Base (_::Top::_),_) => false
        | (Base (Base e_i::Base e_a::_),L) => 
          match AbNum.is_constant (AbEnv.eval_expr e_a L) with
            | None => false
            | Some length => AbNum.check_in_range (AbEnv.eval_expr e_i L) (-1)%Z length
          end
        | _ => true
      end
    | Arraystore =>
      match GlobalLattice.get F i with
        | (Top,_) 
        | (Base (_::Top::_::_),_) 
        | (Base (_::_::Top::_),_) => false
        | (Base (_::Base e_i::Base e_a::_),L) => 
          match AbNum.is_constant (AbEnv.eval_expr e_a L) with
            | None => false
            | Some length => AbNum.check_in_range (AbEnv.eval_expr e_i L) (-1)%Z length
          end
        | _ => true
      end
    | _ => true
  end.

Lemma check_instr_correct : forall p F i h s l ins,
  gamma p F (St i s l h) ->
  instr_at p i = Some ins ->
  check_instr F i ins = true ->
  ~ pre_state_error p (St i s l h).
Proof.
  red; intros.
  simpl in *.
  inv H2;
  match goal with 
    [ h1:instr_at ?p ?i = Some ?ins1, h2:instr_at ?p ?i = Some ?ins2 |- _] =>
    rewrite h1 in h2; inv h2
  end; simpl in *.
  (**)
  destruct (GlobalLattice.get F i) as [S L].
  destruct S; inv H; try congruence.
  inv H2.
  destruct e; simpl in *; try congruence.
  assert (n>=0).
  eapply AbNum.check_positive_correct; eauto.
  assert (gamma_val (AbEnv.eval_expr x L) h (Num n)) by
    (eapply AbEnv.eval_expr_correct; eauto).
  inv H; auto.
  omega.
  (**)
  destruct (GlobalLattice.get F i) as [S L].
  destruct S; inv H; try congruence.
  inv H2.
  destruct e; simpl in *; try congruence.
  inv H7; try congruence.
  inv H7; try congruence.
  destruct e; simpl in *; try congruence.
  generalize (AbNum.check_is_constant_correct (AbEnv.eval_expr x0 L)).
  destruct (AbNum.is_constant (AbEnv.eval_expr x0 L)); try congruence.
  intros; assert (array_length a=i1).
  apply H; auto.
  assert (gamma_val (AbEnv.eval_expr x0 L) h (Ref loc)).
  eapply AbEnv.eval_expr_correct; eauto.
  inv H2.
  replace a0 with a in * by congruence; auto.
  assert (gamma_val (AbEnv.eval_expr x L) h (Num i0)).
  eapply AbEnv.eval_expr_correct; eauto.
  inv H4.
  generalize (AbNum.check_in_range_correct _ _ _ _ H1 H11).
  omega.
  (**)
  destruct (GlobalLattice.get F i) as [S L].
  destruct S; inv H; try congruence.
  inv H2.
  inv H7; try congruence.
  destruct e0; simpl in *; try congruence.
  inv H10; try congruence.
  inv H10; try congruence.
  destruct e0; simpl in *; try congruence.
  generalize (AbNum.check_is_constant_correct (AbEnv.eval_expr x0 L)).
  destruct (AbNum.is_constant (AbEnv.eval_expr x0 L)); try congruence.
  intros; assert (array_length a=i1).
  apply H; auto.
  assert (gamma_val (AbEnv.eval_expr x0 L) h (Ref loc)).
  eapply AbEnv.eval_expr_correct; eauto.
  inv H2.
  replace a0 with a in * by congruence; auto.
  assert (gamma_val (AbEnv.eval_expr x L) h (Num i0)).
  eapply AbEnv.eval_expr_correct; eauto.
  inv H4.
  generalize (AbNum.check_in_range_correct _ _ _ _ H1 H12).
  omega.
Qed.

Definition Checked (p:program) (F:GlobalLattice.t) : Prop :=
 forall pc instr,
   instr_at p pc = Some instr -> 
   check_instr F pc instr = true.

Fixpoint check (l:program) (F:GlobalLattice.t) : bool :=
  match l with
    | nil => true
    | (pc,ins)::q => if check_instr F pc ins then check q F else false
  end.

Lemma check_correct : forall p F,
  check p F = true -> Checked p F.
Proof.
  unfold Checked; induction p; simpl; intros.
  discriminate.
  destruct a.
  destruct eq_word_dec; subst.
  inv H0.
  destruct (check_instr F pc instr); congruence.
  destruct (check_instr F p0 i); auto; congruence.
Qed.


(* Fixpoint solver *)
Module GlobalSolve := SolvePostFixPointNotMonotone GlobalLattice.

Lemma analyzer : forall (p:program), 
  { F:GlobalLattice.t | 
    forall st, ReachableStates p st -> gamma p F st }.
Proof.
  intros p.
  set (F:=Func p).
  destruct (GlobalSolve.pfp F) as [x Hx].
  exists x.
  apply spec_correct.
  apply Func_correct; auto.
Qed.

Lemma Checked_sound : forall p F,
  (forall st, ReachableStates p st -> gamma p F st) ->
  Checked p F ->
  forall st, ReachableStates p st -> Safe st.
Proof.
  intros.
  apply no_pre_state_error_implies_safe_reach with p; auto.
  intros; unfold Checked in *.
  apply check_instr_correct with F ins; auto.
Qed.

Definition verifier (p:program) : 
  { b:bool | b = true -> forall st, ReachableStates p st -> Safe st}.
intros p.
destruct (analyzer p) as [F h].
exists (check p F).
intros; apply Checked_sound with p F; auto.
apply check_correct; auto.
Defined.

End Make.

Module IntervalAnalysis := Make IntervalAbstraction.

Module SignAnalysis := Make SignAbstraction.


(* The BubbleSort program *)
Definition i : var := w 1.
Definition j : var := w 2.
Definition tmp : var := w 3.
Definition n : var := w 4.
Definition t : var := w 5.

Definition bbsort : program :=
(w  1, Push  10 )::
(w  2, Store  n )::
(w  3, Load   n )::
(w  4, Newarray )::
(w  5, Store  t )::
(w  6, Push 0   )::
(w  7, Store  i )::
(w  8, Load   i )::
(w  9, Load   n )::
(w 10, Push 1   )::
(w 11, Binop Sub)::
(w 12, If Ge (w 59))::
(w 13, Push 0  )::
(w 14, Store j  )::
(w 15, Load j   )::
(w 16, Load n   )::
(w 17, Push 1  )::
(w 18, Binop Sub)::
(w 19, Load i   )::
(w 20, Binop Sub)::
(w 21, If Ge (w 54))::
(w 22, Load t   )::
(w 23, Load j   )::
(w 24, Push 1  )::
(w 25, Binop Add)::
(w 26, Arrayload)::
(w 27, Load t   )::
(w 28, Load j   )::
(w 29, Arrayload)::
(w 30, If Ge (w 49))::
(w 31, Load t   )::
(w 32, Load j   )::
(w 33, Arrayload)::
(w 34, Store tmp)::
(w 35, Load t   )::
(w 36, Load j   )::
(w 37, Load t   )::
(w 38, Load j   )::
(w 39, Push 1  )::
(w 40, Binop Add)::
(w 41, Arrayload )::
(w 42, Arraystore)::
(w 43, Load t    )::
(w 44, Load j    )::
(w 45, Push 1   )::
(w 46, Binop Add )::
(w 47, Load tmp  )::
(w 48, Arraystore)::
(w 49, Load j    )::
(w 50, Push 1   )::
(w 51, Binop Add)::
(w 52, Store j  )::
(w 53, Goto (w 15))::
(w 54, Load i   )::
(w 55, Push 1    )::
(w 56, Binop Add)::
(w 57, Store i  )::
(w 58, Goto (w 8))::nil.

Definition prog : program :=
(w  1, Push  10 )::
(w  2, Store  n )::nil.

(* Extraction of the two analysers *)
Extraction "coq.ml" bbsort SignAnalysis.verifier IntervalAnalysis.verifier.

(* 
*** Local Variables: ***
*** coq-prog-name: "coqtop" ***
*** coq-prog-args: ("-emacs-U" "-I" "library") ***
*** End: ***
 *)