Require Export interval_op.

Inductive comp : Set := Eq | Lt.
Inductive sem_comp : comp -> Z -> Z -> Prop :=
 | sem_eq : forall n1 n2, n1=n2 -> sem_comp Eq n1 n2
 | sem_le : forall n1 n2, n1<n2 -> sem_comp Lt n1 n2.
Inductive op : Set := Add | Sub | Mult.
Inductive sem_op : op -> Z -> Z -> Z -> Prop :=
 | sem_add : forall n1 n2, sem_op Add n1 n2 (n1+n2)
 | sem_sub : forall n1 n2, sem_op Sub n1 n2 (n1-n2)
 | sem_mult : forall n1 n2, sem_op Mult n1 n2 (n1*n2).

Module IntervalAbstraction.

  Module Lat := IntervalLattice.

  Import IntervalOp.

  Definition gamma : Lat.t -> Z -> Prop := gamma.
  Definition gamma_monotone : forall N1 N2 n,
    Lat.order N1 N2 -> gamma N1 n -> gamma N2 n := gamma_monotone.
  Definition gamma_meet_morph : forall N1 N2 n,
    gamma N1 n -> gamma N2 n -> gamma (Lat.meet N1 N2) n := gamma_meet_morph.

  Definition backtest_eq := BackTestEq.
  Definition backtest_lt := BackTestLt.
  Definition backtest_eq_correct := BackTestEq_correct.
  Definition backtest_lt_correct := BackTestLt_correct.


  Definition sem_add := sem_op_add.
  Definition sem_sub := sem_op_sub.
  Definition sem_mult := sem_op_mult.
  Definition sem_add_correct := sem_op_add_correct.
  Definition sem_sub_correct := sem_op_sub_correct.
  Definition sem_mult_correct := sem_op_mult_correct.

  Definition backsem_add := backsem_add.
  Definition backsem_add_correct := backsem_add_correct.

  Definition backsem_sub := backsem_sub.
  Definition backsem_sub_correct := backsem_sub_correct.

  Definition backsem_mult := backsem_mult.
  Definition backsem_mult_correct := backsem_mult_correct.


  Definition const : Z -> Lat.t := const.
  Definition const_correct : forall n, gamma (const n) n := const_correct.
  Definition const_complete : forall n N,
    gamma N n -> Lat.order (const n) N:= const_complete.

  Definition top : Lat.t := Lat.top.
  Definition top_correct : forall n, gamma top n := top_correct.

  Definition bottom_empty : forall n, ~ gamma Lat.bottom n := bottom_empty.

  Definition check_positive (N : Lat.t) : bool :=
    match N with
      | None => true 
      | Some (ITV a b _) =>
        if order_Z'_dec (z (-1)) a then true else false
    end.

  Lemma check_positive_correct : forall N n,
    check_positive N = true ->
    gamma N n -> n>=0.
  Proof.
    destruct N; simpl; intros.
    destruct i.
    destruct order_Z'_dec; try discriminate.
    inversion H0; subst.
    inversion o0; subst; inversion H2; subst.
    omega.
    inversion H0.
  Qed.

  Definition check_in_range (N : Lat.t) (z1 z2:Z) : bool :=
    match N with
      | None => true 
      | Some (ITV a b _) =>
        if order_Z'_dec (z z1) a then 
          if order_Z'_dec b (z z2) then true
            else false
          else false
    end.

  Ltac inv H := inversion H; try (subst; clear H).

  Lemma check_in_range_correct : forall N z1 z2 n,
    check_in_range N z1 z2 = true ->
    gamma N n -> z1 < n < z2.
  Proof.
    destruct N; simpl; intros.
    destruct i.
    repeat destruct order_Z'_dec; try discriminate.
    inv H0; simpl in *.
    inv H2; inv o0.
    inv H3; inv o1.
    omega.
    inversion H0.
  Qed.

  Definition is_constant (N : Lat.t) : option Z :=
    match N with
      | None => None
      | Some (ITV (z a) (z b) _) =>
        if Z_eq_dec a b then Some a else None
      | _ => None
    end.

  Lemma check_is_constant_correct : forall N z n,
    is_constant N = Some z ->
    gamma N n -> n = z.
  Proof.
    destruct N; simpl; intros.
    destruct i.
    inv H0.
    simpl in *.
    inv H2; inv H3; try discriminate.
    destruct Z_eq_dec; try discriminate.
    subst; inv H; omega.
    discriminate.
  Qed.



End IntervalAbstraction.

(*
Unset Extraction Optimize.
Extraction IntervalAbstraction.
*)