Require Export lattice_def.

Ltac inv H := inversion H; try (subst; clear H).

Module Type Eq.
  Parameter t : Set.

  Parameter eq : t -> t -> Prop.
  Parameter eq_refl : forall x : t, eq x x.
  Parameter eq_sym : forall x y : t, eq x y -> eq y x.
  Parameter eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Parameter eq_dec : forall x y : t, {eq x y}+{~ eq x y}.
End Eq.

Inductive flat (A:Set) : Set := Top | Base (x:A) | Bot.
Implicit Arguments Top [A].
Implicit Arguments Base [A].
Implicit Arguments Bot [A].

Module MakeFlatLattice (E:Eq) <: Lattice.

  Definition t : Set := flat E.t.

  Inductive eq' : t -> t -> Prop :=
    | eq_Top : eq' Top Top
    | eq_Base : forall x y, E.eq x y -> eq' (Base x) (Base y)
    | eq_Bot : eq' Bot Bot.

  Lemma eq_refl : forall x : t, eq' x x.
  Proof. 
    destruct x; constructor.
    apply E.eq_refl.
  Qed.
  Lemma eq_sym : forall x y : t, eq' x y -> eq' y x.
  Proof. 
    intros x y H; inversion H; try subst; try constructor.
    apply E.eq_sym; auto.
  Qed.

  Lemma eq_trans : forall x y z : t, eq' x y -> eq' y z -> eq' x z.
    intros x y z H; inversion H; subst; intros H1; inversion H1; subst; constructor.
    eapply E.eq_trans; eauto.
  Qed.

  Lemma eq_dec : forall x y : t, {eq' x y}+{~ eq' x y}.
  Proof.
    destruct x; destruct y; 
      try (left;constructor; fail);
        try (right; intros T; inversion T; fail).
    destruct (E.eq_dec x x0).
    left; constructor; auto.
    right; intros T; inversion T; elim n; auto.
  Qed.

  Inductive order_ : t -> t -> Prop :=
    | order_Top : forall x, order_ x Top
    | order_Base : forall x y, E.eq x y -> order_ (Base x) (Base y)
    | order_Bot : forall x, order_ Bot x.

  Lemma order_refl : forall x y : t, eq' x y -> order_ x y.
  Proof.
    intros x y H; inv H; constructor; auto.
  Qed.

  Lemma order_antisym : forall x y : t, order_ x y -> order_ y x -> eq' x y.
  Proof.
    intros x y H H0; inv H; inv H0; try constructor; auto.
  Qed.

  Lemma order_trans : forall x y z : t, order_ x y -> order_ y z -> order_ x z.
  Proof.
    intros x y z H H0; inv H; inv H0; try constructor; auto.
    eapply E.eq_trans; eauto.
  Qed.

  Lemma order_dec : forall x y : t, {order_ x y}+{~ order_ x y}.
  Proof.
    destruct x; destruct y;
      try (left;constructor; fail);
        try (right; intros T; inversion T; fail).
    destruct (E.eq_dec x x0).
    left; constructor; auto.
    right; intros T; inversion T; elim n; auto.
  Qed.

  Hint Immediate eq_sym.
  Hint Resolve eq_refl eq_trans 
               order_refl order_antisym order_trans.

  Function join (x y:t) : t :=
    match x,y with
      | Top, _ => Top
      | _, Top => Top
      | Bot, _ => y
      | _, Bot => x
      | Base x, Base y =>
        if E.eq_dec x y then Base x else Top
    end.
  Lemma join_bound1 : forall x y : t, order_ x (join x y).
  Proof.
    intros x y; functional induction (join x y); try constructor; auto.
    apply E.eq_refl.
  Qed.
  Lemma join_bound2 : forall x y : t, order_ y (join x y).
  Proof.
    intros x y; functional induction (join x y); try constructor; auto.
    apply E.eq_sym; auto.
  Qed.
  Lemma join_least_upper_bound : forall x y z : t, order_ x z -> order_ y z -> order_ (join x y) z.
  Proof.
    intros x y z; functional induction (join x y); try constructor; auto.
    intros.
    inv H; inv H0; try constructor.
    elim _x; eapply E.eq_trans; eauto.
    apply E.eq_sym; auto.
  Qed.

  Function meet (x y:t) : t :=
    match x,y with
      | Top, _ => y
      | _, Top => x
      | Bot, _ => Bot
      | _, Bot => Bot
      | Base x, Base y =>
        if E.eq_dec x y then Base x else Bot
    end.

  Lemma meet_bound1 : forall x y : t, order_ (meet x y) x.
  Proof.
    intros x y; functional induction (meet x y); try constructor; auto.
    apply E.eq_refl; auto.
  Qed.
  Lemma meet_bound2 : forall x y : t, order_ (meet x y) y.
  Proof.
    intros x y; functional induction (meet x y); try constructor; auto.
  Qed.
  Lemma meet_greatest_lower_bound : forall x y z : t, order_ z x -> order_ z y -> order_ z (meet x y).
  Proof.
    intros x y z; functional induction (meet x y); try constructor; auto.
    intros.
    inv H; inv H0; try constructor.
    elim _x; eapply E.eq_trans; eauto.
    apply E.eq_sym; auto.
  Qed.

  Definition bottom : t := Bot.
  Lemma bottom_is_bottom : forall x : t, order_ bottom x.
  Proof. constructor. Qed.

  Definition eq := eq'.
  Definition order := order_.

  Hint Resolve join_bound1 join_bound2 join_least_upper_bound
               meet_bound1 meet_bound2 meet_greatest_lower_bound
               bottom_is_bottom.

  Lemma acc :  well_founded (fun x y => ~ eq' y x /\ order_ y x).
  Proof.
    assert (Acc (fun x y0 : t => (eq y0 x -> False) /\ order y0 x) Top).
    constructor; intuition.
    inv H1; elim H0; constructor.
    assert (forall e, Acc (fun x y0 : t => (eq y0 x -> False) /\ order y0 x) (Base e)).
    constructor; intuition.
    inv H2; auto.
    elim H1; constructor; auto.
    constructor; intuition.
    inv H3; auto.
    destruct y; auto.
    elim H2; constructor; auto.
  Qed.

  Lemma dcc :  well_founded (fun x y => ~ eq' y x /\ order_ x y).
  Proof.
    assert (Acc (fun x y0 : t => (eq y0 x -> False) /\ order x y0) Bot).
    constructor; intuition.
    inv H1; elim H0; constructor.
    assert (forall e, Acc (fun x y0 : t => (eq y0 x -> False) /\ order x y0) (Base e)).
    constructor; intuition.
    inv H2; auto.
    elim H1; constructor; auto.
    apply E.eq_sym; auto.
    constructor; intuition.
    inv H3; auto.
    destruct y; auto.
    elim H2; constructor; auto.
  Qed.

  Definition widening : widening_operator eq order bottom.
  Proof.
    apply ascending_chain_condition_to_widening_operator with (join:=join); 
      unfold eq, order; eauto.
    apply acc.
  Qed.

  Definition narrowing : narrowing_operator eq order meet.
  Proof.
    apply descending_chain_condition_to_narrowing_operator;
      unfold eq, order; eauto.
    apply dcc.
  Qed.

End MakeFlatLattice.
