Require Export ProdWiden.
Require Export FuncSignature.
Require Export wf_prop.
Require Export solve.

Module FiniteSetProp (F:FiniteSet).

  Import F.

  Lemma nat2t_inject : forall t, eq t (nat2t (inject t)).
  Proof.
    intros; apply inject_injective.
    rewrite inject_nat2t; auto.
    apply inject_bounded.
  Qed.

End FiniteSetProp.

Module FuncLattice (F : Func_FiniteSet_Lattice) <: Lattice.

  Definition t := F.t.

  Definition eq := fun (f1 f2 : F.t) =>
       forall a1 a2 : F.A.t, F.A.eq a1 a2 -> F.B.eq (F.get f1 a1) (F.get f2 a2).

  Definition eq_refl : forall x : t, eq x x := F.eq_refl.

  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof.
    unfold eq; intuition.
    apply F.B.eq_sym; auto.
  Qed.

  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof.
    unfold eq; intros; eauto.
  Qed.

  Definition eq_dec : forall x y : t, {eq x y}+{~ eq x y} := F.eq_dec.

  Definition order := fun (f1 f2 : F.t) =>
       forall a1 a2 : F.A.t, F.A.eq a1 a2 -> F.B.order (F.get f1 a1) (F.get f2 a2).

  Lemma order_refl : forall x y : t, eq x y -> order x y.
  Proof.
    unfold eq, order; auto.
  Qed.

  Lemma order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Proof.
    unfold eq, order; auto.
  Qed.

  Lemma order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Proof.
    unfold eq, order; intros.
    apply F.B.order_trans with (F.get y a1); auto.
  Qed.

  Definition order_dec : forall x y : t, {order x y}+{~ order x y} := F.order_dec.

  Definition join  : t -> t -> t := F.map2 F.B.join.

  Lemma join_bound1 : forall x y : t, order x (join x y).
  Proof.
    intros f1 f2 a1 a2 H; unfold join.
    apply F.B.order_trans with (F.B.join (F.get f1 a2) (F.get f2 a2)).
    apply F.B.order_trans with (F.get f1 a2); auto.
    apply F.B.order_refl; apply F.eq_refl; auto.
    apply F.B.order_refl; apply F.B.eq_sym.
    apply F.get_map2; auto.
  Qed.

  Lemma join_bound2 : forall x y : t, order y (join x y).
  Proof.
    intros f1 f2 a1 a2 H; unfold join.
    apply F.B.order_trans with (F.B.join (F.get f1 a2) (F.get f2 a2)).
    apply F.B.order_trans with (F.get f2 a2); auto.
    apply F.B.order_refl; apply F.eq_refl; auto.
    apply F.B.order_refl; apply F.B.eq_sym.
    apply F.get_map2; auto.
  Qed.

  Lemma join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z.
  Proof.
    intros f1 f2 f3 H1 H2 a1 a2 H; unfold join.
    apply F.B.order_trans with (F.B.join (F.get f1 a1) (F.get f2 a1)).
    apply F.B.order_refl; apply F.get_map2; auto.
    apply F.B.order_trans with (F.get f3 a1); auto.
    apply F.B.order_refl; apply F.eq_refl; auto.
  Qed.

  Definition meet  : t -> t -> t := F.map2' F.B.meet.

  Lemma meet_bound1 : forall x y : t, order (meet x y) x.
  Proof.
    intros f1 f2 a1 a2 H; unfold meet.
    apply F.B.order_trans with (F.B.meet (F.get f1 a1) (F.get f2 a1)); auto.
    apply F.B.order_refl; apply F.get_map2'; auto.
    apply F.B.order_trans with (F.get f1 a1); auto.
    apply F.B.order_refl; apply F.eq_refl; auto.
  Qed.

  Lemma meet_bound2 : forall x y : t, order (meet x y) y.
  Proof.
    intros f1 f2 a1 a2 H; unfold meet.
    apply F.B.order_trans with (F.B.meet (F.get f1 a1) (F.get f2 a1)); auto.
    apply F.B.order_refl; apply F.get_map2'; auto.
    apply F.B.order_trans with (F.get f2 a1); auto.
    apply F.B.order_refl; apply F.eq_refl; auto.
  Qed.

  Lemma meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y).
  Proof.
    intros f1 f2 f3 H1 H2 a1 a2 H; unfold meet.
    apply F.B.order_trans with (F.B.meet (F.get f1 a2) (F.get f2 a2)).
    apply F.B.order_trans with (F.get f3 a2); auto.
    apply F.B.order_refl; apply F.eq_refl; auto.
    apply F.B.order_refl; apply F.B.eq_sym; apply F.get_map2'; auto.
  Qed.

  Definition bottom := F.bottom.
 
  Lemma bottom_is_bottom : forall x : t, order bottom x.
  Proof.
    intros x a1 a2 H; unfold bottom.
    apply F.B.order_trans with F.B.bottom; auto.
    apply F.B.order_refl; apply F.bottom_eq_bottom.
  Qed.

  Lemma ascending_chain_condition :
   well_founded (fun x y => ~ F.B.eq y x /\ F.B.order y x) ->
   well_founded (fun x y => ~ eq y x /\ order y x).
  Proof.
    intros wf.
    set (F := fun (f : t) (n : nat) => F.get f (F.A.nat2t (Z_of_nat n))). 
    elim (Z_of_nat_complete_inf F.A.cardinal).
    intros max Hmax.
    apply wf_mes with
       (ltB:=(tab.lt _ (fun x y => ~ F.B.eq y x /\ F.B.order y x) F.B.eq max))
       (mes:=F).
    unfold tab.lt, F; intros f2 f1 (H1,H2).
    unfold eq in H1.
    assert (~ forall n, 0 <= n < F.A.cardinal -> F.B.eq (F.get f1 (F.A.nat2t n)) (F.get f2 (F.A.nat2t n)) ).
    intro; elim H1; clear H1; intros a1 a2 H'.    
    apply F.B.eq_trans with (F.get f1 (F.A.nat2t (F.A.inject a1))).
    apply eq_refl.
    apply F.A.inject_injective.
    rewrite F.A.inject_nat2t; auto.
    apply F.A.inject_bounded.
    apply F.B.eq_trans with (F.get f2 (F.A.nat2t (F.A.inject a2))).
    rewrite (F.A.inject_comp_eq a1 a2); auto.
    apply H; apply F.A.inject_bounded.
    apply eq_refl.
    apply F.A.inject_injective.
    rewrite F.A.inject_nat2t; auto.
    apply F.A.inject_bounded.
    generalize (F.A.inject_nat2t); rewrite Hmax; intros Hyp2.
    rewrite Hmax in H.
    clear F H1 Hmax.
    induction max.
    elim H; intros.
    simpl in H0.
    absurd (n<0); omega.
    case (F.B.eq_dec (F.get f1 (F.A.nat2t (Z_of_nat max))) (F.get f2 (F.A.nat2t (Z_of_nat max)))); intros.
    elim IHmax.
    intros N (H1,(H3,(H4,H5))).
    exists N; intuition.
    elim (le_lt_or_eq k max); auto; intros.
    subst; auto.
    omega.
    intro; elim H; intuition.
    elim (Zle_lt_or_eq n (Z_of_nat max)); auto; intros.
    subst; auto.
    replace (Z_of_nat (S max)) with ((Z_of_nat max)+1) in H4.
    omega.
    rewrite inj_S. 
    omega.
    intros; apply Hyp2; split; intuition.
    rewrite inj_S; omega.
    exists max; intuition.
    absurd (~ (k > max)%nat); omega.
    apply tab.Acc_n.
    apply F.B.eq_dec.
    constructor; intro; eauto.
    intuition.
    elim H1; apply F.B.eq_trans with a3; auto.
    apply F.B.order_trans with a2; auto.    
    apply wf.
    generalize (F.A.cardinal_positive); omega.
  Qed.  


  Notation w := F.B.widening.
  Notation n := F.B.narrowing.

  Definition widen := F.map2 (lattice_def.widen w).

  Lemma widen_bound1 : forall x y : t, order x (widen x y).
  Proof.
    intros f1 f2 a1 a2 H.
    apply F.B.order_trans with (lattice_def.widen w (F.get f1 a2) (F.get f2 a2)).
    apply F.B.order_trans with (F.get f1 a2); auto.
    apply order_refl; auto.
    apply eq_refl.
    apply lattice_def.widen_bound1.
    apply F.B.order_refl; apply F.B.eq_sym.
    unfold widen; apply F.get_map2.
    apply widen_bottom2.
    apply widen_bottom1.
  Qed.

  Lemma widen_bound2 : forall x y : t, order y (widen x y).
  Proof.
    intros f1 f2 a1 a2 H.
    apply F.B.order_trans with (lattice_def.widen w (F.get f1 a2) (F.get f2 a2)).
    apply F.B.order_trans with (F.get f2 a2); auto.
    apply order_refl; auto.
    apply eq_refl.
    apply lattice_def.widen_bound2.
    apply F.B.order_refl; apply F.B.eq_sym.
    unfold widen; apply F.get_map2.
    apply widen_bottom2.
    apply widen_bottom1.
  Qed.

  Lemma widen_eq1 : forall x y z: t, 
    eq x y -> eq (widen x z) (widen y z).
  Proof.
    intros f1 f2 f3 H a1 a2 H0.
    apply F.B.eq_trans with (F.get (widen f1 f3) a2).
    apply F.eq_refl; auto.
    apply F.B.eq_trans with (lattice_def.widen w (F.get f1 a2) (F.get f3 a2)).
    unfold widen; apply F.get_map2.
    apply widen_bottom2.
    apply widen_bottom1.
    apply F.B.eq_trans with (lattice_def.widen w (F.get f2 a2) (F.get f3 a2)); auto.
    apply lattice_def.widen_eq1; auto.
    apply F.B.eq_sym.
    unfold widen; apply F.get_map2.    
    apply widen_bottom2.
    apply widen_bottom1.
  Qed.

  Lemma widen_eq2 : forall x y z: t, 
    eq x y -> eq (widen z x) (widen z y).
  Proof.
    intros f1 f2 f3 H a1 a2 H0.
    apply F.B.eq_trans with (F.get (widen f3 f1) a2).
    apply F.eq_refl; auto.
    apply F.B.eq_trans with (lattice_def.widen w (F.get f3 a2) (F.get f1 a2)).
    unfold widen; apply F.get_map2.
    apply widen_bottom2.
    apply widen_bottom1.
    apply F.B.eq_trans with (lattice_def.widen w (F.get f3 a2) (F.get f2 a2)); auto.
    apply lattice_def.widen_eq2; auto.
    apply F.B.eq_sym.
    unfold widen; apply F.get_map2.    
    apply widen_bottom2.
    apply widen_bottom1.
  Qed.

  Lemma widen_bottom1 : forall x : t, eq x (widen x bottom).
  Proof.
    intros; unfold widen, bottom.
    intros a1 a2 H.
    apply F.B.eq_trans with (lattice_def.widen w (F.get x a2) (F.get F.bottom a2)).
    apply F.B.eq_trans with (lattice_def.widen w (F.get x a2) F.B.bottom).
    apply F.B.eq_trans with (F.get x a2).
    apply eq_refl; auto.
    apply widen_bottom1.
    apply lattice_def.widen_eq2.
    apply F.B.eq_sym; apply F.bottom_eq_bottom.
    apply F.B.eq_sym; apply F.get_map2.
    intros; apply widen_bottom2.
    intros; apply widen_bottom1.
  Qed.

  Lemma widen_bottom2 : forall x : t, eq x (widen bottom x).
  Proof.
    intros; unfold widen, bottom.
    intros a1 a2 H.
    apply F.B.eq_trans with (lattice_def.widen w (F.get F.bottom a2) (F.get x a2)).
    apply F.B.eq_trans with (lattice_def.widen w F.B.bottom (F.get x a2)).
    apply F.B.eq_trans with (F.get x a2).
    apply eq_refl; auto.
    apply widen_bottom2.
    apply lattice_def.widen_eq1.
    apply F.B.eq_sym; apply F.bottom_eq_bottom.
    apply F.B.eq_sym; apply F.get_map2.
    intros; apply widen_bottom2.
    intros; apply lattice_def.widen_bottom1.
  Qed.

  Definition narrow := F.map2'' (lattice_def.narrow n).

  Lemma narrow_bound1 : forall x y : t, order (narrow x y) x.
  Proof.
    intros f1 f2 a1 a2 H'.
    apply F.B.order_trans with (lattice_def.narrow n (F.get f1 a1) (F.get f2 a1)).
    apply F.B.order_refl.
    unfold narrow; apply F.get_map2''.
    apply F.B.eq_sym; auto.
    apply narrow_x_x; eauto.
    apply F.B.order_trans with (F.get f1 a1); auto.
    apply lattice_def.narrow_bound1; auto.
    apply order_refl; auto.
    apply eq_refl.
  Qed.

  Lemma narrow_bound2 : forall x y : t, order (meet x y) (narrow x y).
  Proof.
    intros f1 f2 a1 a2 H.
    apply F.B.order_trans with (F.B.meet (F.get f1 a1) (F.get f2 a1)).
    unfold meet; apply F.B.order_refl; apply F.get_map2'; auto.
    apply F.B.order_trans with (F.get (narrow f1 f2) a1).
    apply F.B.order_trans with (lattice_def.narrow n (F.get f1 a1) (F.get f2 a1)).
    apply lattice_def.narrow_bound2; auto.
    apply F.B.order_refl; apply F.B.eq_sym.
    unfold narrow; apply F.get_map2''.
    apply F.B.eq_sym; apply narrow_x_x; eauto.
    apply F.B.order_refl; apply eq_refl; auto.
  Qed.

  Lemma narrow_eq1 : forall x y z: t, 
    eq x y -> eq (narrow x z) (narrow y z).
  Proof.
    intros f1 f2 f3 H a1 a2 H0.
    apply F.B.eq_trans with (F.get (narrow f1 f3) a2).
    apply F.eq_refl; auto.
    apply F.B.eq_trans with (lattice_def.narrow n (F.get f1 a2) (F.get f3 a2)).
    unfold narrow; apply F.get_map2''.
    apply F.B.eq_sym; apply narrow_x_x; eauto.
    apply F.B.eq_trans with (lattice_def.narrow n (F.get f2 a2) (F.get f3 a2)); auto.
    apply lattice_def.narrow_eq1; auto.
    apply F.B.eq_sym.
    unfold narrow; apply F.get_map2''.    
    apply F.B.eq_sym; apply narrow_x_x; eauto.
  Qed.

  Lemma narrow_eq2 : forall x y z: t, 
    eq x y -> eq (narrow z x) (narrow z y).
  Proof.
    intros f1 f2 f3 H a1 a2 H0.
    apply F.B.eq_trans with (F.get (narrow f3 f1) a2).
    apply F.eq_refl; auto.
    apply F.B.eq_trans with (lattice_def.narrow n (F.get f3 a2) (F.get f1 a2)).
    unfold narrow; apply F.get_map2''.
    apply F.B.eq_sym; apply narrow_x_x; eauto.
    apply F.B.eq_trans with (lattice_def.narrow n (F.get f3 a2) (F.get f2 a2)); auto.
    apply lattice_def.narrow_eq2; auto.
    apply F.B.eq_sym.
    unfold narrow; apply F.get_map2''.    
    apply F.B.eq_sym; apply narrow_x_x; eauto.
  Qed.

  Definition PosWidF := PosWid_constr
    F.t eq eq_refl eq_sym eq_trans eq_dec
        order order_refl order_antisym order_trans order_dec
        meet meet_bound1 meet_bound2 meet_greatest_lower_bound
        widen widen_bound1 widen_bound2 widen_eq1 widen_eq2
        narrow narrow_bound1 narrow_bound2 narrow_eq1 narrow_eq2.

  Definition PosWidB := PosWid_constr
    F.B.t F.B.eq F.B.eq_refl F.B.eq_sym F.B.eq_trans F.B.eq_dec
      F.B.order F.B.order_refl F.B.order_antisym F.B.order_trans F.B.order_dec
      F.B.meet F.B.meet_bound1 F.B.meet_bound2 F.B.meet_greatest_lower_bound
      (lattice_def.widen w) (lattice_def.widen_bound1 w) (lattice_def.widen_bound2 w) (lattice_def.widen_eq1 w) (lattice_def.widen_eq2 w)
      (lattice_def.narrow n) (lattice_def.narrow_bound1 n) (lattice_def.narrow_bound2 n) (lattice_def.narrow_eq1 n) (lattice_def.narrow_eq2 n).

  Open Scope Z_scope.

  Lemma Z_of_nat_Zabs_nat : forall x, x >=0 ->
    Z_of_nat (Zabs_nat x) = x.
  Proof.
    destruct x; intros; simpl; auto.
    rewrite Zpos_eq_Z_of_nat_o_nat_of_P; auto.
    destruct H; auto.
  Qed.

  Lemma Zabs_nat_Z_of_nat : forall k,
    Zabs_nat (Z_of_nat k) = k.
  Proof.
    induction k; simpl; intros; auto.
    rewrite nat_of_P_o_P_of_succ_nat_eq_succ; auto.
  Qed.

  Definition widen_rel : (t*t) -> (t*t) -> Prop := fun x y =>
    eq (snd x) (widen (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Lemma widen_acc_property : forall x:t, Acc widen_rel (x,x).
  Proof.
    cut (forall x:t, Acc (ProdWiden.widen_rel PosWidF) (x,x)); intros; auto.
    set (F := fun (f : t) (n : nat) => F.get f (F.A.nat2t (Z_of_nat n))). 
    elim (Z_of_nat_complete_inf F.A.cardinal).
    intros max Hmax.
    apply widen_acc1 with
       (P2:=(tabwid.FuncPosWid _ PosWidB max))
       (mes:=F).
    unfold F; intros f2 f1 H1 i H2.
    apply H1; auto.
    unfold F; intros f1 f2 H1.
    unfold PosWidF, eq in H1; simpl in H1.
    assert (~ forall n, 0 <= n < F.A.cardinal -> F.B.eq (F.get f1 (F.A.nat2t n)) (F.get f2 (F.A.nat2t n)) ).
    intro; elim H1; clear H1; intros.
    apply F.B.eq_trans with (F.get f1 (F.A.nat2t (F.A.inject a1))).
    apply eq_refl.
    apply F.A.inject_injective.
    rewrite F.A.inject_nat2t; auto.
    apply F.A.inject_bounded.
    apply F.B.eq_trans with (F.get f2 (F.A.nat2t (F.A.inject a2))).
    rewrite (F.A.inject_comp_eq a1 a2); auto.
    apply H; apply F.A.inject_bounded.
    apply eq_refl.
    apply F.A.inject_injective.
    rewrite F.A.inject_nat2t; auto.
    apply F.A.inject_bounded.
    intro; elim H; intros.
    generalize (H0 (Zabs_nat n)).
    rewrite Z_of_nat_Zabs_nat; intros.
    apply H3.
    rewrite <- (Zabs_nat_Z_of_nat max).
    apply Zabs_nat_lt.
    rewrite <- Hmax; omega.
    omega.
    intros t1 t2; unfold F.
    intros n' Hn; unfold tabwid.FuncPosWid in *; simpl in *.
    unfold widen; apply F.get_map2.
    apply lattice_def.widen_bottom2.
    apply lattice_def.widen_bottom1.
    apply tabwid.Acc_wid_n.
    unfold widen_acc; intros.
    apply (lattice_def.widen_acc_property w x0).
    generalize (F.A.cardinal_positive); omega.
  Qed. 

  Definition narrow_rel : (t*t) -> (t*t) -> Prop := fun x y =>
    eq (snd x) (narrow (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Lemma narrow_acc_property : forall x:t, Acc narrow_rel (x,x).
  Proof.
    cut (forall x:t, Acc (ProdWiden.narrow_rel PosWidF) (x,x)); intros; auto.
    set (F := fun (f : t) (n : nat) => F.get f (F.A.nat2t (Z_of_nat n))). 
    elim (Z_of_nat_complete_inf F.A.cardinal).
    intros max Hmax.
    apply narrow_acc1 with
       (P2:=(tabwid.FuncPosWid _ PosWidB max))
       (mes:=F).
    unfold F; intros f2 f1 H1 i H2.
    apply H1; auto.
    unfold F; intros f1 f2 H1.
    unfold PosWidF, eq in H1; simpl in H1.
    assert (~ forall n, 0 <= n < F.A.cardinal -> F.B.eq (F.get f1 (F.A.nat2t n)) (F.get f2 (F.A.nat2t n)) ).
    intro; elim H1; clear H1; intros.
    apply F.B.eq_trans with (F.get f1 (F.A.nat2t (F.A.inject a1))).
    apply eq_refl.
    apply F.A.inject_injective.
    rewrite F.A.inject_nat2t; auto.
    apply F.A.inject_bounded.
    apply F.B.eq_trans with (F.get f2 (F.A.nat2t (F.A.inject a2))).
    rewrite (F.A.inject_comp_eq a1 a2); auto.
    apply H; apply F.A.inject_bounded.
    apply eq_refl.
    apply F.A.inject_injective.
    rewrite F.A.inject_nat2t; auto.
    apply F.A.inject_bounded.
    intro; elim H; intros.
    generalize (H0 (Zabs_nat n)).
    rewrite Z_of_nat_Zabs_nat; intros.
    apply H3.
    rewrite <- (Zabs_nat_Z_of_nat max).
    apply Zabs_nat_lt.
    rewrite <- Hmax; omega.
    omega.
    intros t1 t2; unfold F.
    intros n' Hn; unfold tabwid.FuncPosWid in *; simpl in *.
    unfold narrow; apply F.get_map2''.
    apply F.B.eq_sym; apply narrow_x_x; eauto.
    apply tabwid.Acc_nar_n.
    unfold narrow_acc; intros.
    apply (lattice_def.narrow_acc_property n).
    generalize (F.A.cardinal_positive); omega.
  Qed. 

  Definition widening : widening_operator eq order bottom :=
   WidOp (eq:=eq) order bottom
     widen_bound1 widen_bound2 widen_eq1 widen_eq2
     widen_bottom1 widen_bottom2 widen_acc_property. 
 
  Definition narrowing : narrowing_operator eq order meet :=
   NarOp (eq:=eq) order meet
     narrow_bound1 narrow_bound2 narrow_eq1 narrow_eq2 narrow_acc_property. 
 
End FuncLattice.



