Require Export ZArith.
Require Export lattice_def.
Require Export Word.

Open Scope Z_scope.

Module Type FiniteSet.

  Parameter t : Set.

  Parameter eq : t -> t -> Prop.
  Parameter eq_refl : forall x : t, eq x x.
  Parameter eq_sym : forall x y : t, eq x y -> eq y x.
  Parameter eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Parameter eq_dec : forall x y : t, {eq x y}+{~ eq x y}.

  Parameter cardinal : Z.
  Parameter cardinal_positive : cardinal > 0.

  Parameter inject : t -> Z.
  Parameter nat2t : Z -> t.
  Parameter inject_bounded : forall x : t, 0 <= (inject x) < cardinal.
  Parameter inject_nat2t : forall n : Z, 
     0 <= n < cardinal -> inject (nat2t n) = n.      
  Parameter inject_injective : forall x y : t, inject x = inject y -> eq x y.
  Parameter inject_comp_eq : forall x y : t, eq x y -> inject x = inject y.

  Hint Resolve eq_refl eq_sym eq_trans cardinal_positive.

End FiniteSet.

Module Type Func_FiniteSet_Lattice.

  Declare Module A : FiniteSet.
  Declare Module B : Lattice.

  Parameter t : Set.

  Parameter get : t -> A.t -> B.t.

  Definition eq : t -> t -> Prop := fun f1 f2 =>
    forall a1 a2 : A.t, A.eq a1 a2 -> B.eq (get f1 a1) (get f2 a2).

  Parameter eq_refl : forall x : t, eq x x.
  Parameter eq_dec : forall x y : t, {eq x y}+{~ eq x y}.

  Definition order : t -> t -> Prop := fun f1 f2 =>
    forall a1 a2 : A.t, A.eq a1 a2 -> B.order (get f1 a1) (get f2 a2).

  Parameter order_dec : forall x y : t, {order x y}+{~ order x y}.

  Parameter map2 : (B.t -> B.t -> B.t) -> t -> t -> t.
  Parameter get_map2 : forall (f : B.t -> B.t -> B.t),
    (forall x : B.t, B.eq x (f B.bottom x)) ->
    (forall x : B.t, B.eq x (f x B.bottom)) ->
    forall (t1 t2 : t) (a : A.t), B.eq (get (map2 f t1 t2) a) (f (get t1 a) (get t2 a)).

  Parameter map2' : (B.t -> B.t -> B.t) -> t -> t -> t.
  Parameter get_map2' : forall (f : B.t -> B.t -> B.t),
    (forall x : B.t, B.eq B.bottom (f B.bottom x)) ->
    (forall x : B.t, B.eq B.bottom (f x B.bottom)) ->
    forall (t1 t2 : t) (a : A.t), B.eq (get (map2' f t1 t2) a) (f (get t1 a) (get t2 a)).

  Parameter map2'' : (B.t -> B.t -> B.t) -> t -> t -> t.
  Parameter get_map2'' : forall (f : B.t -> B.t -> B.t),
    B.eq B.bottom (f B.bottom B.bottom) ->
    forall (t1 t2 : t) (a : A.t), B.eq (get (map2'' f t1 t2) a) (f (get t1 a) (get t2 a)).

  Parameter map2i : (A.t -> B.t -> B.t -> B.t) -> t -> t -> t.
  Parameter get_map2i : forall (f : A.t -> B.t -> B.t -> B.t),
    (forall a1 a2, A.eq a1 a2 -> f a1 = f a2) ->
    (forall x a, B.eq x (f a B.bottom x)) ->
    (forall x a, B.eq x (f a x B.bottom)) ->
    forall (t1 t2 : t) (a : A.t),
       B.eq (get (map2i f t1 t2) a) (f a (get t1 a) (get t2 a)).

  Parameter map2i' : (A.t -> B.t -> B.t -> B.t) -> t -> t -> t.
  Parameter get_map2i' : forall (f : A.t -> B.t -> B.t -> B.t),
    (forall a1 a2, A.eq a1 a2 -> f a1 = f a2) ->
    (forall a, B.eq B.bottom (f a B.bottom B.bottom)) ->
    forall (t1 t2 : t) (a : A.t),
       B.eq (get (map2i' f t1 t2) a) (f a (get t1 a) (get t2 a)).

  Parameter modify : t -> A.t -> (B.t -> B.t) -> t.
  Parameter get_modify1 : forall f a2 a1 fv,
    A.eq a2 a1 ->
    get (modify f a1 fv) a2 = fv (get f a2).
  Parameter get_modify2 : forall f a2 a1 fv,
    ~ A.eq a2 a1 ->
    get (modify f a1 fv) a2 = get f a2.

  Parameter bottom : t.
  Parameter bottom_eq_bottom : forall a, B.eq (get bottom a) B.bottom. 

End Func_FiniteSet_Lattice.

Module WordFiniteSet <: FiniteSet
  with Definition t := Word
  with Definition eq := @eq Word.

  Definition t := Word.

  Definition eq := @eq Word.

  Definition eq_refl : forall x : t, eq x x := @refl_equal _.
  Definition eq_sym : forall x y : t, eq x y -> eq y x := @sym_eq _.
  Definition eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z := @trans_eq _.
  Definition eq_dec : forall x y : t, {eq x y}+{~ eq x y} := eq_word_dec.

  Definition inject : t -> Z := fun w =>
    let (p,_) := w in (Zpred (Zpos p)).

  Definition nat2t : Z -> t := fun z =>
    match z with
       Z0 => Word_1
     | Zpos p => match (inf_log_dec (Psucc p) WordSize) with
                    left H => exP (Psucc p) (inf_log_inf_log_bool _ _ H)
                  | right _ => Word_1
                 end
     | Zneg p => Word_1
    end.

  Fixpoint height_pos (p:positive) : nat :=
    match p with
      xH => O
    | xI p => S (height_pos p)
    | xO p => S (height_pos p)
    end.

  Fixpoint pow2P (n:nat) : positive :=
    match n with
      O => xH
    | S p => xI (pow2P p)
    end.

  Lemma Zpos_xO : forall p, Zpos (xO p) = 2 * (Zpos p).
  Proof.
    auto with zarith.
  Qed.

  Lemma Zpos_xI : forall p, Zpos (xI p) = 2 * (Zpos p) + 1.
  Proof.
    auto with zarith.
  Qed.

  Lemma pow2P_bigger_height : forall p,
    (Zpos p) <= (Zpos (pow2P (height_pos p))).
  Proof.
    induction p; simpl; auto.
    rewrite Zpos_xI; rewrite Zpos_xO.
    auto with zarith.
    auto with zarith.
  Qed.

  Lemma height_word : forall w : Word, (height_pos w <= WordSize)%nat.
  Proof.
    destruct w; simpl.
    generalize dependent WordSize.
    intros n H.
    generalize (inf_log_bool_inf_log _ _ H).
    generalize n; clear n H.
    induction x; simpl; intros.
    inversion_clear H; auto with arith.
    inversion_clear H; auto with arith.
    auto with arith.
  Qed.

  Lemma pow2P_monotone : forall n1 n2, 
    (n1 <= n2)%nat ->(Zpos (pow2P n1)) <= (Zpos (pow2P n2)).
  Proof.
    induction 1; simpl.
    auto with zarith.
    rewrite Zpos_xI.
    generalize (Zle_0_pos (pow2P m)); intros.
    omega.
  Qed.

  Lemma word_bounded : forall w : Word,
    (Zpos w) <= (Zpos (pow2P WordSize)).
  Proof. 
    intros w; apply Zle_trans with (Zpos (pow2P (height_pos w))).
    apply pow2P_bigger_height.
    apply pow2P_monotone.
    apply height_word.
  Qed.

  Definition cardinal : Z := (Zpos (pow2P WordSize)).

  Lemma cardinal_positive : cardinal > 0.
  Proof.
    unfold cardinal; generalize (Zgt_pos_0 (pow2P WordSize)).
    omega.
  Qed.

  Lemma inject_bounded : forall x : t, 0 <= (inject x) < cardinal.
  Proof.
    destruct x; unfold inject, cardinal in *.
    split.
    destruct x; simpl; auto with zarith.
    cut (Zpos x <= Zpos (pow2P WordSize)); intros.
    unfold Zpred.
    omega.
    generalize (word_bounded (exP x e)); auto.
  Qed.


  Lemma le_pow2P_inf_log : forall p n,
    Zpos p <= Zpos (pow2P n) -> inf_log p n.
  Proof.
    induction p; intros.
    rewrite Zpos_xI in H.
    destruct n.
    simpl in H.
    rewrite Zpos_xI in H.
    generalize (Zgt_pos_0 p); intros.
    absurd (Zpos p <= 0); omega.
    simpl in H.
    do 2 rewrite Zpos_xI in H.
    constructor; apply IHp; omega.
    destruct n.
    simpl in H; rewrite Zpos_xO in H.
    generalize (Zgt_pos_0 p); intros.
    absurd (Zpos p <= 0); omega.
    simpl in H.
    rewrite Zpos_xI in H; rewrite Zpos_xO in H.
    constructor; apply IHp; omega.
    constructor.
  Qed.

  Lemma inject_nat2t : forall n : Z, 
     0 <= n < cardinal -> inject (nat2t n) = n.      
  Proof.
    intros; unfold inject, nat2t.
    destruct n.
    unfold Word_1; auto.
    case inf_log_dec; unfold exP; intros.
    rewrite Zpos_succ_morphism.
    rewrite Zpred_succ; auto.
    elim n; clear n.
    apply le_pow2P_inf_log.
    rewrite Zpos_succ_morphism. 
    unfold cardinal in H.
    unfold Zsucc; omega.
    generalize (Zlt_neg_0 p); intros.
    absurd (Zneg p >= 0); omega.
  Qed.

  Lemma inject_injective : forall x y : t, inject x = inject y -> eq x y.
  Proof.
    intros (x,Hx) (y,hy); unfold inject; intros; simpl.
    assert (Zpos x = Zpos y).
    rewrite (Zsucc_pred (Zpos x)).
    rewrite (Zsucc_pred (Zpos y)).
    rewrite H; auto.
    inversion H0; subst.
    replace hy with Hx; auto.
    unfold eq; auto.
    apply bool_proof_irrelevance.
  Qed.

  Lemma inject_comp_eq : forall x y : t, eq x y -> inject x = inject y.
  Proof.
    intros (x,Hx) (y,hy); unfold eq,inject; intros.
    inversion H.
    subst; auto.
  Qed.

End WordFiniteSet.