Require Export partial_lattice.
Require Export change_implem.
Require Export FlatLattice.
Require Export List.

Module list.

Section eq.

Variable A : Set.
Variable eqA : A -> A -> Prop.

Inductive eq : list A -> list A -> Prop :=
  eq_nil : eq nil nil
| eq_cons : forall a1 a2 q1 q2,
         eqA a1 a2 ->
         eq q1 q2 ->
         eq (a1::q1) (a2::q2).

Variable orderA : A -> A -> Prop.

Inductive order : list A -> list A -> Prop :=
  order_nil : order nil nil
| order_cons : forall a1 a2 q1 q2,
         orderA a1 a2 ->
         order q1 q2 ->
         order (a1::q1) (a2::q2).

Lemma app_monotone : forall l1 l2 l3 l4,
  order l1 l2 -> order l3 l4 -> order (l1++l3) (l2++l4).
Proof.
  induction 1; simpl; try constructor; auto.
Qed.

Lemma order_same_length : forall l1 l2,
  order l1 l2 -> length l1 = length l2.
Proof.
  induction 1; simpl; auto.
Qed.

Lemma eq_app_1 : forall l1 l2 l3 l4,
  order l1 l2 -> eq (l1++l3) (l2++l4) -> eq l1 l2.
Proof.
  induction 1; simpl; constructor.
  inv H1; auto.
  inv H1; auto.
Qed.

Lemma eq_app_2 : forall l1 l2 l3 l4,
  order l1 l2 -> eq (l1++l3) (l2++l4) -> eq l3 l4.
Proof.
  induction 1; simpl; intros; auto.
  inv H1; auto.
Qed.

End eq.

Section listf.

Variable A : Set.
Variable f : A -> A -> A.

Fixpoint list_f (l1 l2 : list A) {struct l1} : option (list A) := 
  match (l1,l2) with
    (nil,nil) => Some nil
  | (a1::q1,a2::q2) => match (list_f q1 q2) with
      None => None
    | Some l => Some ((f a1 a2)::l)
    end
  | _ => None
  end.

End listf.

End list.

Module ListPartialLattice (L:Lattice) <: PartialLattice
 with Definition t := list L.t
 with Definition eq := list.eq L.t L.eq
 with Definition order := list.order L.t L.order.

  Definition t := list L.t.

  Definition eq : t -> t -> Prop := list.eq L.t L.eq.
 
  Lemma eq_refl : forall x : t, eq x x.
  Proof.
    induction x; constructor; auto.
  Qed.

  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof.
    induction 1; constructor; auto.
  Qed.

  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof.
    intros x y z H; generalize dependent z.
    induction H; intros z H'; inversion_clear H'; constructor; eauto.
    apply IHeq; auto.
  Qed.

  Lemma eq_dec : forall x y : t, {eq x y}+{~ eq x y}.
  Proof.
    induction x; destruct y.
    left; constructor.
    right; intros H; inversion_clear H.
    right; intros H; inversion_clear H.
    case (L.eq_dec a t0); intros.
    case (IHx y); intros.
    left; constructor; auto.
    right; intros H; inversion_clear H; elim n; auto.
    right; intros H; inversion_clear H; elim n; auto.
  Qed.

  Definition order : t -> t -> Prop := list.order L.t L.order.
 
  Lemma order_refl : forall x y : t, eq x y -> order x y.
  Proof.
    induction 1; constructor; auto.
  Qed.

  Lemma order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Proof.
    induction 1; constructor; auto.
    inversion_clear H1; auto.
    inversion_clear H1; apply IHorder; auto.
  Qed.

  Lemma order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Proof.
    intros x y z H; generalize dependent z.
    induction H; intros z H'; inversion_clear H'; constructor; eauto.
    apply IHorder; auto.
  Qed.

  Lemma order_dec : forall x y : t, {order x y}+{~ order x y}.
  Proof.
    induction x; destruct y.
    left; constructor.
    right; intros H; inversion_clear H.
    right; intros H; inversion_clear H.
    case (L.order_dec a t0); intros.
    case (IHx y); intros.
    left; constructor; auto.
    right; intros H; inversion_clear H; elim n; auto.
    right; intros H; inversion_clear H; elim n; auto.
  Qed.

  Definition join  : t -> t -> option t := list.list_f _ L.join.

  Lemma join_bound1 : forall x y j : t, join x y = Some j ->  order x j.
  Proof.
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail).
    intros H; inversion_clear H; constructor.
    generalize (IHx y); destruct (join x y); intros.
    inversion_clear H0; constructor; auto.
    apply H; auto.
    inversion_clear H0.
  Qed.

  Lemma join_bound2 : forall x y j : t, join x y = Some j ->  order y j.
  Proof.
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail).
    intros H; inversion_clear H; constructor.
    generalize (IHx y); destruct (join x y); intros.
    inversion_clear H0; constructor; auto.
    apply H; auto.
    inversion_clear H0.
  Qed.

  Lemma join_least_upper_bound : forall x y z : t, 
    order x z -> order y z ->
    exists j, join x y = Some j /\ order j z.
  Proof.
    intros x y z H.
    generalize y; clear y.
    induction H; intros y HH; inversion_clear HH; simpl.
    exists (nil (A:=L.t)); split; auto.
    constructor.
    destruct (IHorder _ H2) as [j (H3,H4)].
    rewrite H3.
    exists (L.join a1 a0 :: j); split; auto.
    constructor; auto.
  Qed.

  Definition meet  : t -> t -> option t := list.list_f _ L.meet.

  Lemma meet_bound1 : forall x y m : t, meet x y = Some m -> order m x.
  Proof.
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail).
    intros H; inversion_clear H; constructor.
    generalize (IHx y); destruct (meet x y); intros.
    inversion_clear H0; constructor; auto.
    apply H; auto.
    inversion_clear H0.
  Qed.

  Lemma meet_bound2 : forall x y m : t, meet x y = Some m -> order m y.
  Proof.
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail).
    intros H; inversion_clear H; constructor.
    generalize (IHx y); destruct (meet x y); intros.
    inversion_clear H0; constructor; auto.
    apply H; auto.
    inversion_clear H0.
  Qed.

  Lemma meet_greatest_lower_bound : forall x y z : t, 
    order z x -> order z y -> 
    exists m, meet x y = Some m /\ order z m.
  Proof.
    intros x y z H.
    generalize y; clear y.
    induction H; intros y HH; inversion_clear HH; simpl.
    exists (nil (A:=L.t)); split; auto.
    constructor.
    destruct (IHorder _ H2) as [j (H3,H4)].
    rewrite H3.
    exists (L.meet a2 a3 :: j); split; auto.
    constructor; auto.
  Qed.

  Lemma ascending_chain_condition :
   well_founded (fun x y => ~ L.eq y x /\ L.order y x) ->
   well_founded (fun x y => ~ eq y x /\ order y x).
  Proof.
    intros wf x; induction x.
    constructor; intros y (H1,H2).
    elim H1; inversion_clear H2; constructor.
    generalize dependent a.
    induction IHx.
    clear H.
    intros a; induction (wf a).
    clear H.
    constructor; intros y (H3,H2).
    inversion_clear H2 in H3.
    case (eq_dec x q2); intros.
    apply acc_eq with (eq:=eq) (x:=(a2 :: x)).
    intros x' y' z' (H1',H2') H3'; split.
    red; intros; elim H1'; apply eq_trans with z'; auto.
    apply eq_sym; auto.
    apply order_trans with y'; auto. 
    apply order_refl; auto.
    apply H1; split; auto.
    intro; elim H3; constructor; auto.
    constructor; auto.
    apply H0; split; auto.
  Qed.  


  Inductive relmes : (t*t) -> ((L.t*L.t)*(t*t)) -> Prop :=
   | relmes_def : forall a1 l1 a2 l2,
       relmes (a1::l1,a2::l2) ((a1,a2),(l1,l2)). 

  Definition widening : partial_widening_operator eq order.
  Proof.
    set (w:=L.widening).
    assert(
     forall x y z : t,
     eq x y ->
     option.eq eq (list.list_f L.t (lattice_def.widen w) x z)
       (list.list_f L.t (lattice_def.widen w) y z)).
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail); intros H; inversion_clear H.
    destruct j; repeat constructor.
    destruct j; repeat constructor.
    generalize (IHx _ j H1); intros.
    inversion_clear H; repeat constructor; auto.
    apply lattice_def.widen_eq1; auto.
    rename H into HHH.
    exists (list.list_f _ (lattice_def.widen w)).
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail).
    intros H; inversion_clear H; constructor.
    generalize (IHx y); destruct (list.list_f L.t (lattice_def.widen w) x y); intros.
    inversion_clear H0; constructor; auto.
    apply lattice_def.widen_bound1.
    apply H; auto.
    inversion_clear H0.
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail).
    intros H; inversion_clear H; constructor.
    generalize (IHx y); destruct (list.list_f L.t (lattice_def.widen w) x y); intros.
    inversion_clear H0; constructor; auto.
    apply lattice_def.widen_bound2.
    apply H; auto.
    inversion_clear H0.
    auto.
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail); intros H; inversion_clear H.
    destruct (list.list_f L.t (lattice_def.widen w) j nil); constructor.
    apply eq_refl.
    destruct j; simpl; repeat constructor.
    generalize (IHx _ j H1); intros.
    inversion_clear H; repeat constructor; auto.
    apply (lattice_def.widen_eq2 w ); auto.
    induction x.
    constructor; intros.
    destruct H as [H H0]; simpl in *.
    CaseEq (fst y); intros.
    rewrite H1 in H; inversion_clear H.
    elim H0; apply eq_sym; auto.
    rewrite H1 in H; inversion_clear H.
    set (widen1_rel_stop := fun x y : L.t*L.t => L.eq (snd x) (snd y)).
    set (widen_eq_rel := fun x y : t*t =>
           option.eq eq (Some (snd x)) ((list.list_f L.t (lattice_def.widen w))  (snd y) (fst x))).
    set (widen_rel := (partial_widen_rel eq 
        (list.list_f L.t (lattice_def.widen w)))).
    apply acc_rel with (B:=((L.t*L.t)*(t*t))%type)
                    (rel:=relmes)
                    (ltB := (R _ _ (lattice_def.widen_rel L.eq (lattice_def.widen w))
        widen1_rel_stop widen_rel widen_eq_rel))
                    (b:=((a,a),(x,x))).
    intros.
    inversion_clear H0 in H; inversion_clear H1 in H. 
    destruct H as [H H0]; simpl in *.
    generalize dependent H; CaseEq (list.list_f L.t (lattice_def.widen w) l3 l1); intros.
    rewrite H in H1.
    inversion_clear H1.
    inversion_clear H2.
    destruct (L.eq_dec a5 a3).
    constructor 2.
    repeat split; auto.
    simpl.
    rewrite H; constructor; auto.
    simpl; intro; elim H0; constructor; auto.
    unfold widen1_rel_stop; simpl; auto.
    constructor 1.
    repeat split; auto.
    unfold widen_eq_rel; simpl; auto.
    rewrite H; constructor; auto.
    rewrite H in H1; inversion_clear H1.
    intros.
    inversion_clear H0 in H.
    destruct H as [H H0]; simpl in *.
    destruct a1; simpl in *.
    destruct t0; try (inversion_clear H; fail).
    CaseEq (list.list_f L.t (lattice_def.widen w) l2 t2); intros; rewrite H1 in H; inversion_clear H.
    inversion_clear H2.
    exists ((t0,a1),(t2,q1)); constructor.
    constructor.
    apply R_acc; auto.
    intros; destruct H as [H H1].
    constructor 1.
    repeat split; eauto.
    apply L.eq_trans with (1:= H).
    apply lattice_def.widen_eq1; auto.
    intros (x3,y3) (x2,y2) (x1,y1) (H3,H4) H1; simpl in *. 
    case (eq_dec y2 y1); intros.
    constructor 1; simpl; repeat split; simpl in *.
    generalize (HHH _ _ x3 e); generalize H3.
    case (list.list_f L.t (lattice_def.widen w) y2 x3); intros.
    inversion_clear H0; inversion_clear H; constructor.
    apply eq_trans with l; auto.
    inversion_clear H; auto.
    intro; elim H4; apply eq_trans with y1; auto.
    constructor 2 with (x2,y2); constructor 1; simpl; repeat split; simpl; auto. 
    intro; elim n; apply eq_sym; auto.
    apply (lattice_def.widen_acc_property w).
  Defined.

  Definition narrowing : partial_narrowing_operator eq order meet.
  Proof.
    set (w:=L.narrowing).
    assert(
     forall x y z : t,
     eq x y ->
     option.eq eq (list.list_f L.t (lattice_def.narrow w) x z)
       (list.list_f L.t (lattice_def.narrow w) y z)).
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail); intros H; inversion_clear H.
    destruct j; repeat constructor.
    destruct j; repeat constructor.
    generalize (IHx _ j H1); intros.
    inversion_clear H; repeat constructor; auto.
    apply lattice_def.narrow_eq1; auto.
    rename H into HHH.
    exists (list.list_f _ (lattice_def.narrow w)).
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail).
    intros H; inversion_clear H; constructor.
    generalize (IHx y); destruct (list.list_f L.t (lattice_def.narrow w) x y); intros.
    inversion_clear H0; constructor; auto.
    apply lattice_def.narrow_bound1.
    apply H; auto.
    inversion_clear H0.
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail).
    intros H; inversion_clear H.
    exists (nil (A:=L.t)); repeat constructor.
    generalize (IHx y).
    destruct (meet x y); intros.
    inversion_clear H0.
    destruct (H t1); auto.
    destruct H0.
    rewrite H0.
    exists (lattice_def.narrow w a t0 :: x0); split; auto.
    constructor; auto.
    apply lattice_def.narrow_bound2.
    inversion_clear H0.
    auto.
    induction x; destruct y; simpl; intros j;
    try (intros H; inversion_clear H; fail); intros H; inversion_clear H.
    destruct (list.list_f L.t (lattice_def.narrow w) j nil); constructor.
    apply eq_refl.
    destruct j; simpl; repeat constructor.
    generalize (IHx _ j H1); intros.
    inversion_clear H; repeat constructor; auto.
    apply (lattice_def.narrow_eq2 w ); auto.
    induction x.
    constructor; intros.
    destruct H as [H H0]; simpl in *.
    CaseEq (fst y); intros.
    rewrite H1 in H; inversion_clear H.
    elim H0; apply eq_sym; auto.
    rewrite H1 in H; inversion_clear H.
    set (narrow1_rel_stop := fun x y : L.t*L.t => L.eq (snd x) (snd y)).
    set (narrow_eq_rel := fun x y : t*t =>
           option.eq eq (Some (snd x)) ((list.list_f L.t (lattice_def.narrow w))  (snd y) (fst x))).
    set (narrow_rel := (partial_narrow_rel eq 
        (list.list_f L.t (lattice_def.narrow w)))).
    apply acc_rel with (B:=((L.t*L.t)*(t*t))%type)
                    (rel:=relmes)
                    (ltB := (R _ _ (lattice_def.narrow_rel L.eq (lattice_def.narrow w))
        narrow1_rel_stop narrow_rel narrow_eq_rel))
                    (b:=((a,a),(x,x))).
    intros.
    inversion_clear H0 in H; inversion_clear H1 in H. 
    destruct H as [H H0]; simpl in *.
    generalize dependent H; CaseEq (list.list_f L.t (lattice_def.narrow w) l3 l1); intros.
    rewrite H in H1.
    inversion_clear H1.
    inversion_clear H2.
    destruct (L.eq_dec a5 a3).
    constructor 2.
    repeat split; auto.
    simpl.
    rewrite H; constructor; auto.
    simpl; intro; elim H0; constructor; auto.
    unfold narrow1_rel_stop; simpl; auto.
    constructor 1.
    repeat split; auto.
    unfold narrow_eq_rel; simpl; auto.
    rewrite H; constructor; auto.
    rewrite H in H1; inversion_clear H1.
    intros.
    inversion_clear H0 in H.
    destruct H as [H H0]; simpl in *.
    destruct a1; simpl in *.
    destruct t0; try (inversion_clear H; fail).
    CaseEq (list.list_f L.t (lattice_def.narrow w) l2 t2); intros; rewrite H1 in H; inversion_clear H.
    inversion_clear H2.
    exists ((t0,a1),(t2,q1)); constructor.
    constructor.
    apply R_acc; auto.
    intros; destruct H as [H H1].
    constructor 1.
    repeat split; eauto.
    apply L.eq_trans with (1:= H).
    apply lattice_def.narrow_eq1; auto.
    intros (x3,y3) (x2,y2) (x1,y1) (H3,H4) H1; simpl in *. 
    case (eq_dec y2 y1); intros.
    constructor 1; simpl; repeat split; simpl in *.
    generalize (HHH _ _ x3 e); generalize H3.
    case (list.list_f L.t (lattice_def.narrow w) y2 x3); intros.
    inversion_clear H0; inversion_clear H; constructor.
    apply eq_trans with l; auto.
    inversion_clear H; auto.
    intro; elim H4; apply eq_trans with y1; auto.
    constructor 2 with (x2,y2); constructor 1; simpl; repeat split; simpl; auto. 
    intro; elim n; apply eq_sym; auto.
    apply (lattice_def.narrow_acc_property w).
  Defined.

End ListPartialLattice.

Module ListLattice (LL:Lattice) <: Lattice
  with Definition t := flat (list LL.t).

  Module ListPL := ListPartialLattice LL.
  Module ListL := Lattice_of_PartialLattice ListPL.

  Module Implem <: NewImplemLattice.
    Module L <: Lattice := ListL.
    Definition t : Set := flat (list LL.t).

    Inductive eq' : t -> t -> Prop :=
    | eq_Top : eq' Top Top
    | eq_Base : forall x y, list.eq _ LL.eq x y -> eq' (Base x) (Base y)
    | eq_Bot : eq' Bot Bot.

    Definition eq : t -> t -> Prop := eq'.

    Lemma eq_dec : forall x y : t, {eq x y}+{~ eq x y}.
    Proof.
      destruct x; destruct y; 
        try (left;constructor; fail);
          try (right; intros T; inversion T; fail).
      destruct (ListPL.eq_dec x x0).
      left; constructor; auto.
      right; intros T; inversion T; elim n; auto.
    Qed.

    Inductive order_ : t -> t -> Prop :=
    | order_Top : forall x, order_ x Top
    | order_Base : forall x y, list.order _ LL.order x y -> order_ (Base x) (Base y)
    | order_Bot : forall x, order_ Bot x.

    Definition order : t -> t -> Prop := order_.

    Lemma order_eq1 : forall x y : t, eq x y -> order x y /\ order y x.
    Proof.
      intros.
      inv H; repeat constructor.
      induction H0; constructor; auto.
      induction H0; constructor; auto.
    Qed.

    Lemma order_eq2 : forall x y : t, order x y -> order y x -> eq x y.
    Proof.
      intros.
      inv H; inv H0; repeat constructor.
      generalize H3; clear H3.
      induction H1; intros H3; inv H3; constructor; auto.
    Qed.

    Lemma order_dec : forall x y : t, {order x y}+{~ order x y}.
    Proof.
      destruct x; destruct y;
        try (left;constructor; fail);
          try (right; intros T; inversion T; fail).
      destruct (ListPL.order_dec x x0).
      left; constructor; auto.
      right; intros T; inversion T; elim n; auto.
    Qed.

    Definition join (x y:t) : t :=
      match x,y with
        | Top, _ => Top
        | _, Top => Top
        | Bot, _ => y
        | _, Bot => x
        | Base x, Base y => 
          match ListPL.join x y with
            | Some res => Base res
            | None => Top
          end
      end.

    Definition meet (x y:t) : t :=
      match x,y with
        | Top, _ => y
        | _, Top => x
        | Bot, _ => Bot
        | _, Bot => Bot
        | Base x, Base y => 
          match ListPL.meet x y with
            | Some res => Base res
            | None => Bot
          end
      end.

    Definition widen (x y:t) : t :=
      match x,y with
        | Top, _ => Top
        | _, Top => Top
        | Bot, _ => y
        | _, Bot => x
        | Base x, Base y => 
          match ListPL.widening x y with
            | Some res => Base res
            | None => Top
          end
      end.

    Definition narrow (x y:t) : t :=
      match x,y with
        | Top, _ => y
        | _, Top => x
        | Bot, _ => Bot
        | _, Bot => Bot
        | Base x, Base y => 
          match ListPL.narrowing x y with
            | Some res => Base res
            | None => Bot
          end
      end.

    Definition i (x:t) : L.t :=
      match x with
        | Top => lift_top_bot.top
        | Base x => lift_top_bot.inj x
        | Bot => lift_top_bot.bot
      end.

    Lemma i_order_preserve : forall x y, order x y -> L.order (i x) (i y).
    Proof.
      intros.
      inv H; constructor; auto.
    Qed.

    Lemma inv_i_order_preserve : forall x y, L.order (i x) (i y) -> order x y.
    Proof.
      destruct x; destruct y; intros H; inv H; constructor.
      induction H2; constructor; auto.
    Qed.

    Lemma join_commut : forall x y,
      L.eq (i (join x y)) (L.join (i x) (i y)).
    Proof.
      destruct x; destruct y; simpl; try constructor.
      destruct ListPL.join; constructor.
      apply ListPL.eq_refl.
      apply ListPL.eq_refl.
      apply ListPL.eq_refl.
    Qed.

    Lemma meet_commut : forall x y,
      L.eq (i (meet x y)) (L.meet (i x) (i y)).
    Proof.
      destruct x; destruct y; simpl; try constructor.
      apply ListPL.eq_refl.
      apply ListPL.eq_refl.
      destruct ListPL.meet; constructor.
      apply ListPL.eq_refl.
    Qed.

    Lemma widen_commut : forall x y,
      L.eq (i (widen x y)) (L.widening (i x) (i y)).
    Proof.
      destruct x; destruct y; simpl; try constructor.
      destruct list.list_f; constructor.
      apply ListPL.eq_refl.
      apply ListPL.eq_refl.
      apply ListPL.eq_refl.
    Qed.

    Lemma narrow_commut : forall x y,
      L.eq (i (narrow x y)) (L.narrowing (i x) (i y)).
    Proof.
      destruct x; destruct y; simpl; try constructor.
      apply ListPL.eq_refl.
      apply ListPL.eq_refl.
      destruct list.list_f; constructor.
      apply ListPL.eq_refl.
    Qed.

    Definition bottom : t := Bot.

    Lemma bottom_is_bottom : L.eq L.bottom (i bottom).
    Proof. constructor. Qed.

  End Implem.

  Module New := Lattice_of_NewImplemLattice Implem.

  Include New.

End ListLattice.




