Require Export lattice_def.

Open Scope type_scope.

Module ProdLattice (P1:Lattice) (P2:Lattice) : Lattice
  with Definition t := (P1.t * P2.t)%type
  with Definition eq := fun (x y : (P1.t * P2.t)%type) =>
    P1.eq (fst x) (fst y) /\ P2.eq (snd x) (snd y)
  with Definition order := fun (x y : (P1.t * P2.t)%type) =>
    P1.order (fst x) (fst y) /\ P2.order (snd x) (snd y)
  with Definition join := fun x y =>
    (P1.join (fst x) (fst y),P2.join (snd x) (snd y))
  with Definition meet := fun x y =>
    (P1.meet (fst x) (fst y),P2.meet (snd x) (snd y))
  with Definition bottom := (P1.bottom,P2.bottom).

  Definition t := P1.t * P2.t.

  Definition eq (x y : t) :=
    P1.eq (fst x) (fst y) /\ P2.eq (snd x) (snd y).

  Lemma eq_refl : forall x : t, eq x x.
  Proof.
    intros (x,y); split; auto.
  Qed.

  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof.
    intros (x1,x2) (y1,y2) (H1,H2); split; auto.
  Qed.

  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof.
    intros (x1,x2) (y1,y2) (z1,z2) (H1,H2) (H3,H4); split; eauto.
  Qed.

  Lemma eq_dec : forall x y : t, {eq x y}+{~ eq x y}.
  Proof.
    intros (x1,x2) (y1,y2).
    case (P1.eq_dec x1 y1); intros.
    case (P2.eq_dec x2 y2); intros.
    left; constructor; auto.
    right; intros (H1,H2); elim n; auto. 
    right; intros (H1,H2); elim n; auto. 
  Qed.

  Definition order (x y : t) :=
    P1.order (fst x) (fst y) /\ P2.order (snd x) (snd y).

  Lemma order_refl : forall x y : t, eq x y -> order x y.
  Proof.
    intros (x1,x2) (y1,y2) (H1,H2); split; auto.    
  Qed.

  Lemma order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Proof.
    intros (x1,x2) (y1,y2) (H1,H2) (H3,H4); split; auto.    
  Qed.

  Lemma order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Proof.
    intros (x1,x2) (y1,y2) (z1,z2) (H1,H2) (H3,H4); split; eauto.
  Qed.

  Lemma order_dec : forall x y : t, {order x y}+{~ order x y}.
  Proof.
    intros (x1,x2) (y1,y2).
    case (P1.order_dec x1 y1); intros.
    case (P2.order_dec x2 y2); intros.
    left; constructor; auto.
    right; intros (H1,H2); elim n; auto. 
    right; intros (H1,H2); elim n; auto. 
  Qed.

  Definition join  : t -> t -> t := fun x y =>
    (P1.join (fst x) (fst y),P2.join (snd x) (snd y)).

  Lemma join_bound1 : forall x y : t, order x (join x y).
  Proof.
    intros; unfold join; split; simpl; auto.
  Qed.

  Lemma join_bound2 : forall x y : t, order y (join x y).
  Proof.
    intros; unfold join; split; simpl; auto.
  Qed.

  Lemma join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z.
  Proof.
    intros x y z (H1,H2) (H3,H4); unfold join; split; simpl in *; auto.
  Qed.

  Definition meet  : t -> t -> t := fun x y =>
    (P1.meet (fst x) (fst y),P2.meet (snd x) (snd y)).

  Lemma meet_bound1 : forall x y : t, order (meet x y) x.
  Proof.
    intros; unfold meet; split; simpl; auto.
  Qed.

  Lemma meet_bound2 : forall x y : t, order (meet x y) y.
  Proof.
    intros; unfold meet; split; simpl; auto.
  Qed.

  Lemma meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y).
  Proof.
    intros x y z (H1,H2) (H3,H4); unfold meet; split; simpl in *; auto.
  Qed.

  Definition bottom := (P1.bottom,P2.bottom).

  Lemma bottom_is_bottom : forall x : t, order bottom x.
  Proof.
    unfold bottom; split; simpl; auto.
  Qed.

  Definition w1 := P1.widening.
  Definition w2 := P2.widening.
  Definition n1 := P1.narrowing.
  Definition n2 := P2.narrowing.

  Require ProdWiden.

  Definition PosWid1 := ProdWiden.PosWid_constr
    P1.t P1.eq P1.eq_refl P1.eq_sym P1.eq_trans P1.eq_dec
      P1.order P1.order_refl P1.order_antisym P1.order_trans P1.order_dec
      P1.meet P1.meet_bound1 P1.meet_bound2 P1.meet_greatest_lower_bound
      (lattice_def.widen w1) (lattice_def.widen_bound1 w1) (lattice_def.widen_bound2 w1) (lattice_def.widen_eq1 w1) (lattice_def.widen_eq2 w1)
      (lattice_def.narrow n1) (lattice_def.narrow_bound1 n1) (lattice_def.narrow_bound2 n1) (lattice_def.narrow_eq1 n1) (lattice_def.narrow_eq2 n1).
  Definition PosWid2 := ProdWiden.PosWid_constr
    P2.t P2.eq P2.eq_refl P2.eq_sym P2.eq_trans P2.eq_dec
      P2.order P2.order_refl P2.order_antisym P2.order_trans P2.order_dec
      P2.meet P2.meet_bound1 P2.meet_bound2 P2.meet_greatest_lower_bound
      (lattice_def.widen w2) (lattice_def.widen_bound1 w2) (lattice_def.widen_bound2 w2) (lattice_def.widen_eq1 w2) (lattice_def.widen_eq2 w2)
      (lattice_def.narrow n2) (lattice_def.narrow_bound1 n2) (lattice_def.narrow_bound2 n2) (lattice_def.narrow_eq1 n2) (lattice_def.narrow_eq2 n2).

  Definition Pos := (ProdWiden.ProdPosWid _ _ PosWid1 PosWid2).

  Definition widen := fun (x y : (P1.t * P2.t)) => 
     (widen w1 (fst x) (fst y), widen w2 (snd x) (snd y)).

  Definition widen_bound1 : forall x y : t, order x (widen x y) :=
     ProdWiden.widen_bound1 Pos.
  Definition widen_bound2 : forall x y : t, order y (widen x y) :=
     ProdWiden.widen_bound2 Pos.
  Definition widen_eq1 : forall x y z: t, 
    eq x y -> eq (widen x z) (widen y z) :=
     ProdWiden.widen_eq1 Pos.
  Definition widen_eq2 : forall x y z: t, 
    eq x y -> eq (widen z x) (widen z y) :=
     ProdWiden.widen_eq2 Pos.
  Lemma widen_bottom1 : forall x : t, eq x (widen x bottom).
  Proof.
    intros; unfold widen, bottom; destruct x; constructor; simpl.
    apply widen_bottom1.
    apply widen_bottom1.
  Qed.
  Lemma widen_bottom2 : forall x : t, eq x (widen bottom x).
  Proof.
    intros; unfold widen, bottom; destruct x; constructor; simpl.
    apply widen_bottom2.
    apply widen_bottom2.
  Qed.

  Definition narrow := fun (x y : (P1.t * P2.t)) => 
     (narrow n1 (fst x) (fst y), narrow n2 (snd x) (snd y)).

  Definition narrow_bound1 : forall x y : t, order (narrow x y) x :=
     ProdWiden.narrow_bound1 Pos.
  Definition narrow_bound2 : forall x y : t, order (meet x y) (narrow x y) :=
     ProdWiden.narrow_bound2 Pos.
  Definition narrow_eq1 : forall x y z: t, 
    eq x y -> eq (narrow x z) (narrow y z) :=
     ProdWiden.narrow_eq1 Pos.
  Definition narrow_eq2 : forall x y z: t, 
    eq x y -> eq (narrow z x) (narrow z y) :=
     ProdWiden.narrow_eq2 Pos.

  Definition widen_rel : (t*t) -> (t*t) -> Prop := fun x y =>
    eq (snd x) (widen (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Lemma widen_acc_property : forall x:t, Acc widen_rel (x,x).
  Proof.
    apply (ProdWiden.widen_acc_property_prod _ _ PosWid1 PosWid2).
    intros x; apply (lattice_def.widen_acc_property w1 x).
    intros x; apply (lattice_def.widen_acc_property w2 x).
  Qed.

  Definition narrow_rel : (t*t) -> (t*t) -> Prop := fun x y =>
    eq (snd x) (narrow (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Lemma narrow_acc_property : forall x:t, Acc narrow_rel (x,x).
  Proof.
    apply (ProdWiden.narrow_acc_property_prod _ _ PosWid1 PosWid2).
    intros x; apply (lattice_def.narrow_acc_property n1 x).
    intros x; apply (lattice_def.narrow_acc_property n2 x).
  Qed.

  Definition widening : widening_operator eq order bottom :=
   WidOp (eq:=eq) order bottom
     widen_bound1 widen_bound2 widen_eq1 widen_eq2
     widen_bottom1 widen_bottom2 widen_acc_property. 
 
  Definition narrowing : narrowing_operator eq order meet :=
   NarOp (eq:=eq) order meet
     narrow_bound1 narrow_bound2 narrow_eq1 narrow_eq2 narrow_acc_property. 

  
End ProdLattice.

