Require Export lattice_def.
Require Export wf_prop.

Set Implicit Arguments.

Section A_eq_order.

Variable A : Set.
Variable eq order : A -> A -> Prop.

Inductive eqbot  : option A  -> option A -> Prop :=
  eqbot1 : eqbot None None
| eqbot2 : forall x1 x2, eq x1 x2 -> eqbot (Some x1) (Some x2).

Inductive orderbot  : option A -> option A -> Prop :=
  orderbot1 : forall x, orderbot None x
| orderbot2 : forall x1 x2, order x1 x2 ->
   orderbot (Some x1) (Some x2).

Inductive ordertop  : option A -> option A -> Prop :=
  ordertop1 : forall x, ordertop x None
| ordertop2 : forall x1 x2, order x1 x2 ->
   ordertop (Some x1) (Some x2).

End A_eq_order.

Section partial_meet.

Variable A : Set.
Variable eq order : A -> A -> Prop.
Variable meet : A -> A -> option A.
Variable meet_bound1 : forall x y, orderbot order (meet x y) (Some x).
Variable meet_bound2 : forall x y, orderbot order (meet x y) (Some y).
Variable meet_greatest_lower_bound :
  forall x y z, order z x -> order z y -> orderbot order (Some z) (meet x y).


Record total_meet : Set := {
  total_meet_op : option A -> option A -> option A;
  total_meet_bound1 : forall x y,
   orderbot order (total_meet_op x y) x;
  total_meet_bound2 : forall x y,
   orderbot order (total_meet_op x y) y;
  total_meet_greatest_lower_bound : forall x y z,
   orderbot order z x -> orderbot order z y ->
   orderbot order z (total_meet_op x y)
}.

Lemma build_total_meet : total_meet.
Proof.
  exists 
   (fun x y =>
       match x,y with
         None,_ => None
       | _,None => None
       | Some x,Some y => meet x y
       end);
  destruct x; try constructor;
  destruct y; try constructor; auto.
  intros z H; inversion_clear H; 
  intros H1; inversion_clear H1; try constructor; auto.
Defined.

End partial_meet.

Module LatRecord.

Record lattice_record (t:Set) : Type := {
 eq  : t -> t -> Prop;
 eq_refl : forall x : t, eq x x;
 eq_sym : forall x y : t, eq x y -> eq y x;
 eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z;
 eq_dec : forall x y : t, {eq x y}+{~ eq x y};

 order : t -> t -> Prop;
 order_refl : forall x y : t, eq x y -> order x y;
 order_antisym : forall x y : t, order x y -> order y x -> eq x y;
 order_trans : forall x y z : t, order x y -> order y z -> order x z;
 order_dec : forall x y : t, {order x y}+{~ order x y};

 join  : t -> t -> t;
 join_bound1 : forall x y : t, order x (join x y);
 join_bound2 : forall x y : t, order y (join x y);
 join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z;

 meet  : t -> t -> t;
 meet_bound1 : forall x y : t, order (meet x y) x;
 meet_bound2 : forall x y : t, order (meet x y) y;
 meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y);

 bottom : t;
 bottom_is_bottom : forall x : t, order bottom x;

 top : t;
 top_is_top : forall x : t, order x top
}.

End LatRecord.

Section add_top.

Variable t : Set.
Variable eq  : t -> t -> Prop.
Variable eq_refl : forall x : t, eq x x.
Variable eq_sym : forall x y : t, eq x y -> eq y x.
Variable eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
Variable eq_dec : forall x y : t, {eq x y}+{~ eq x y}.

Variable order : t -> t -> Prop.
Variable order_refl : forall x y : t, eq x y -> order x y.
Variable order_antisym : forall x y : t, order x y -> order y x -> eq x y.
Variable order_trans : forall x y z : t, order x y -> order y z -> order x z.
Variable order_dec : forall x y : t, {order x y}+{~ order x y}.

Variable join  : t -> t -> t.
Variable join_bound1 : forall x y : t, order x (join x y).
Variable join_bound2 : forall x y : t, order y (join x y).
Variable join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z.

Variable meet  : t -> t -> t.
Variable meet_bound1 : forall x y : t, order (meet x y) x.
Variable meet_bound2 : forall x y : t, order (meet x y) y.
Variable meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y).

Variable bottom : t.
Variable bottom_is_bottom : forall x : t, order bottom x.

Definition add_top_lattice : LatRecord.lattice_record (option t).
refine (LatRecord.Build_lattice_record
 (eqbot eq) _ _ _ _
 (ordertop order) _ _ _ _
 (fun x y =>
   match x,y with
    | Some x,Some y => Some (join x y)
    | _,_ => None
   end) _ _ _
 (fun x y =>
   match x,y with
    | None,_ => y
    | _,None => x
    | Some x,Some y => Some (meet x y)
   end) _ _ _
  (Some bottom) _
  None _
).
destruct x; constructor; auto.
intros x y H; inversion_clear H; constructor; auto.
intros x y z H; inversion_clear H; 
intros H1; inversion_clear H1; constructor; eauto.
destruct x; destruct y; try (right;intros H;inversion_clear H;fail).
destruct (eq_dec t0 t1).
left; constructor; auto.
right; intros H; inversion_clear H; elim n; auto.
left; constructor.
intros x y H; inversion_clear H; constructor; auto.
intros x y H; inversion_clear H; 
intros H1; inversion_clear H1; constructor; eauto.
intros x y z H; inversion_clear H; 
intros H1; inversion_clear H1; constructor; eauto.
destruct x; destruct y; try (right;intros H;inversion_clear H;fail).
destruct (order_dec t0 t1).
left; constructor; auto.
right; intros H; inversion_clear H; elim n; auto.
left; constructor.
left; constructor.
destruct x; try constructor.
destruct y; constructor; auto.
destruct x; try constructor.
destruct y; constructor; auto.
intros x y z H; inversion_clear H; 
intros H1; inversion_clear H1; constructor; eauto.
destruct x; try constructor.
destruct y; constructor; auto.
destruct x; try constructor.
destruct y; constructor; auto.
destruct y; constructor; auto.
intros x y z H; inversion_clear H; 
intros H1; inversion_clear H1; constructor; eauto.
destruct x; constructor; auto.
constructor.
Defined.

Variable w : widening_operator eq order bottom.
Variable n : narrowing_operator eq order meet.
Variable acc : well_founded (fun x y => ~ eq y x /\ order y x).

Lemma add_top_w : 
  widening_operator (LatRecord.eq add_top_lattice)  (LatRecord.order add_top_lattice)  (LatRecord.bottom add_top_lattice).
Proof.
  exists (fun x y =>
   match x,y with
    | Some x,Some y => Some (widen w x y)
    | _,_ => None
   end).
  destruct x; try constructor.
  destruct y; constructor; auto.
  apply widen_bound1.
  destruct x; try constructor.
  destruct y; constructor; auto.
  apply widen_bound2.
  intros x y z H; inversion_clear H; try constructor.
  destruct z; constructor.
  apply widen_eq1; auto.
  intros x y z H; inversion_clear H; 
  destruct z; constructor.
  apply widen_eq2; auto.
  destruct x; simpl; try constructor.
  apply widen_bottom1.
  destruct x; simpl; try constructor.
  apply widen_bottom2.
  assert (   Acc
     (widen_rel (eqbot eq)
        (fun x y : option t =>
         match x with
         | Some x0 =>
             match y with
             | Some y0 => Some (widen w x0 y0)
             | None => None (A:=t)
             end
         | None => None (A:=t)
         end)) (None, None)).
  constructor; destruct y; intros (H1,H2); simpl in *.
  elim H2; inversion_clear H1; constructor.
  destruct x.
  replace (Some t0,Some t0) with (Some (fst (t0,t0)),Some (snd (t0,t0))); auto.
  induction (widen_acc_property w t0).
  constructor.
  destruct x; destruct y; intros (H3,H2); simpl in *; auto. 
  destruct o; inversion_clear H3 in H2.
  apply (H1 (t3,x1)).
  split; simpl; auto.
  intro; elim H2; constructor; auto.
  exact H.
  exact H.
Qed.

Inductive eq' : (option t)*(option t) -> (option t)*(option t) -> Prop :=
  eq'_def : forall x y z,
    eq y z -> eq' (Some x,Some y) (Some x,Some z).

Lemma add_top_n : 
  narrowing_operator (LatRecord.eq add_top_lattice)  (LatRecord.order add_top_lattice)  (LatRecord.meet add_top_lattice).
Proof.
  exists (fun x y =>
   match x,y with
    | None,_ => y
    | _,None => x
    | Some x,Some y => Some (narrow n x y)
   end).
  destruct x; try constructor.
  destruct y; constructor; auto.
  apply narrow_bound1.
  destruct x; 
  destruct y; simpl; try constructor; auto.
  apply narrow_bound2.
  intros x y z H; inversion_clear H; try constructor.
  destruct z; constructor; auto.
  destruct z; constructor; auto.
  apply narrow_eq1; auto.
  intros x y z H; inversion_clear H; 
  destruct z; constructor; auto.
  apply narrow_eq2; auto.
  assert (forall x,
   Acc
     (narrow_rel (LatRecord.eq add_top_lattice)
        (fun x0 y : option t =>
         match x0 with
         | Some x1 =>
             match y with
             | Some y0 => Some (narrow n x1 y0)
             | None => x0
             end
         | None => y
         end)) (Some (fst (x,x)), Some (snd (x,x)))).
  intros x; induction (narrow_acc_property n x).
  constructor.
  destruct x0; destruct y; intros (H3,H2); simpl in *; auto. 
  destruct o; inversion_clear H3 in H2.
  apply (H0 (t2,x1)).
  split; simpl; auto.
  intro; elim H2; constructor; auto.
  elim H2; constructor; auto.
  destruct x.
  apply (H t0).
  constructor.
  destruct y; intros (H3,H2); simpl in *; auto. 
  inversion_clear H3 in H2.
  elim H2; constructor.
  apply acc_eq with (eq := eq') (x:=(Some x2,Some x2)); auto.
  intros.
  inversion_clear H3 in H1; destruct H1; simpl in *.
  destruct x; split; simpl in *; auto.
  destruct o1; inversion_clear H1 in H3; try constructor.
  apply eq_trans with (narrow n z0 t0); auto.
  apply narrow_eq1;auto.
  apply eq_trans with z0; auto.
  intro; elim H3.
  inversion_clear H5; constructor.
  eauto.
  constructor; auto.
Qed.

Lemma add_top_acc : well_founded (fun x y => ~ LatRecord.eq add_top_lattice y x /\ LatRecord.order add_top_lattice y x).
Proof.
  intros; unfold well_founded.
  assert (Acc
     (fun x y : option t =>
      ~ LatRecord.eq add_top_lattice y x /\
      LatRecord.order add_top_lattice y x) None).
  constructor; destruct y; intros (H1,H2). 
  inversion_clear H2.
  elim H1; auto; constructor. 
  destruct a; auto.
  induction (acc t0).
  constructor.
  intros y (H3,H2).
  inversion_clear H2 in H3; auto.
  apply H1; split; auto. 
  intro; elim H3; constructor; auto.
Qed.

End add_top.

