Require Export solve.
Require Export wf_prop.

Module Type NewImplemLattice.

  Declare Module L : Lattice.

  Parameter t : Set.

  Parameter eq : t -> t -> Prop.
  Parameter eq_dec : forall x y : t, {eq x y}+{~ eq x y}.

  Parameter order : t -> t -> Prop.
  Parameter order_eq1 : forall x y : t, eq x y -> order x y /\ order y x.
  Parameter order_eq2 : forall x y : t, order x y -> order y x -> eq x y.
  Parameter order_dec : forall x y : t, {order x y}+{~ order x y}.

  Parameter join  : t -> t -> t.
  Parameter meet  : t -> t -> t.
  Parameter widen : t -> t -> t.
  Parameter narrow : t -> t -> t.

  Parameter i : t -> L.t.
  Parameter i_order_preserve : forall x y, order x y -> L.order (i x) (i y).
  Parameter inv_i_order_preserve : forall x y, L.order (i x) (i y) -> order x y.

  Parameter join_commut : forall x y,
    L.eq (i (join x y)) (L.join (i x) (i y)).
  Parameter meet_commut : forall x y,
    L.eq (i (meet x y)) (L.meet (i x) (i y)).
  Parameter widen_commut : forall x y,
    L.eq (i (widen x y)) (L.widening (i x) (i y)).
  Parameter narrow_commut : forall x y,
    L.eq (i (narrow x y)) (L.narrowing (i x) (i y)).

  Parameter bottom : t.
  Parameter bottom_is_bottom : L.eq L.bottom (i bottom).

  Hint Resolve order_eq1 order_eq2 i_order_preserve inv_i_order_preserve
     join_commut meet_commut widen_commut narrow_commut bottom_is_bottom.

End NewImplemLattice.

Module Lattice_of_NewImplemLattice (N:NewImplemLattice) <: Lattice
  with Definition t := N.t
  with Definition eq := N.eq
  with Definition order := N.order
  with Definition join := N.join
  with Definition meet := N.meet
  with Definition bottom := N.bottom.

  Definition t := N.t.
  Definition eq := N.eq.
  Definition order := N.order.
  Definition join := N.join.
  Definition meet := N.meet.
  Definition bottom := N.bottom.
  
  Lemma eq_refl : forall x : t, eq x x.
  Proof.
    unfold eq; auto.
  Qed.

  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof.
    unfold eq; intros.
    destruct (N.order_eq1 _ _ H); auto.
  Qed.

  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof.
    unfold eq; intros.
    destruct (N.order_eq1 _ _ H).
    destruct (N.order_eq1 _ _ H0).
    apply N.order_eq2; eauto.
  Qed.

  Definition eq_dec : forall x y : t, {eq x y}+{~ eq x y} := N.eq_dec.

  Lemma order_refl : forall x y : t, eq x y -> order x y.
  Proof.
    unfold eq, order; intros.
    destruct (N.order_eq1 _ _ H); auto.    
  Qed.

  Lemma order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Proof.
    unfold eq, order; intros.
    apply N.order_eq2; auto.
  Qed.

  Lemma order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Proof.
    unfold order; eauto.
  Qed.

  Definition order_dec : forall x y : t, {order x y}+{~ order x y} := N.order_dec.

  Lemma i_injective : forall x y, N.L.eq (N.i x) (N.i y) -> eq x y.
  Proof.
    intros; apply order_antisym; apply N.inv_i_order_preserve; auto.
  Qed.

  Lemma i_preserve_eq : forall x y, eq x y -> N.L.eq (N.i x) (N.i y).
  Proof.
    unfold eq; intros.
    destruct (N.order_eq1 _ _ H); auto.    
  Qed.

  Lemma join_bound1 : forall x y : t, order x (join x y).
  Proof.
    intros.
    apply N.inv_i_order_preserve.
    generalize (N.L.join_bound1 (N.i x) (N.i y)); intros. 
    generalize (N.join_commut x y); eauto.
  Qed.
 
  Lemma join_bound2 : forall x y : t, order y (join x y).
  Proof.
    intros.
    apply N.inv_i_order_preserve.
    generalize (N.L.join_bound2 (N.i x) (N.i y)); intros. 
    generalize (N.join_commut x y); eauto.
  Qed.
 
  Lemma join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z.
  Proof.
    intros.
    generalize (N.L.join_least_upper_bound (N.i x) (N.i y) (N.i z)); intros.
    generalize (N.join_commut x y); intros.
    apply N.inv_i_order_preserve.
    apply N.L.order_trans with  (N.L.join (N.i x) (N.i y)); auto.
  Qed.

  Lemma meet_bound1 : forall x y : t, order (meet x y) x.
  Proof.
    intros.
    apply N.inv_i_order_preserve.
    generalize (N.L.meet_bound1 (N.i x) (N.i y)); intros. 
    generalize (N.meet_commut x y); eauto.
  Qed.

  Lemma meet_bound2 : forall x y : t, order (meet x y) y. 
  Proof.
    intros.
    apply N.inv_i_order_preserve.
    generalize (N.L.meet_bound2 (N.i x) (N.i y)); intros. 
    generalize (N.meet_commut x y); eauto.
  Qed.

  Lemma meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y).
  Proof.
    intros.
    generalize (N.L.meet_greatest_lower_bound (N.i x) (N.i y) (N.i z)); intros.
    generalize (N.meet_commut x y); intros.
    apply N.inv_i_order_preserve.
    apply N.L.order_trans with (N.L.meet (N.i x) (N.i y)); auto.
  Qed.

  Lemma bottom_is_bottom : forall x : t, order bottom x.
  Proof.
    intros; apply N.inv_i_order_preserve.
    apply N.L.order_trans with N.L.bottom; auto.
  Qed.

  Lemma ascending_chain_condition :
      well_founded (fun x y => ~ N.L.eq y x /\ N.L.order y x) ->
      well_founded (fun x y => ~ eq y x /\ order y x).
  Proof.
    intros.
    apply wf_mes with (mes:=N.i) (2:=H).
    intuition.
    elim H1; apply i_injective; auto.
  Qed.

  Definition widening : widening_operator eq order bottom.
  Proof.
    intros.
    set (w:=N.L.widening).
    exists N.widen.
    intros.
    apply N.inv_i_order_preserve.
    generalize (widen_bound1 w (N.i x) (N.i y)); eauto. 
    intros.
    apply N.inv_i_order_preserve.
    generalize (widen_bound2 w (N.i x) (N.i y)); eauto.
    intros.
    apply i_injective.
    generalize (widen_eq1 w (N.i x) (N.i y) (N.i z) (i_preserve_eq _ _ H)); eauto. 
    intros.
    apply i_injective.
    generalize (widen_eq2 w (N.i x) (N.i y) (N.i z) (i_preserve_eq _ _ H)); eauto.
    intros; apply i_injective.
    apply N.L.eq_trans with (widen w (N.i x) (N.i bottom) ); auto.
    apply N.L.eq_trans with (widen w (N.i x) N.L.bottom); auto.
    apply (widen_bottom1 w).
    apply (widen_eq2 w); auto.
    intros; apply i_injective.
    apply N.L.eq_trans with (widen w (N.i bottom) (N.i x)); auto.
    apply N.L.eq_trans with (widen w N.L.bottom (N.i x)); auto.
    apply (widen_bottom2 w).
    apply (widen_eq1 w); auto.
    intros.
    apply acc_mes with (mes:=fun x => (N.i (fst x),N.i (snd x)))
                       (ltB:=widen_rel N.L.eq (widen w)).
    destruct a1; destruct a2; intros; simpl.
    destruct H as [HH HH2]; simpl in *; repeat split; simpl; auto.
    generalize (N.widen_commut t3 t0); intros.
    apply N.L.eq_trans with (2:=H).
    apply i_preserve_eq; auto.
    intro; elim HH2; apply i_injective; auto.
    simpl; apply (widen_acc_property w).
  Defined.

  Lemma narrowing : narrowing_operator eq order meet.
  Proof.
    intros.
    set (n:=N.L.narrowing).
    exists N.narrow.
    intros.
    apply N.inv_i_order_preserve.
    generalize (narrow_bound1 n (N.i x) (N.i y)); eauto.
    intros.
    apply N.inv_i_order_preserve.
    generalize (narrow_bound2 n (N.i x) (N.i y)); intros.
    generalize (N.meet_commut x y); intros.
    apply N.L.order_trans with (N.L.meet (N.i x) (N.i y)); auto.
    apply N.L.order_trans with (narrow n (N.i x) (N.i y)); auto.
    intros.
    apply i_injective.
    generalize (narrow_eq1 n (N.i x) (N.i y) (N.i z) (i_preserve_eq _ _ H)); eauto. 
    intros.
    apply i_injective.
    generalize (narrow_eq2 n (N.i x) (N.i y) (N.i z) (i_preserve_eq _ _ H)); eauto.
    intros.
    apply acc_mes with (mes:=fun x => (N.i (fst x),N.i (snd x)))
                       (ltB:=narrow_rel N.L.eq (narrow n)).
    destruct a1; destruct a2; intros; simpl.
    destruct H as [HH HH1]; simpl in *; repeat split; simpl; auto.
    generalize (N.narrow_commut t3 t0); intros.
    apply N.L.eq_trans with (2:=H).
    apply i_preserve_eq; auto.
    intro; elim HH1; apply i_injective; auto.
    simpl; apply (narrow_acc_property n).
  Defined.

End Lattice_of_NewImplemLattice.

