Require Export lattice_def.
Require Export add_option.
Require Export solve.


Require Import ZArith.

Inductive Z' : Set :=
 | z : Z -> Z'
 | infty_pos : Z'
 | infty_neg : Z'.

Inductive order_Z' : Z' -> Z' -> Prop :=
 | order_z_z : forall a b : Z, (Zlt a b) -> (order_Z' (z a) (z b))
 | order_infty: (order_Z' infty_neg infty_pos)
 | order_infty_pos : forall x: Z, (order_Z' (z x) infty_pos)
 | order_infty_neg : forall x: Z, (order_Z' infty_neg (z x)).

Notation "a <' b" := (order_Z' a b) (at level 50).
Notation "a <=' b" := (a <' b \/ a=b) (at level 50).

Lemma infty_neg_min : forall a : Z', infty_neg <=' a.
destruct a; intros; auto || left; constructor; auto.
Qed.

Lemma infty_pos_max : forall a : Z', a <=' infty_pos.
destruct a; intros; auto || left; constructor; auto.
Qed.

Lemma order_Z'_antisym : forall a b: Z', a<'b -> ~b<'a.
intros a b H; inversion_clear H; red; intros H; inversion_clear H.
apply (Zlt_asym _ _ H0 H1).
Qed.

Lemma order_Z'_strict : forall a b: Z', a<'b -> ~a=b.
intros a b H; inversion_clear H; red; intros H.
injection H; intros; subst; omega.
discriminate H.
discriminate H.
discriminate H.
Qed.

Lemma order_Z'_trans : forall a b c: Z', a<'b -> b<'c -> a<'c.
intros a b c H; inversion_clear H; intros H'; inversion_clear H'; constructor.
auto with zarith.
Qed.

Lemma order_Z'_trans1 : forall a b c: Z', a<'b -> b<='c -> a<='c.
intros a b c H; inversion_clear H; 
 (intros [H'|H']; [inversion_clear H';left;constructor|rewrite <- H'; left; constructor; auto]).
auto with zarith.
Qed. 

Lemma order_Z'_trans2 : forall a b c: Z', 
 a<'b -> b<='c -> a<'c.
intros a b c H H1; destruct H1.
apply order_Z'_trans with b; auto.
subst; auto.
Qed. 

Lemma order_Z'_trans3 : forall a b c: Z', 
 b<='c -> c<'a -> b<'a.
intros a b c H H1; destruct H.
apply order_Z'_trans with c; auto.
subst; auto.
Qed. 

Lemma order_Z'_disj : forall b c: Z', 
 b<='c -> c<'b -> False.
intros.
assert (b <' b).
apply order_Z'_trans3 with c; auto.
apply (order_Z'_strict _ _ H1); auto.
Qed. 

Lemma not_order_Z' : forall a b: Z', ~ a<'b -> b<='a.
destruct a; intros.
destruct b; intros.
case (Z_dec' z1 z0); intros.
case s; intros.
left; constructor; auto.
elim H; constructor; auto.
rewrite e; auto.
elim H; constructor; auto.
left; constructor; auto.
destruct b; intros; auto.
left; constructor; auto.
left; constructor; auto.
destruct b; intros; auto.
elim H; constructor; auto.
elim H; constructor; auto.
Qed.

Lemma order_Z'_dec : forall a b : Z', {a<'b}+{~a<'b}.
destruct a; intros.
destruct b; intros.
case (Z_lt_dec z0 z1); intros.
left; constructor; auto.
right; intros H; inversion H; elim n; auto.
left; constructor; auto.
right; intros H; inversion H.
right; intros H; inversion H.
destruct b; intros.
left; constructor; auto.
left; constructor; auto.
right; intros H; inversion H.
Qed.

Inductive interval : Set := 
 | ITV : forall a b : Z', a <=' b -> interval.

Definition min : interval -> Z' :=
 fun i => match i with
  | ITV a _ _ => a
 end.

Definition max : interval -> Z' :=
 fun i => match i with
  | ITV _ a _ => a
 end.

Lemma min_order_max : forall i:interval, (min i)<='(max i).
destruct i; simpl; auto.
Qed.

Definition eq_interval : interval -> interval -> Prop :=
 fun i1 i2 => (min i1)=(min i2) /\ (max i1)=(max i2).

Lemma eq_interval_refl : forall i : interval, eq_interval i i.
unfold eq_interval; auto.
Qed.

Lemma eq_interval_sym : forall i1 i2 : interval, 
 eq_interval i1 i2 -> eq_interval i2 i1.
unfold eq_interval; intuition.
Qed.

Lemma eq_interval_trans : forall i1 i2 i3 : interval, 
 eq_interval i1 i2 -> eq_interval i2 i3 -> eq_interval i1 i3.
unfold eq_interval; intuition.
rewrite <- H; auto.
rewrite <- H3; auto.
Qed.

Definition order_interval : interval -> interval -> Prop :=
 fun i1 i2 => 
   ((min i2)<='(min i1)) /\ ((max i1)<='(max i2)).

Lemma order_interval_refl : forall i1 i2 : interval, 
  eq_interval i1 i2 -> order_interval i1 i2.
unfold eq_interval,order_interval; intuition.
Qed.

Lemma order_interval_antisym : forall i1 i2 : interval, 
 order_interval i1 i2 -> order_interval i2 i1 -> eq_interval i1 i2.
generalize order_Z'_antisym; intros Hantisym.
generalize order_Z'_strict; intros Hstrict.
unfold eq_interval,order_interval; intuition; solve [apply False_ind; eauto].
Qed.

Ltac Rew := match goal with
 | id : _ = _ |- _ => try rewrite id; clear id
end.

Lemma order_interval_trans : forall i1 i2 i3 : interval, 
 order_interval i1 i2 -> order_interval i2 i3 -> order_interval i1 i3.
generalize order_Z'_trans1; intros H1.
unfold order_interval; intuition eauto; try (repeat Rew; auto; fail).
right; rewrite H2; auto.
right; rewrite H2; auto.
right; rewrite H2; auto.
right; rewrite H2; auto.
Qed.

Definition join_interval : interval -> interval -> interval.
refine (fun i1 i2 => 
 let min1 := (min i1) in
 let min2 := (min i2) in
 let max1 := (max i1) in
 let max2 := (max i2) in 
 let min' := (match (order_Z'_dec min1 min2) with
      (left _) => min1
    | (right _) => min2
   end) in 
 let max' := (match (order_Z'_dec max1 max2) with
      left _ => max2
    | right _ => max1
   end) in
 (ITV min' max' _)).
unfold min',max'.
cut (min1<'max1 \/ min1=max1); intros.
cut (min2<'max2 \/ min2=max2); intros.
case (order_Z'_dec min1 min2); case (order_Z'_dec max1 max2); intros; auto.
apply order_Z'_trans1 with min2; auto.
elim (not_order_Z' _ _ n0); intros.
apply order_Z'_trans1 with min1; auto.
rewrite H1; auto.
unfold min2,max2 in * |- *; apply min_order_max.
unfold min1,max1 in * |- *; apply min_order_max.
Defined.

Lemma join_interval_bound1 : forall i1 i2 : interval, 
  order_interval i1 (join_interval i1 i2).
intros i1 i2.
unfold order_interval,join_interval.
case (order_Z'_dec (min i1) (min i2)); intros;
case (order_Z'_dec (max i1) (max i2)); simpl; intros; split; auto.
apply not_order_Z'; auto.
apply not_order_Z'; auto.
Qed.

Lemma join_interval_bound2 : forall i1 i2 : interval, 
  order_interval i2 (join_interval i1 i2).
intros i1 i2.
unfold order_interval,join_interval.
case (order_Z'_dec (min i1) (min i2)); intros;
case (order_Z'_dec (max i1) (max i2)); simpl; intros; split; auto.
apply not_order_Z'; auto.
apply not_order_Z'; auto.
Qed.

Lemma join_interval_least_bound : forall i1 i2 i: interval, 
  order_interval i1 i ->
  order_interval i2 i ->
 order_interval (join_interval i1 i2) i.
intros i1 i2 i.
unfold order_interval,join_interval.
case (order_Z'_dec (min i1) (min i2)); intros H (H1,H2) (H3,H4);
case (order_Z'_dec (max i1) (max i2)); simpl; intros; split; auto.
Qed.

Definition meet_interval : interval -> interval -> option interval.
refine (fun i1 i2 => 
 let min1 := (min i1) in
 let min2 := (min i2) in
 let max1 := (max i1) in
 let max2 := (max i2) in 
 let min' := (match (order_Z'_dec min1 min2) with
      (left _) => min2
    | (right _) => min1
   end) in 
 let max' := (match (order_Z'_dec max1 max2) with
      left _ => max1
    | right _ => max2
   end) in
 match (order_Z'_dec max' min') with
   left _ => None
 | right _ => Some (ITV min' max' _)
 end).
generalize dependent n.
unfold min',max'.
cut (min1<'max1 \/ min1=max1); [intro|idtac].
cut (min2<'max2 \/ min2=max2); [intro|idtac].
case (order_Z'_dec min1 min2); case (order_Z'_dec max1 max2); intros; auto.
apply not_order_Z'; auto.
apply not_order_Z'; auto.
unfold min2,max2 in * |- *; apply min_order_max.
unfold min1,max1 in * |- *; apply min_order_max.
Defined.

Notation eqbot := (eqbot eq_interval).
Notation orderbot := (orderbot order_interval).

Lemma meet_interval_bound1 : forall i1 i2 : interval, 
  orderbot (meet_interval i1 i2) (Some i1).
intros i1 i2.
unfold order_interval,meet_interval.
case (order_Z'_dec (min i1) (min i2)); intros;
case (order_Z'_dec (max i1) (max i2)); simpl; intros;
case order_Z'_dec; auto; intros; try constructor; split; simpl; auto.
apply not_order_Z'; auto.
apply not_order_Z'; auto.
Qed.

Lemma meet_interval_bound2 : forall i1 i2 : interval, 
  orderbot (meet_interval i1 i2) (Some i2).
intros i1 i2.
unfold order_interval,meet_interval.
case (order_Z'_dec (min i1) (min i2)); intros;
case (order_Z'_dec (max i1) (max i2)); simpl; intros; 
case order_Z'_dec; auto; intros; try constructor; split; simpl; auto.
apply not_order_Z'; auto.
apply not_order_Z'; auto.
Qed.

Lemma meet_interval_least_bound : forall i1 i2 i: interval, 
  order_interval i i1 ->
  order_interval i i2 ->
  orderbot (Some i) (meet_interval i1 i2).
intros i1 i2 i.
unfold order_interval,meet_interval.
case (order_Z'_dec (min i1) (min i2)); intros H (H1,H2) (H3,H4);
case (order_Z'_dec (max i1) (max i2)); simpl; intros;
case order_Z'_dec; auto; intros; try (constructor; split; simpl; auto);
apply False_ind;
apply (order_Z'_disj (min i) (max i));
try (apply min_order_max).
apply order_Z'_trans2 with (min i2); auto.
apply order_Z'_trans3 with (max i1); auto.
apply order_Z'_trans2 with (min i2); auto.
apply order_Z'_trans3 with (max i2); auto.
apply order_Z'_trans2 with (min i1); auto.
apply order_Z'_trans3 with (max i1); auto.
apply order_Z'_trans2 with (min i1); auto.
apply order_Z'_trans3 with (max i2); auto.
Qed.

Definition widen_interval : interval -> interval -> interval.
refine (fun i1 i2 => 
 let min1 := (min i1) in
 let min2 := (min i2) in
 let max1 := (max i1) in
 let max2 := (max i2) in 
 let min' := (match (order_Z'_dec min2 min1) with
      (left _) => infty_neg
    | (right _) => min1
   end) in 
 let max' := (match (order_Z'_dec max1 max2) with
      left _ => infty_pos
    | right _ => max1
   end) in
 (ITV min' max' _)).
unfold min',max'.
case (order_Z'_dec min2 min1); case (order_Z'_dec max1 max2); intros; auto.
left; constructor.
case max1; intros; (left;constructor; auto) || auto.
case min1; intros; (left;constructor; auto) || auto.
unfold min1,max1 in * |- *; apply min_order_max.
Defined.

Lemma order_Z'_eq_trans : forall x y t, x <=' y -> y <=' t -> x <=' t.
Proof.
  destruct 1.
  intros; apply order_Z'_trans1 with y; auto.
  subst; auto.
Qed.

Lemma not_order_Z'' : forall a b: Z', ~ a<'b -> b<='a.
Proof.
  intros; apply not_order_Z'; auto.
Qed.

Lemma order_interval_dec : forall i1 i2, {order_interval i1 i2}+{~order_interval i1 i2}.
destruct i1; destruct i2.
case (order_Z'_dec a a0); intros.
right; intros H; inversion_clear H; simpl in *.
destruct H0.
apply (order_Z'_antisym _ _ H o1).
elim (order_Z'_strict _ _ o1); auto.
case (order_Z'_dec b0 b); intros.
right; intros H; inversion_clear H; simpl in *.
destruct H1.
apply (order_Z'_antisym _ _ H o1).
elim (order_Z'_strict _ _ o1); auto.
left; constructor; simpl; intros.
apply not_order_Z'; auto.
apply not_order_Z'; auto.
Qed.

Definition narrow_interval : interval -> interval -> option interval :=
  fun i1 i2 => 
  let min := (match (min i1) with
                infty_neg => min i2
              | _ => min i1
              end) in
  let max :=  (match (max i1) with
                 infty_pos => max i2
               | _ => max i1
               end) in
   match order_Z'_dec max min with
    | left _ => None
    | right h => Some (ITV min max (not_order_Z' _ _ h))
   end.

Definition top_interval : interval.
refine 
 (ITV infty_neg infty_pos _).
left; constructor.
Defined.

Lemma top_interval_is_top : forall i : interval, order_interval i top_interval.
unfold top_interval,order_interval; intros; split; auto.
apply infty_neg_min.
apply infty_pos_max.
Qed.

Lemma widen_interval_eq1 : forall i1 i2 i3,
 eq_interval i1 i2 ->
 eq_interval (widen_interval i1 i3) (widen_interval i2 i3).
intros i1 i2 i3.
unfold eq_interval,widen_interval.
split.
case (order_Z'_dec (min i1) (min i2)); intro;
case (order_Z'_dec (min i3) (min i1)); intro;
case (order_Z'_dec (min i3) (min i2)); simpl; intuition.
elim (order_Z'_strict _ _ o H0).
elim (order_Z'_strict _ _ o H0).
elim n0; rewrite <- H0; auto.
elim n0; rewrite H0; auto.
case (order_Z'_dec (max i1) (max i2)); intro;
case (order_Z'_dec (max i1) (max i3)); intro;
case (order_Z'_dec (max i2) (max i3)); simpl; intuition.
elim (order_Z'_strict _ _ o H1).
elim (order_Z'_strict _ _ o H1).
elim n0; rewrite <- H1; auto.
elim n0; rewrite H1; auto.
Qed.

Lemma widen_interval_eq2 : forall i1 i2 i3,
 eq_interval i1 i2 ->
 eq_interval (widen_interval i3 i1) (widen_interval i3 i2).
intros i1 i2 i3.
unfold eq_interval,widen_interval.
split.
case (order_Z'_dec (min i1) (min i2)); intro;
case (order_Z'_dec (min i1) (min i3)); intro;
case (order_Z'_dec (min i2) (min i3)); simpl; intuition.
elim (order_Z'_strict _ _ o H0).
elim (order_Z'_strict _ _ o H0).
elim n0; rewrite <- H0; auto.
elim n0; rewrite H0; auto.
case (order_Z'_dec (max i1) (max i2)); intro;
case (order_Z'_dec (max i3) (max i1)); intro;
case (order_Z'_dec (max i3) (max i2)); simpl; intuition.
elim (order_Z'_strict _ _ o H1).
elim (order_Z'_strict _ _ o H1).
elim n0; rewrite <- H1; auto.
elim n0; rewrite H1; auto.
Qed.

Lemma widen_interval_bound1 : forall i1 i2 : interval,
  order_interval i1 (widen_interval i1 i2).
intros i1 i2.
unfold order_interval,widen_interval.
case (order_Z'_dec (min i2) (min i1)); intros;
case (order_Z'_dec (max i1) (max i2)); simpl; intros; split; auto;
apply infty_neg_min || apply infty_pos_max.
Qed.

Lemma widen_interval_bound2 : forall i1 i2 : interval,
  order_interval i2 (widen_interval i1 i2).
intros i1 i2.
unfold order_interval,widen_interval.
case (order_Z'_dec (min i2) (min i1)); intros;
case (order_Z'_dec (max i1) (max i2)); simpl; intros; split; auto;
apply infty_neg_min || apply infty_pos_max || apply not_order_Z'; auto.
Qed.

Hint Resolve infty_pos_max infty_neg_min.

Lemma narrow_interval_bound1 : forall i1 i2 : interval,
  orderbot (narrow_interval i1 i2) (Some i1).
intros i1 i2.
unfold order_interval,narrow_interval; simpl.
CaseEq (max i1); intros;
CaseEq (min i1); intros;
destruct order_Z'_dec;
constructor; simpl; split; auto;
try (rewrite H0; auto; fail);
try (rewrite H; auto; fail).
Qed.

Lemma narrow_interval_eq_def : forall i1 i2 i,
  narrow_interval i1 i2 = Some i ->
  eq_interval i i1 \/ 
  (eq_interval i i2 /\ min i1 = infty_neg /\ max i1 = infty_pos) \/
  (min i = min i1 /\ max i = max i2 /\ max i1 = infty_pos) \/
  (min i = min i2 /\ max i = max i1 /\ min i1 = infty_neg).
Proof.
  unfold narrow_interval; intros i1 i2 i.
  unfold eq_interval.
  destruct (max i1); destruct (min i1); 
  destruct order_Z'_dec;
  intros HH; inversion_clear HH; 
  unfold eq_interval; simpl; intuition.
Qed.

Lemma narrow_interval_eq_def' : forall i1 i2,
  narrow_interval i1 i2 = None ->
  (max i2 <' min i1 /\ max i1 = infty_pos) \/
  (max i1 <' min i2 /\ min i1 = infty_neg).
Proof.
  unfold narrow_interval; intros i1 i2.
  generalize (min_order_max i1); intro.
  destruct (max i1); destruct (min i1); 
  destruct order_Z'_dec;
  intros HH; inversion_clear HH;
  try (elim (order_Z'_disj _ _ H);assumption); auto.
  elim (order_Z'_disj _ _ (min_order_max i2)); auto.
Qed.

Lemma eqbot_orderbot_trans1 : forall x y z,
  eqbot x y -> orderbot y z -> orderbot x z.
Proof.
  intros.
  inversion_clear H in H0; inversion_clear H0; constructor.
  apply order_interval_trans with x2; auto.
  apply order_interval_refl; auto.
Qed.

Lemma eqbot_orderbot_trans2 : forall x y z,
  orderbot x y -> eqbot y z -> orderbot x z.
Proof.
  intros.
  inversion_clear H in H0; inversion_clear H0; constructor.
  apply order_interval_trans with x2; auto.
  apply order_interval_refl; auto.
Qed.

Hint Resolve eqbot1 eqbot2 eq_interval_sym
  meet_interval_bound2 meet_interval_bound1
  min_order_max.
Lemma narrow_interval_bound2 : forall i1 i2 : interval,
  orderbot (meet_interval i1 i2) (narrow_interval i1 i2).
intros i1 i2.
CaseEq (narrow_interval i1 i2); intros.
destruct (narrow_interval_eq_def _ _ _ H) as [H1 | [(H1,H2) | [(H1,(H2,H3)) | (H1,(H2,H3))]]].
apply eqbot_orderbot_trans2 with (Some i1); auto.
rewrite H; auto.
apply eqbot_orderbot_trans2 with (Some i2); auto.
rewrite H; auto.
unfold meet_interval.
destruct (order_Z'_dec (min i1) (min i2));
destruct (order_Z'_dec (max i1) (max i2));
case order_Z'_dec; intros; simpl;
try rewrite H; 
try (constructor; split; simpl; auto);
try rewrite H1; try rewrite H2; auto.
unfold meet_interval.
destruct (order_Z'_dec (min i1) (min i2));
destruct (order_Z'_dec (max i1) (max i2));
case order_Z'_dec; intros; simpl;
try rewrite H;
constructor; split; simpl; auto;
try rewrite H1; try rewrite H2; try rewrite H3; auto.
apply  not_order_Z'; auto.
destruct (min i2); try (right;auto;fail).
elim n; rewrite H3; constructor.
elim n; rewrite H3; constructor.
destruct (min i2); try (right;auto;fail).
elim n; rewrite H3; constructor.
elim n; rewrite H3; constructor.
apply  not_order_Z'; auto.
destruct (narrow_interval_eq_def' _ _ H) as [(H0,H1)|(H0,H1)].
unfold meet_interval.
destruct (order_Z'_dec (min i1) (min i2));
destruct (order_Z'_dec (max i1) (max i2));
case order_Z'_dec; intros; simpl;
try constructor.
elim (order_Z'_antisym _ _ H0).
apply order_Z'_trans2 with (min i2); auto.
elim n0; apply order_Z'_trans with (1:=H0); auto.
elim n0; apply order_Z'_trans with (2:=H0); auto.
elim n1; auto.
unfold meet_interval.
destruct (order_Z'_dec (min i1) (min i2));
destruct (order_Z'_dec (max i1) (max i2));
case order_Z'_dec; intros; simpl;
try constructor.
elim n; auto.
elim n; apply order_Z'_trans2 with (min i2); auto.
elim n; apply order_Z'_trans3 with (max i1); auto.
elim n0; apply order_Z'_trans2 with (min i2); auto.
Qed.

Definition left_interval : Z -> interval.
refine 
 (fun b => ITV infty_neg (z b) _).
left; constructor.
Defined.

Definition right_interval : Z -> interval.
refine 
 (fun a => ITV (z a) infty_pos _).
left; constructor.
Defined.

Definition z_interval : forall a b :Z, (Zle a b) -> interval.
refine 
 (fun a b H => ITV (z a) (z b) _).
elim (Zle_lt_or_eq _ _ H); intros.
left; constructor; auto.
right; subst; auto.
Defined.

Definition infty_neg_interval : interval.
refine (ITV infty_neg infty_neg _).
auto.
Defined.

Definition infty_pos_interval : interval.
refine (ITV infty_pos infty_pos _).
auto.
Defined.

Lemma six_kind_of_interval : forall i : interval,
    (eq_interval i top_interval)
 \/ (eq_interval i infty_neg_interval)
 \/ (eq_interval i infty_pos_interval)
 \/ (exists b, (eq_interval i (left_interval b)))
 \/ (exists a, (eq_interval i (right_interval a)))
 \/ (exists a, exists b, exists H:Zle a b, (eq_interval i (z_interval a b H))).
destruct i.
destruct a; destruct b.
right; right; right; right; right.
exists z0; exists z1.
cut (z0 <= z1)%Z; intros.
exists H.
unfold eq_interval, z_interval; simpl; auto.
decompose [or] o.
inversion H; intros; auto with zarith.
inversion H; intros; auto with zarith.
right; right; right; right; left.
exists z0; unfold eq_interval, right_interval; simpl; auto.
elim o; intros H; inversion H.
elim o; intros H; inversion H.
right; right; left; unfold eq_interval,infty_pos_interval; simpl; auto.
elim o; intros H; inversion H.
right; right; right; left; unfold eq_interval,left_interval; simpl; eauto.
left; unfold eq_interval,top_interval; simpl; auto.
right; left; unfold eq_interval,infty_neg_interval; simpl; auto.
Qed.

Definition acc_interval i := forall int : interval,
    Acc
      (fun p1 p2 =>
       eq_interval (snd p1) (widen_interval (snd p2) (fst p1)) /\ 
         ~ eq_interval (snd p2) (snd p1)) 
      (int,i).

Lemma Acc_eq_acc : forall (A:Set) (Ord Eq:A->A->Prop),
 (forall x y z, Ord x y -> Eq y z -> Ord x z) ->
 forall x y, Acc Ord x -> Eq y x -> Acc Ord y.
intros.
inversion_clear H0.
constructor; intros.
apply H2.
apply H with y; auto.
Qed.

Lemma order_eq_order_interval : forall i1 i2 i3,
 order_interval i1 i2 -> eq_interval i2 i3 -> order_interval i1 i3.
intros; apply order_interval_trans with i2; auto.
apply order_interval_refl; auto.
Qed.

Lemma acc_interval_eq : forall i1 i2,
 acc_interval i1 -> eq_interval i1 i2 -> acc_interval i2.
unfold acc_interval; intros.
apply Acc_eq_acc with
 (fun p1 p2 : interval*interval => eq_interval (fst p1) (fst p2) /\
               eq_interval (snd p1) (snd p2))
 (int,i1); auto.
intros; intuition.
apply eq_interval_trans with (1:=H3).
apply widen_interval_eq1; auto.
elim H4; apply eq_interval_trans with (snd z0); auto.
split; simpl.
apply eq_interval_refl.
apply eq_interval_sym; auto.
Qed.

Lemma acc_interval_property_top : acc_interval top_interval.
unfold acc_interval; constructor.
unfold order_interval,eq_interval; simpl.
destruct y; simpl.
case (order_Z'_dec (min i) infty_neg); intro;
case (order_Z'_dec infty_pos (max i)); intro.
inversion_clear o.
inversion_clear o.
inversion_clear o.
clear n n0; intros ((H2,H3),H4).
elim H4; rewrite H2; rewrite H3; auto.
Qed.

Lemma acc_interval_property_left : forall b : Z,
 acc_interval (left_interval b).
constructor.
unfold left_interval,order_interval,eq_interval; simpl.
destruct y; simpl.
case (order_Z'_dec (min i) infty_neg); intro;
case (order_Z'_dec (z b) (max i)); intro.
inversion_clear o.
inversion_clear o.
intros ((H1,H1'),H2).
apply acc_interval_eq with top_interval.
apply acc_interval_property_top.
split; simpl; auto.
intros ((H1,H3),H2).
elim H2; split; auto.
Qed.

Lemma acc_interval_property_right : forall b : Z,
 acc_interval (right_interval b).
constructor.
unfold right_interval,order_interval,eq_interval; simpl.
destruct y; simpl.
case (order_Z'_dec infty_pos (max i)); intro;
case (order_Z'_dec (min i) (z b)); intro;
intros ((H2,H3),H4).
inversion_clear o.
inversion_clear o.
apply acc_interval_eq with top_interval.
apply acc_interval_property_top.
unfold eq_interval; auto.
elim H4; auto.
Qed.

Lemma acc_interval_property_infty_neg : acc_interval infty_neg_interval.
constructor.
unfold infty_neg_interval,order_interval,eq_interval; simpl.
destruct y; simpl.
case (order_Z'_dec (min i) infty_neg); intro;
case (order_Z'_dec infty_neg (max i)); intro;
intros ((H2,H3),H4).
inversion_clear o.
inversion_clear o.
apply acc_interval_eq with top_interval.
apply acc_interval_property_top.
unfold eq_interval; auto.
elim H4; auto.
Qed.

Lemma acc_interval_property_infty_pos : acc_interval infty_pos_interval.
constructor.
unfold infty_pos_interval,order_interval,eq_interval; simpl.
destruct y; simpl.
case (order_Z'_dec (min i) infty_pos); intro;
case (order_Z'_dec infty_pos (max i)); intro;
intros ((H2,H3),H4).
inversion_clear o0.
apply acc_interval_eq with top_interval.
apply acc_interval_property_top.
unfold eq_interval; auto.
inversion_clear o.
elim H4; auto.
Qed.

Lemma acc_interval_property_z : forall a b H,
 acc_interval (z_interval a b H).
constructor.
unfold z_interval,order_interval,eq_interval; simpl.
destruct y; simpl.
case (order_Z'_dec (min i) (z a)); intro;
case (order_Z'_dec (z b) (max i)); intro;
intros ((H2,H3),H4).
apply acc_interval_eq with top_interval.
apply acc_interval_property_top.
unfold eq_interval; auto.
apply acc_interval_eq with (left_interval b).
apply acc_interval_property_left.
unfold eq_interval; auto.
apply acc_interval_eq with (right_interval a).
apply acc_interval_property_right.
unfold eq_interval; auto.
elim H4; auto.
Qed.

Lemma acc_interval_property : forall i:interval, acc_interval i.
intros i; elim (six_kind_of_interval i); intuition.
apply acc_interval_eq with top_interval.
apply acc_interval_property_top.
apply eq_interval_sym; auto.
apply acc_interval_eq with infty_neg_interval.
apply acc_interval_property_infty_neg.
apply eq_interval_sym; auto.
apply acc_interval_eq with infty_pos_interval.
apply acc_interval_property_infty_pos.
apply eq_interval_sym; auto.
elim H0; intros.
apply acc_interval_eq with (left_interval x).
apply acc_interval_property_left.
apply eq_interval_sym; auto.
elim H; intros.
apply acc_interval_eq with (right_interval x).
apply acc_interval_property_right.
apply eq_interval_sym; auto.
decompose [ex] H.
elim H0; intros.
apply acc_interval_eq with (z_interval x x0 x1).
apply acc_interval_property_z.
apply eq_interval_sym; auto.
Qed. 

Definition acc_interval' i := forall int : interval,
    Acc
      (fun p1 p2 =>
       eqbot (Some (snd p1)) (narrow_interval (snd p2) (fst p1)) /\ 
         ~ eq_interval (snd p2) (snd p1)) 
      (int,i).

Lemma case_interval: forall i,
 ( min i = infty_neg /\ max i = infty_pos ) \/
 ( min i = infty_neg /\ max i <' infty_pos ) \/
 ( infty_neg <' min i /\ max i = infty_pos ) \/
 ( infty_neg <' min i /\ max i <' infty_pos ).
Proof.
  destruct i; simpl.
  destruct a; destruct b; simpl; auto;
  try (do 3 right; split; constructor) ||
      (do 2 right; left; split; constructor) ||
      (right; left; split; constructor).
Qed.

Definition sym_interval : nat -> interval.
intros n.
refine 
 (ITV (z (- (Z_of_nat n))) (z (Z_of_nat n)) _).
destruct n.
right; constructor; reflexivity.
left; constructor.
assert (0<Z_of_nat (S n))%Z.
simpl; constructor.
omega.
Defined.


Lemma meet_interval_eq1_order : forall i1 i2 j,
 eq_interval i1 i2 ->
 orderbot (meet_interval i1 j) (meet_interval i2 j).
Proof.
  intros.
  destruct H.
  generalize (meet_interval_bound1 i1 j).
  generalize (meet_interval_bound2 i1 j).
  generalize (meet_interval_least_bound i2 j).
  CaseEq (meet_interval i2 j);
  CaseEq (meet_interval i1 j); intros;
  rewrite H1 in *; rewrite H2 in *.
  apply H3; auto.
  inversion_clear H5.
  destruct H6; split. 
  rewrite <- H; auto.
  rewrite <- H0; auto.
  inversion_clear H4; auto.
  constructor.
  inversion_clear H4; inversion_clear H5.
  apply H3; auto.
  destruct H4; split. 
  rewrite <- H; auto.
  rewrite <- H0; auto. 
  constructor.
Qed.

Lemma orderbot_antisym : forall x y,
  orderbot x y -> orderbot y x -> eqbot x y.
Proof.
  intros.
  inversion_clear H in H0; inversion_clear H0; constructor.
  apply order_interval_antisym; auto.
Qed.

Lemma meet_interval_eq1 : forall i1 i2 j,
 eq_interval i1 i2 ->
 eqbot (meet_interval i1 j) (meet_interval i2 j).
Proof.
  intros; apply orderbot_antisym.
  apply meet_interval_eq1_order; auto.
  apply meet_interval_eq1_order; auto.
Qed.

Hint Resolve eq_interval_refl.

Lemma eqbot_refl : forall x, eqbot x x.
Proof.
  destruct x; constructor; auto.
Qed.

Hint Resolve eqbot_refl.

Lemma narrow_interval_eq1 : forall i1 i2 j,
  eq_interval i1 i2 ->
  eqbot (narrow_interval i1 j) (narrow_interval i2 j).
Proof.
  intros.
  destruct H.
  unfold narrow_interval.
  rewrite H; rewrite H0.
  destruct (max i2); destruct (min i2); auto.
Qed.

Lemma narrow_interval_eq2 : forall i1 i2 j,
  eq_interval i1 i2 ->
  eqbot (narrow_interval j i1) (narrow_interval j i2).
Proof.
  intros.
  destruct H.
  unfold narrow_interval.
  rewrite H; rewrite H0.
  destruct (max i2); destruct (min i2); auto.
Qed.

Lemma eqbot_trans : forall x y z,
  eqbot x y -> eqbot y z -> eqbot x z.
Proof.
  intros.
  inversion_clear H in H0; inversion_clear H0; constructor.
  apply eq_interval_trans with x2; auto.
Qed.

Lemma acc_interval'_eq : forall i1 i2,
 acc_interval' i1 -> eq_interval i1 i2 -> acc_interval' i2.
unfold acc_interval'; intros.
apply Acc_eq_acc with
 (fun p1 p2 : interval*interval => eq_interval (fst p1) (fst p2) /\
               eq_interval (snd p1) (snd p2))
 (int,i1); auto.
intros; intuition.
apply eqbot_trans with (narrow_interval (snd y) (fst x)); auto.
apply  narrow_interval_eq1; auto.
elim H4; apply eq_interval_trans with (snd z0); auto.
Qed.

Lemma order_Z'_eq_antisym : forall x y,
 x<='y -> y<='x -> x=y.
Proof.
  intros.
  destruct H; auto.
  elim (order_Z'_disj _ _ H0); auto.
Qed.

Inductive bound_length_interval (i:interval )(b:Z) : Prop :=
 bound_length_interval_def : forall z1 z2,
  min i = z z1 -> max i = z z2 -> ((z2-z1)<=b)%Z ->
  bound_length_interval i b.

Lemma inj_order_Z'_eq : forall x y, (z x)<='(z y) -> (x<=y)%Z.
Proof.
  intros.
  inversion_clear H.
  inversion_clear H0; omega.
  inversion_clear H0; omega.
Qed.

Lemma acc_interval'_property1 : forall i:interval, 
 infty_neg <' min i /\ max i <' infty_pos ->
 acc_interval' i.
Proof.
  intros i2 (H1,H2) i1.
  constructor; destruct y; simpl.
  CaseEq (narrow_interval i2 i).
  intros i3 H; rewrite H in *; generalize (narrow_interval_eq_def _ _ _ H); intuition.
  inversion_clear H4; elim H5; apply eq_interval_trans with i3; auto.
  rewrite H7 in H2; inversion_clear H2.
  rewrite H7 in H2; inversion_clear H2.
  rewrite H7 in H1; inversion_clear H1.
  intros H; generalize (narrow_interval_eq_def' _ _ H); intuition.
  rewrite H6 in H2; inversion_clear H2.
  rewrite H6 in H1; inversion_clear H1.
Qed.

Lemma acc_interval'_property2 : forall i:interval, 
 infty_neg <' min i /\ max i = infty_pos ->
 acc_interval' i.
Proof.
  intros i2 (H1,H2) i1.
  constructor; intros (j1,j2); simpl.
  unfold narrow_interval.
  unfold eq_interval.
  rewrite H2; destruct (min i2); 
  case order_Z'_dec; simpl; intros h (H,H3);
  inversion_clear H; destruct H0; simpl in *.
  apply acc_interval'_property1.
  rewrite H; split.
  constructor.
  destruct (max j2); try constructor.
  elim H3; split; auto.
  generalize (min_order_max j2); rewrite H; intros.
  inversion_clear H4.
  inversion_clear H5.
  elim H3; split; auto.
  inversion_clear H1.
Qed.

Lemma acc_interval'_property3 : forall i:interval, 
 infty_neg = min i /\ max i <' infty_pos ->
 acc_interval' i.
Proof.
  intros i2 (H1,H2) i1.
  constructor; intros (j1,j2); simpl.
  unfold narrow_interval.
  unfold eq_interval.
  rewrite <- H1; destruct (max i2); 
  case order_Z'_dec; simpl; intros h (H,H3);
  inversion_clear H; destruct H0; simpl in *.
  apply acc_interval'_property1.
  rewrite H0; split.
  destruct (min j2); try constructor.
  elim H3; split; auto.
  constructor.
  inversion_clear H2.
  generalize (min_order_max j2); rewrite H0; intros.
  inversion_clear H4.
  inversion_clear H5.
  elim H3; split; auto.
Qed.

Lemma eq_Z'_dec : forall x y : Z', {x=y}+{~x=y}.
Proof.
  destruct x; destruct y; try (left;constructor;reflexivity) ||
  (right; intros H; inversion_clear H;fail).
  destruct (Z_eq_dec z0 z1); intros.
  subst; left; constructor.
  right; intro; elim n.
  inversion_clear H; auto.
Qed.

Lemma acc_interval'_property4 : forall i:interval, 
 min i = infty_neg /\ max i = infty_pos ->
 acc_interval' i.
Proof.
  intros i2 (H1,H2) i1.
  constructor; intros (j1,j2); simpl.
  unfold narrow_interval.
  unfold eq_interval.
  rewrite H1; rewrite H2.
  case order_Z'_dec; simpl; intros h (H3,H4).
  elim (order_Z'_disj _ _ (min_order_max j1)); auto.
  inversion_clear H3.
  destruct H; simpl in *.
  destruct (eq_Z'_dec infty_neg (min j2)).
  apply acc_interval'_property3.
  split; auto.
  destruct (max j2); try constructor.
  elim H4; auto.
  destruct (eq_Z'_dec infty_pos (max j2)).
  apply acc_interval'_property2.
  split; auto.
  destruct (min j2); try constructor.
  elim H4; auto.
  apply acc_interval'_property1; split.
  destruct (min j2); try constructor; intuition.
  destruct (max j2); try constructor; intuition.
Qed.
  
Lemma acc_interval'_property : forall i:interval, acc_interval' i.
Proof.
  intros i1 i2.
  destruct (case_interval i1) as [H7|[H7|[H7|H7]]]; destruct H7.
  apply acc_interval'_property4; split; auto.
  apply acc_interval'_property3; split; auto.
  apply acc_interval'_property2; split; auto.
  apply acc_interval'_property1; split; auto.
Qed.

Lemma eq_interval_dec : forall i1 i2, {eq_interval i1 i2}+{~eq_interval i1 i2}.
destruct i1; destruct i2.
case (eq_Z'_dec a a0); intros.
case (eq_Z'_dec b b0); intros.
left; constructor; simpl; auto.
right; intros H; inversion_clear H; elim n; auto.
right; intros H; inversion_clear H; elim n; auto.
Qed.

Module IntervalLattice <: Lattice.

  Definition t : Set := option interval.

  Definition eq : t -> t -> Prop := eqbot.

  Lemma eq_refl : forall x : t, eq x x.
  Proof.
    destruct x; constructor.
    apply eq_interval_refl.
  Qed.

  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof.
    intros.
    inversion_clear H; constructor.
    apply eq_interval_sym; auto.
  Qed.

  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof.
    intros.
    inversion_clear H in H0; inversion_clear H0; constructor.
    apply eq_interval_trans with x2; auto.
  Qed.

  Lemma eq_dec : forall x y : t, {eq x y}+{~ eq x y}.
  Proof.
    destruct x; destruct y; try (right; intros H; inversion_clear H; fail).
    destruct (eq_interval_dec i i0).
    left; constructor; auto.
    right; intros H; inversion_clear H; elim n; auto.
    left; constructor.
  Qed.

  Definition order : t -> t -> Prop := orderbot.

  Lemma order_refl : forall x y : t, eq x y -> order x y.
  Proof.
    intros.
    inversion_clear H; constructor.
    apply order_interval_refl; auto.
  Qed.

  Lemma order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Proof.
    intros.
    inversion_clear H in H0; inversion_clear H0; constructor.
    apply order_interval_antisym; auto.
  Qed.

  Lemma order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Proof.
    intros.
    inversion_clear H in H0; inversion_clear H0; constructor.
    apply order_interval_trans with x2; auto.
  Qed.

  Lemma order_dec : forall x y : t, {order x y}+{~ order x y}.
  Proof.
    destruct x; destruct y; try (right; intros H; inversion_clear H; fail).
    destruct (order_interval_dec i i0).
    left; constructor; auto.
    right; intros H; inversion_clear H; elim n; auto.
    left; constructor.
    left; constructor.
  Qed.  

  Definition join : t -> t -> t := fun x y =>
    match (x,y) with
    | (None,_) => y
    | (_,None) => x
    | (Some x,Some y) => Some (join_interval x y)
    end.

  Lemma join_bound1 : forall x y : t, order x (join x y).
  Proof.
    destruct x; destruct y; unfold join; constructor.
    apply join_interval_bound1.
    apply order_interval_refl; apply eq_interval_refl.
  Qed.

  Lemma join_bound2 : forall x y : t, order y (join x y).
  Proof.
    destruct x; destruct y; unfold join; constructor.
    apply join_interval_bound2.
    apply order_interval_refl; apply eq_interval_refl.
  Qed.

  Lemma join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z.
  Proof.
    intros.
    inversion_clear H in H0; inversion_clear H0; unfold join; constructor; auto.
    apply join_interval_least_bound; auto.
  Qed.

  Definition meet  : t -> t -> t := fun x y =>
    match (x,y) with
    | (None,_) => None
    | (_,None) => None
    | (Some x,Some y) => meet_interval x y
    end.

  Lemma meet_bound1 : forall x y : t, order (meet x y) x.
  Proof.
    destruct x; destruct y; unfold meet; try constructor.
    apply meet_interval_bound1.
  Qed.

  Lemma meet_bound2 : forall x y : t, order (meet x y) y.
  Proof.
    destruct x; destruct y; unfold meet; try constructor.
    apply meet_interval_bound2.    
  Qed.

  Lemma meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y).
  Proof.
    intros.
    inversion_clear H in H0; inversion_clear H0; unfold meet; try constructor; auto.
    apply meet_interval_least_bound; auto.
  Qed.

  Definition bottom : t := None.

  Lemma bottom_is_bottom : forall x, order bottom x.
  Proof.
    intros; unfold order, bottom; constructor.
  Qed.

  Definition top : t := Some top_interval.
 
  Lemma top_is_top : forall x, order x top.
  Proof.
    intros; unfold order, top.
    destruct x; constructor.
    split; unfold top_interval; simpl; auto.
  Qed.

  Definition wid : t -> t -> t := fun x y =>
    match (x,y) with
    | (None,_) => y
    | (_,None) => x
    | (Some x,Some y) => Some (widen_interval x y)
    end.

Definition narr: t -> t -> t := fun x y =>
  match x,y with
    Some x,Some y => narrow_interval x y
  | _,_ => None
  end.
 
  Definition widen : t -> t -> t := fun x y =>
    match (x,y) with
    | (None,_) => y
    | (_,None) => x
    | (Some x,Some y) => Some (widen_interval x y)
    end.
 
  Definition widen_rel : (t*t) -> (t*t) -> Prop := fun x y =>
    eq (snd x) (widen (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Lemma widen_bound1 : forall x y : t, order x (widen x y).
  Proof.
    destruct x; destruct y; unfold widen; constructor.
    apply widen_interval_bound1.
    apply order_interval_refl; apply eq_interval_refl.
  Qed.

  Lemma widen_bound2 : forall x y : t, order y (widen x y).
  Proof.
    destruct x; destruct y; unfold widen; constructor.
    apply widen_interval_bound2.
    apply order_interval_refl; apply eq_interval_refl.
  Qed.

  Lemma widen_eq1 : forall x y z: t, eq x y -> eq (widen x z) (widen y z).
  Proof.
    intros x y l H; inversion_clear H; destruct l; unfold widen; constructor; auto.
    apply widen_interval_eq1; auto.
  Qed.

  Lemma widen_eq2 : forall x y z: t, eq x y -> eq (widen z x) (widen z y).
  Proof.
    intros x y l H; inversion_clear H; destruct l; unfold widen; constructor; auto.
    apply widen_interval_eq2; auto.
  Qed.

Definition widening : widening_operator eq order bottom.
Proof.
  exists widen.
  apply widen_bound1.
  apply widen_bound2.
  apply widen_eq1.
  apply widen_eq2.

  unfold widen, bottom.
  destruct x; constructor; auto.

  unfold widen, bottom; apply eq_refl.

  constructor; destruct y.
  unfold lattice_def.widen_rel; simpl.
  destruct t0. 
  destruct t1.
  intros _.
  assert (forall p,
      Acc
     (fun x0 y : t * t =>
      eq (snd x0) (widen (snd y) (fst x0)) /\ ~ eq (snd y) (snd x0))
     (Some (fst p), Some (snd p))).
  intros p.
  assert ( Acc
      (fun p1 p2 =>
       eq_interval (snd p1) (widen_interval (snd p2) (fst p1)) /\ 
         ~ eq_interval (snd p2) (snd p1)) p).
  destruct p; apply acc_interval_property.
  induction H as [p _ Hind].
  destruct p; simpl; constructor.
  destruct y; simpl.
  intros (H1,H2).
  destruct t0.
  destruct t1.
  apply Hind with (y:=(i3,i4)).
  simpl in H1; inversion_clear H1.
  split; simpl; auto.
  intro; elim H2; constructor; auto.
  simpl in H1; inversion_clear H1.
  unfold widen in H1; elim H2.
  apply eq_sym; auto.
  apply (H (i,i0)).
  simpl; intros (H,H1).
  destruct x.
  simpl in H; inversion_clear H.
  elim H1; apply eq_refl.
  intros (H1,H2); simpl in *.
  elim H2.
  replace x with (widen x None).
  apply eq_sym; auto.
  destruct x; unfold widen; auto.
Defined.

Definition narrow: t -> t -> t := fun x y =>
  match x,y with
    Some x,Some y => narrow_interval x y
  | _,_ => None
  end.

Lemma narrow_bound1 : forall x y, order (narrow x y) x.
Proof.
  destruct x; destruct y; simpl; try constructor.
  apply narrow_interval_bound1.
Qed.

Lemma narrow_bound2 : forall x y : t, order (meet x y) (narrow x y).
Proof.
  destruct x; destruct y; unfold meet; simpl; try constructor.
  apply narrow_interval_bound2.
Qed.

Lemma narrow_eq1 : forall x y z: t,
      eq x y -> eq (narrow x z) (narrow y z).
Proof.
  destruct x; destruct y; destruct z0; simpl; intros;
  try constructor.
  inversion_clear H; apply narrow_interval_eq1; auto.
  inversion_clear H.
  inversion_clear H.
Qed.

Lemma narrow_eq2 : forall x y z: t, 
      eq x y -> eq (narrow z x) (narrow z y).
Proof.
  destruct x; destruct y; destruct z0; simpl; intros;
  try constructor.
  inversion_clear H; apply narrow_interval_eq2; auto.
  inversion_clear H.
  inversion_clear H.
Qed.

Lemma narrow_acc_property : forall x:t, Acc (lattice_def.narrow_rel eq narrow) (x,x).
Proof.
  constructor; destruct y.
  unfold lattice_def.narrow_rel; simpl.
  destruct t1.
  destruct t0.
  intros _.
  assert (forall p, Acc
     (fun x0 y : t * t =>
      eq (snd x0) (narrow (snd y) (fst x0)) /\ ~ eq (snd y) (snd x0))
     (Some (fst p), Some (snd p))).
  intros p.
  assert ( Acc
      (fun p1 p2 =>
       eqbot (Some (snd p1)) (narrow_interval (snd p2) (fst p1)) /\ 
         ~ eq_interval (snd p2) (snd p1)) p).
  destruct p; apply acc_interval'_property.
  induction H as [p _ Hind].
  destruct p; constructor.
  destruct y; simpl; intros (H1,H2).
  destruct t1.
  destruct t0.
  apply (Hind (i4,i3)).
  simpl; split; auto.
  intro; elim H2; constructor; auto.
  inversion_clear H1.
  constructor; destruct y; simpl; intros (H3,H4).
  elim H4; apply eq_sym; auto.
  apply (H (i0,i)).
  unfold narrow; destruct x; intros (H,_); inversion_clear H.
  constructor; destruct y; simpl; intros (H3,H4).
  elim H4; apply eq_sym; auto.
Qed.

Definition narrowing : narrowing_operator eq order meet.
Proof.
  exists narrow.
  apply narrow_bound1.
  apply narrow_bound2.
  apply narrow_eq1.
  apply narrow_eq2.
  apply narrow_acc_property.
Defined.

End IntervalLattice.