Require Export interval.
Require Export z_extra_props.

Module IntervalOp.

Module Lat := IntervalLattice.

Inductive ZleI (x:Z) : Z' -> Prop :=
  ZleI_def0 : ZleI x infty_pos
| ZleI_def1 : forall y, x<=y -> ZleI x (z y).
Inductive ZgeI (x:Z) : Z' -> Prop :=
  ZgeI_def0 : ZgeI x infty_neg
| ZgeI_def1 : forall y, y<=x -> ZgeI x (z y).

Inductive gamma_ : Lat.t -> Z -> Prop :=
  gamma_def0 : forall i x, 
    ZgeI x (min i) -> ZleI x (max i) -> gamma_ (Some i) x.
Definition gamma := gamma_.

Lemma ZleI_trans1 : forall x y z,
  ZleI x y -> y <' z -> ZleI x z.
Proof.
  intros.
  inversion_clear H0 in H; inversion_clear H; try constructor.
  omega.
Qed.

Lemma ZleI_trans2 : forall x y z,
  ZleI x y -> y <=' z -> ZleI x z.
Proof.
  destruct 2.
  apply ZleI_trans1 with y; auto.
  subst; auto.
Qed.

Lemma ZgeI_trans1 : forall x y z,
  ZgeI x z -> y <' z -> ZgeI x y.
Proof.
  intros.
  inversion_clear H0 in H; inversion_clear H; try constructor.
  omega.
Qed.

Lemma ZgeI_trans2 : forall x y z,
  ZgeI x z -> y <=' z -> ZgeI x y.
Proof.
  destruct 2.
  apply ZgeI_trans1 with z; auto.
  subst; auto.
Qed.

Lemma gamma_monotone : forall N1 N2 n,
  Lat.order N1 N2 ->
  gamma N1 n ->
  gamma N2 n.
Proof.
  intros.
  inversion_clear H in H0; inversion_clear H0; constructor.
  apply ZgeI_trans2 with (min x1); inversion_clear H1; auto.
  apply ZleI_trans2 with (max x1); inversion_clear H1; auto.
Qed.

Lemma not_ZgeI_ZleI : forall x y, ZgeI x y -> ZleI x y -> y = z x.
Proof.
  destruct 1; intros H'; inversion_clear H'.
  replace x with y; auto.
  omega.
Qed.

Hint Resolve not_ZgeI_ZleI ZgeI_trans1 ZgeI_trans2 
 ZleI_trans1 ZleI_trans2.

Lemma gamma_meet_morph : forall N1 N2 n,
  gamma N1 n -> gamma N2 n -> gamma (Lat.meet N1 N2) n.
Proof.
  intros N1 N2 n H1; inversion_clear H1; intros H2; inversion_clear H2.

  unfold Lat.meet, meet_interval.
  case (order_Z'_dec (max i) (max i0));
  case (order_Z'_dec (min i) (min i0)); 
  case order_Z'_dec; intros;
  try (constructor; simpl; auto; fail).
  assert (min i0 = z n); [eauto|idtac].
  assert (max i = z n); [eauto|idtac].
  elim (order_Z'_strict _ _ o); rewrite H2; auto.
  assert (min i = z n); [eauto|idtac].
  assert (max i = z n); [eauto|idtac].
  elim (order_Z'_strict _ _ o); rewrite H2; auto.
  assert (min i0 = z n); [eauto|idtac].
  assert (max i0 = z n); [eauto|idtac].
  elim (order_Z'_strict _ _ o); rewrite H2; auto.
  assert (min i = z n); [eauto|idtac].
  assert (max i0 = z n); [eauto|idtac].
  elim (order_Z'_strict _ _ o); rewrite H2; auto.
Qed.

Definition sub1' (x:Z') : Z' :=
  match x with
    infty_neg => infty_neg
  | infty_pos => infty_pos
  | z a => z (a-1)
  end.

Definition add1' (x:Z') : Z' :=
  match x with
    infty_neg => infty_neg
  | infty_pos => infty_pos
  | z a => z (a+1)
  end.

Definition min' (x y : Z') : Z' :=
  match x with
    infty_neg => x
  | infty_pos => y
  | z a => match y with
             infty_neg => infty_neg
           | infty_pos => x
           | z b => z (Zmin a b) 
           end
  end.

Lemma order_Z'_eq_refl : forall x, x <=' x.
Proof.
  intros; right; auto.
Qed.
Hint Resolve order_Z'_eq_refl.

Lemma order_Z'_eq_Zle : forall x y,
 x <= y -> (z x) <=' (z y).
Proof.
  intros.
  destruct (Zle_lt_or_eq _ _ H).
  left; constructor; auto.
  right; subst; auto.
Qed.
Hint Resolve order_Z'_eq_Zle.

Lemma le_min'_l : forall x y, (min' x y) <=' x.
Proof.
  destruct x; destruct y; simpl; auto; try (repeat constructor; fail).
Qed.

Lemma le_min'_r : forall x y, (min' x y) <=' y.
Proof.
  destruct x; destruct y; simpl; auto; try (repeat constructor; fail).
Qed.

Lemma case_min' : forall (P:Z'->Prop) x y, 
  (x <=' y -> P x) ->
  (y <=' x -> P y) -> P (min' x y).
Proof.
  destruct x; destruct y; simpl; intros; auto;
  (apply H0; repeat constructor) ||
  (apply H; repeat constructor) || 
  (apply case_Zmin; auto).
Qed.

Lemma order_Z'_eq_infty_neg : forall x, infty_neg <=' x.
Proof.
  destruct x; (constructor 1; constructor) || (constructor 2; auto).
Qed.
Lemma order_Z'_eq_infty_pos : forall x, x <=' infty_pos.
Proof.
  destruct x; (constructor 1; constructor) || (constructor 2; auto).
Qed.
Hint Resolve order_Z'_eq_infty_pos order_Z'_eq_infty_neg.

Definition max' (x y : Z') : Z' :=
  match x with
    infty_neg => y
  | infty_pos => x
  | z a => match y with
             infty_neg => z a
           | infty_pos => infty_pos
           | z b => z (Zmax a b) 
           end
  end.

Lemma le_max'_l : forall x y, x <=' (max' x y).
Proof.
  destruct x; destruct y; simpl; auto; try (repeat constructor; fail).
Qed.

Lemma le_max'_r : forall x y, y <=' (max' x y).
Proof.
  destruct x; destruct y; simpl; auto; try (repeat constructor; fail).
Qed.

Lemma case_max' : forall (P:Z'->Prop) x y, 
  (y <=' x -> P x) ->
  (x <=' y -> P y) -> P (max' x y).
Proof.
  destruct x; destruct y; simpl; intros; auto;
  (apply H0; repeat constructor) ||
  (apply H; repeat constructor) || 
  (apply case_Zmax; auto).
Qed.

Lemma order_Z'_eq_min' : forall a b d,
  a <' d -> a <=' b -> a <=' (min' b (sub1' d)).
Proof.
  do 4 intro; apply case_min'; intros; auto.
  destruct d; simpl in *; auto.
  inversion_clear H; auto.
  apply order_Z'_eq_Zle; omega.
Qed.

Lemma order_Z'_eq_max' : forall a c d,
  a <' d -> c <=' d -> (max' (add1' a) c) <=' d.
Proof.
  do 4 intro; apply case_max'; intros; auto.
  destruct a; simpl in *; auto.
  inversion_clear H; auto.
  apply order_Z'_eq_Zle; omega.
Qed.

Definition BackTestEq (N1 N2:Lat.t) : (Lat.t*Lat.t) :=
  (Lat.meet N1 N2,Lat.meet N1 N2).

Definition BackTestLt (N1 N2:Lat.t) : (Lat.t*Lat.t) :=
  match N1,N2 with
  | Some (ITV a b h1),Some (ITV c d h2) =>
      (Lat.meet N1 (Some (ITV infty_neg (sub1' d) (order_Z'_eq_infty_neg _))),
       Lat.meet N2 (Some (ITV (add1' a) infty_pos (order_Z'_eq_infty_pos _))))
  | _,_ => (None,None)
  end.

Lemma Zmin_l : forall x y, x<=y -> Zmin x y = x.
Proof.
  intros; apply case_Zmin; auto with zarith.
Qed.

Lemma Zmin_r : forall x y, y<=x -> Zmin x y = y.
Proof.
  intros; apply case_Zmin; auto with zarith.
Qed.


Lemma min'_eq : forall b d,
  min' b d = (if order_Z'_dec b d then b else d).
Proof.
  destruct b; destruct d; simpl; case order_Z'_dec; 
  intros; 
  try solve[inversion_clear o|elim n;constructor]; auto.
  inversion_clear o.
  rewrite Zmin_l; try omega; auto.
  assert (z0<=z).
  case (Z_le_dec z0 z); auto.
  intro; elim n; constructor; omega.
  rewrite Zmin_r; auto.
Qed.

Lemma BackTestEq_correct : forall N1 N2 N1' N2' n1 n2,
  n1 = n2 ->
  gamma N1 n1 -> gamma N2 n2 ->
  BackTestEq N1 N2 = (N1',N2') ->
  gamma N1' n1 /\ gamma N2' n2.
Proof.
  unfold BackTestEq; intros N1 N2 N1' N2' n1 n2 H.
  intros; subst.
  inversion_clear H2; split; apply gamma_meet_morph; auto.
Qed.

Lemma BackTestLt_correct : forall N1 N2 N1' N2' n1 n2,
  n1 < n2 ->
  gamma N1 n1 -> gamma N2 n2 ->
  BackTestLt N1 N2 = (N1',N2') ->
  gamma N1' n1 /\ gamma N2' n2.
Proof.
  unfold BackTestLt; intros N1 N2 N1' N2' n1 n2 H.
  destruct N1; destruct N2; intros;
  injection H2; intros; clear H2; subst; auto;
  try solve [inversion_clear H1;inversion_clear H0].
  generalize H3; clear H3; destruct i; destruct i0.
  intros H3; injection H3; intros; subst; clear H3.
  inversion_clear H0; inversion_clear H1; simpl in *.
  split.
  apply gamma_meet_morph; auto; repeat (constructor;simpl); auto.
  inversion_clear H4; simpl; constructor; omega.
  apply gamma_meet_morph; auto; repeat (constructor;simpl); auto.
  inversion_clear H2; simpl; constructor; omega.
Qed.

Definition minus_Z' (x:Z') : Z' :=
 match x with
 | z y => z (-y)
 | infty_pos => infty_neg
 | infty_neg => infty_pos
 end.

Lemma minus_Z'_inv_lt : forall x y,
 x <' y -> (minus_Z' y) <' (minus_Z' x).
Proof.
  destruct 1; simpl; try constructor.
  omega.
Qed.

Lemma minus_Z'_inv_le : forall x y,
 x <=' y  -> (minus_Z' y) <=' (minus_Z' x).
Proof.
  destruct 1.
  left; apply minus_Z'_inv_lt; auto.
  subst; auto.
Qed.

Definition Minus (N:Lat.t) : Lat.t :=
  match N with
    None => None
  | Some (ITV x y H) => Some (ITV (minus_Z' y) (minus_Z' x) (minus_Z'_inv_le _ _ H))
  end.

Lemma Minus_correct : forall N n,
  gamma N n -> gamma (Minus N) (-n).
Proof.
  destruct 1.
  destruct i; simpl in *; constructor; simpl.
  inversion_clear H0; simpl; constructor; omega.
  inversion_clear H; simpl; constructor; omega.
Qed.

Definition add' (x y:Z') : Z' :=
  match (x,y) with
   | (infty_pos,infty_neg) => infty_pos
   | (infty_pos,_) => infty_pos
   | (infty_neg,infty_pos) => infty_pos
   | (infty_neg,_) => infty_neg
   | (_,infty_neg) => infty_neg
   | (_,infty_pos) => infty_pos
   | (z a,z b) => (z (a+b))
  end.

Inductive optionprop (A:Set) (P:A->Prop) : option A -> Prop :=
 optionprop_def : forall a, P a -> optionprop A P (Some a).
Implicit Arguments optionprop [A].
Inductive optionprop2 (A:Set) (P:A->A->Prop) : option A -> option A -> Prop :=
 optionprop2_def : forall a1 a2, P a1 a2 -> optionprop2 A P (Some a1) (Some a2).
Implicit Arguments optionprop2 [A].

Lemma add'_monotone_lt : forall x y t u, 
  x <' t -> y <' u -> (add' x y) <' (add' t u).
Proof.
  destruct 1; destruct 1; unfold add'; repeat constructor.
  omega.
Qed.

Lemma add'_monotone : forall x y t u, 
  x <=' t-> y <=' u -> (add' x y) <=' (add' t u).
Proof.
  destruct 1; destruct 1; subst.
  destruct H; destruct H0; unfold add'; repeat constructor.
  omega.
  destruct H; destruct u; unfold add'; auto; try constructor;
  try (repeat constructor; intuition;fail); auto.
  destruct H0; destruct t; unfold add';  auto; try constructor;
  try (repeat constructor; intuition;fail); auto.
  auto.
Qed.

Lemma add'_monotone1 : forall x y t u, 
  ZleI x t -> ZleI y u -> ZleI (x+y) (add' t u).
Proof.
  destruct 1; destruct 1; unfold add'; repeat constructor.
  omega.
Qed.

Lemma add'_monotone2 : forall x y t u, 
  ZgeI x t -> ZgeI y u -> ZgeI (x+y) (add' t u).
Proof.
  destruct 1; destruct 1; unfold add'; repeat constructor.
  omega.
Qed.

Definition sub' (x y:Z') : Z' :=
  match (x,y) with
   | (infty_pos,infty_pos) => infty_pos
   | (infty_pos,_) => infty_pos
   | (infty_neg,infty_neg) => infty_neg
   | (infty_neg,_) => infty_neg
   | (_,infty_neg) => infty_pos
   | (_,infty_pos) => infty_neg
   | (z a,z b) => (z (a-b))
  end.

Lemma sub'_monotone : forall x y t u, 
  x <=' t-> y <=' u -> (sub' x u) <=' (sub' t y).
Proof.
  destruct 1; destruct 1; subst.
  destruct H; destruct H0; unfold sub'; repeat constructor.
  omega.
  destruct H; destruct u; unfold sub'; auto; try constructor;
  try (repeat constructor; intuition;fail); auto.
  destruct H0; destruct t; unfold sub'; auto; try constructor;
  try (repeat constructor; intuition;fail); auto.
  auto.
Qed.

Lemma sub'_monotone1 : forall x y t u, 
  ZleI x t -> ZgeI y u -> ZleI (x-y) (sub' t u).
Proof.
  destruct 1; destruct 1; unfold sub'; repeat constructor.
  omega.
Qed.

Lemma sub'_monotone2 : forall x y t u, 
  ZgeI x t -> ZleI y u -> ZgeI (x-y) (sub' t u).
Proof.
  destruct 1; destruct 1; unfold sub'; repeat constructor.
  omega.
Qed.

Inductive Zcomp0 (x:Z) : Set :=
 | Zcomp0_lt : x<0 -> Zcomp0 x
 | Zcomp0_0 : x=0 -> Zcomp0 x
 | Zcomp0_gt : 0<x -> Zcomp0 x.
Lemma Zcomp0_dec : forall z, Zcomp0 z.
Proof.
  destruct z.
  constructor 2; auto.
  constructor 3; constructor.
  constructor 1; constructor.
Qed.

Inductive Case_Z' : Z' -> Set :=
   Case_Z'_infty_neg : Case_Z' infty_neg
 | Case_Z'_z_neg : forall x, x<0 -> Case_Z' (z x)
 | Case_Z'_z_0 : Case_Z' (z 0)
 | Case_Z'_z_pos : forall x, x>0 -> Case_Z' (z x)
 | Case_Z'_infty_pos : Case_Z' infty_pos.
Lemma Case_Z'_dec : forall z, Case_Z' z.
Proof.
  destruct z.
  destruct z.
  constructor 3.
  constructor 4; constructor.
  constructor 2; constructor.
  constructor.
  constructor.
Qed.

Definition mult' (x y:Z') : Z' :=
  match (x,y) with
   | (infty_pos,infty_pos) => infty_pos
   | (infty_pos,infty_neg) => infty_neg
   | (infty_pos,z b) => 
       match Zcomp0_dec b with
         Zcomp0_lt _ => infty_neg
       | Zcomp0_0 _ => z 0
       | Zcomp0_gt _ => infty_pos
       end
   | (infty_neg,infty_pos) => infty_neg
   | (infty_neg,infty_neg) => infty_pos
   | (infty_neg,z b) => 
       match Zcomp0_dec b with
         Zcomp0_lt _ => infty_pos
       | Zcomp0_0 _ => z 0
       | Zcomp0_gt _ => infty_neg
       end
   | (z a,infty_pos) => 
       match Zcomp0_dec a with
         Zcomp0_lt _ => infty_neg
       | Zcomp0_0 _ => z 0
       | Zcomp0_gt _ => infty_pos
       end
   | (z a,infty_neg) => 
       match Zcomp0_dec a with
         Zcomp0_lt _ => infty_pos
       | Zcomp0_0 _ => z 0
       | Zcomp0_gt _ => infty_neg
       end
   | (z a,z b) => z (a*b)
  end.

Definition mult_min' (a b c d:Z') :=
  (min' (min' (mult' a c) (mult' b d))
        (min' (mult' b c) (mult' a d))).
Definition mult_max' (a b c d:Z') :=
  (max' (max' (mult' a c) (mult' b d))
        (max' (mult' b c) (mult' a d))).

Lemma order_Z'_eq_trans : forall x y t, x <=' y -> y <=' t -> x <=' t.
Proof.
  destruct 1.
  intros; apply order_Z'_trans1 with y; auto.
  subst; auto.
Qed.
Hint Resolve order_Z'_eq_trans.

Lemma mult'_monotone : forall a b c d, 
  (mult_min' a b c d) <=' (mult_max' a b c d).
Proof.
  unfold mult_max', mult_min'; intros.
  repeat (apply case_min');
  repeat (apply case_max'); eauto.
Qed.

Ltac prod0 x y := 
  cut (x*y=0); [intros HHH; destruct (Zmult_integral _ _ HHH)|omega].

Lemma min'_n_n : forall x, min' x x = x.
Proof.
  intros; apply case_min'; auto.
Qed.

Lemma Zmin_Zmin_1 : forall x y,
  Zmin x (Zmin y x) = Zmin y x.
Proof.
  intros; repeat (apply case_Zmin); auto with zarith.
Qed.

Lemma Zmin_Zmin_2 : forall x y,
  Zmin x (Zmin x y) = Zmin x y.
Proof.
  intros; repeat (apply case_Zmin); auto with zarith.
Qed.


Lemma prod_le_0 : forall x y, 0 <= x*y -> 
 (x<=0/\y<=0) \/ (x>=0/\y>=0).
Proof.
  intros.
  destruct (Zpos_or_not x); destruct (Zpos_or_not y); intuition;
  try (assert (x=0); [omega|subst]);
  try (assert (y=0); [omega|subst]).
  assert (x*y<=0).
  auto with zarith sign.
  prod0 x y; subst; omega.
  assert (x*y<=0).
  auto with zarith sign.
  prod0 x y; subst; omega.
Qed.

Lemma prod_ge_0 : forall x y, x*y <= 0 -> 
 (x<=0/\y>=0) \/ (x>=0/\y<=0).
Proof.
  intros.
  destruct (Zpos_or_not x); destruct (Zpos_or_not y); intuition;
  try (assert (x=0); [omega|subst]);
  try (assert (y=0); [omega|subst]).
  assert (0<=x*y).
  auto with zarith sign.
  prod0 x y; subst; omega.
  assert (0<=x*y).
  auto with zarith sign.
  prod0 x y; subst; omega.
Qed.

Ltac prodLe0 := 
  match goal with
   [id:0 <= (?x*?y)|- _] => 
        destruct (prod_le_0 x y id) as [HH|HH];
        destruct HH; clear id
 | [id:(?x*?y) <= 0|- _] => 
        destruct (prod_ge_0 x y id) as [HH|HH];
        destruct HH; clear id
  end.

Lemma mult_min'_monotone : forall a b c d x y,
  ZgeI x a -> ZleI x b -> 
  ZgeI y c -> ZleI y d -> ZgeI (x*y) (mult_min' a b c d).
Proof.
  unfold mult_min'; intros.
  destruct a as [a| | ]; inversion_clear H;
  destruct b as [b| | ]; inversion_clear H0;
  destruct c as [c| | ]; inversion_clear H1;
  destruct d as [d| | ]; inversion_clear H2; unfold mult';
  try (unfold min'; constructor;
       destruct (Mult_interval_correct_min_max a b c d x y); intuition)
  ||
  (repeat (case Zcomp0_dec; intros); subst; 
  try (apply False_ind; omega) || constructor; 
  try (assert (x=0); [omega|subst]);
  try (assert (y=0); [omega|subst]);
  repeat rewrite Zmult_0_l;
  repeat rewrite Zmult_0_r;
  repeat rewrite min'_n_n; unfold min'; constructor;
  try rewrite Zmin_Zmin_1;
  try rewrite Zmin_Zmin_2; 
  auto).
  apply case_Zmin; intros; try omega.
  repeat  prodLe0; try (apply False_ind; omega).
  destruct (Zpos_or_not y).
  apply Zle_trans with 0; auto with sign zarith.
  apply Zmult_ineq1; apply Zmult_bounds1; split; try omega.
  prodLe0; try (apply False_ind; omega).
  apply sign_rule2; omega.
  destruct (Zpos_or_not c).
  rewrite Zmin_l; auto with sign zarith.
  rewrite Zmin_r; auto with sign zarith.
  destruct (Zpos_or_not y).
  apply Zle_trans with 0; auto with sign zarith.
  apply Zmult_ineq1.
  apply Zmult_bounds1; split; try omega.
  destruct (Zpos_or_not d).
  rewrite Zmin_r; auto with sign zarith.
  destruct (Zpos_or_not y).
  apply Zmult_ineq1; auto with sign zarith.
  apply Zle_trans with 0; auto with sign zarith.
  rewrite Zmin_l; auto with sign zarith.
  destruct (Zpos_or_not d).
  rewrite Zmin_r; auto with sign zarith.
  destruct (Zpos_or_not y).
  apply Zmult_ineq2; auto with sign zarith.
  apply Zle_trans with 0; auto with sign zarith.
  rewrite Zmin_l; auto with sign zarith.
  destruct (Zpos_or_not a).
  rewrite Zmin_l; auto with sign zarith.
  rewrite Zmin_r; auto with sign zarith.
  destruct (Zpos_or_not x).
  apply Zle_trans with 0; auto with sign zarith.
  apply Zmult_ineq2; auto with sign zarith.
  destruct (Zpos_or_not a).
  rewrite Zmin_l; auto with sign zarith.
  rewrite Zmin_r; auto with sign zarith.
  destruct (Zpos_or_not x).
  apply Zle_trans with 0; auto with sign zarith.
  auto with sign zarith.
  auto with sign zarith.
  auto with sign zarith.
  auto with sign zarith.
  auto with sign zarith.
  destruct (Zpos_or_not b).
  rewrite Zmin_r; auto with sign zarith.
  destruct (Zpos_or_not x).
  auto with sign zarith.
  apply Zle_trans with 0; auto with sign zarith.
  rewrite Zmin_l; auto with sign zarith.
  destruct (Zpos_or_not b).
  rewrite Zmin_l; auto with sign zarith.
  destruct (Zpos_or_not x).
  auto with sign zarith.
  apply Zle_trans with 0; auto with sign zarith.
  rewrite Zmin_r; auto with sign zarith.
  auto with sign zarith.
  auto with sign zarith.
  auto with sign zarith.
  auto with sign zarith.
Qed.

Lemma order_Z'_eq_geI : forall x y, ZgeI x y -> y <=' (z x).
Proof.
  destruct 1; auto.
Qed.

Lemma geI_order_Z'_eq : forall x y, y <=' (z x) -> ZgeI x y.
Proof.
  destruct y; destruct 1; try constructor.
  inversion_clear H; omega.
  injection H; intros; omega.
  inversion_clear H.
  inversion_clear H.
Qed.

Lemma order_Z'_eq_leI : forall x y, ZleI x y -> (z x) <=' y.
Proof.
  destruct 1; auto.
Qed.

Lemma leI_order_Z'_eq : forall x y, (z x) <=' y -> ZleI x y.
Proof.
  destruct y; destruct 1; try constructor.
  inversion_clear H; try omega.
  injection H; intros; omega.
  inversion_clear H.
  inversion_clear H.
Qed.

Hint Constructors ZgeI ZleI.
Hint Resolve geI_order_Z'_eq leI_order_Z'_eq.

Lemma order_Z'_eq_infty_pos' : forall x,
 infty_pos <=' x -> x = infty_pos.
Proof.
  destruct 1; auto.
  inversion_clear H; auto.
Qed.
Lemma order_Z'_eq_infty_neg' : forall x,
 x <=' infty_neg -> x = infty_neg.
Proof.
  destruct 1; auto.
  inversion_clear H; auto.
Qed.

Ltac simp_order := 
  match goal with
    [id : infty_neg <=' _ |- _] => clear id
  | [id : infty_pos <=' _ |- _] => generalize (order_Z'_eq_infty_pos' _ id); clear id; intros id; subst
  | [id : _ <=' infty_pos |- _] => clear id
  | [id : _ <=' infty_neg |- _] => generalize (order_Z'_eq_infty_neg' _ id); clear id; intros id; subst
  end.

Lemma min'_l : forall x y, x<='y -> min' x y = x.
Proof.
  intros; apply case_min'; intros; auto.
  inversion_clear H in H0; inversion_clear H0; auto.
  inversion_clear H in H1; inversion_clear H1.
  apply False_ind; omega.
Qed.
Lemma min'_r : forall x y, y<='x -> min' x y = y.
Proof.
  intros; apply case_min'; intros; auto.
  inversion_clear H in H0; inversion_clear H0; auto.
  inversion_clear H in H1; inversion_clear H1.
  apply False_ind; omega.
Qed.

Lemma order_Z'_eq_prop : forall x y, z x <=' z y -> x <= y.
Proof.
  destruct 1; inversion_clear H; omega.
Qed.

Lemma mult'_x_pos1 : forall a x,
  a <=' z x -> x < 0 -> mult' a infty_pos = infty_neg.
Proof.
  unfold mult'; intros.
  destruct a; auto.
  generalize (order_Z'_eq_prop _ _ H); clear H; intros.
  case Zcomp0_dec; intros; auto; apply False_ind; omega.
  simp_order.
  discriminate H.
Qed.

Lemma le_min'1 : forall a b c d, mult_min' a b c d <=' mult' a c.
Proof.
  unfold mult_min'; intros.
  apply order_Z'_eq_trans with (min' (mult' a c) (mult' b d));
  apply le_min'_l.
Qed.
Lemma le_min'2 : forall a b c d, mult_min' a b c d <=' mult' a d.
Proof.
  unfold mult_min'; intros.
  apply order_Z'_eq_trans with (min' (mult' b c) (mult' a d));
  apply le_min'_l || apply le_min'_r.
Qed.
Lemma le_min'3 : forall a b c d, mult_min' a b c d <=' mult' b c.
Proof.
  unfold mult_min'; intros.
  apply order_Z'_eq_trans with (min' (mult' b c) (mult' a d));
  apply le_min'_l || apply le_min'_r.
Qed.
Lemma le_min'4 : forall a b c d, mult_min' a b c d <=' mult' b d.
Proof.
  unfold mult_min'; intros.
  apply order_Z'_eq_trans with (min' (mult' a c) (mult' b d));
  apply le_min'_l || apply le_min'_r.
Qed.
Hint Resolve le_min'1 le_min'2 le_min'3 le_min'4.

Lemma mult'_pos1 : forall a b y, z 0 <=' y -> a <=' b -> mult' a y <=' mult' b y.
Proof.
  destruct 1.
  inversion_clear H; unfold mult'; intros.
  destruct a; destruct b; destruct H; inversion_clear H;
  try (destruct Zcomp0_dec; try (apply False_ind; omega); auto).
  repeat constructor; apply Zmult_gt_0_lt_compat_r; omega.
  auto.
  destruct a; destruct b; destruct H; inversion_clear H; 
  repeat (case Zcomp0_dec; intro); auto;
  try (apply False_ind; omega).
  subst.
  intros H; unfold mult'; destruct a; destruct b; destruct H; inversion_clear H;
  repeat rewrite Zmult_0_l;
  repeat rewrite Zmult_0_r; 
  try (destruct Zcomp0_dec; try (apply False_ind; omega); auto);
  auto.
Qed.
Lemma Zlt_neg : forall x y, -x < -y -> y < x.
Proof.
  intros; omega.
Qed.

Lemma mult'_neg1 : forall a b y, y <=' z 0 -> a <=' b -> mult' b y <=' mult' a y.
Proof.
  destruct 1.
  inversion_clear H; unfold mult'; intros.
  destruct a; destruct b; destruct H; inversion_clear H;
  try (destruct Zcomp0_dec; try (apply False_ind; omega); auto).
  repeat constructor; apply Zlt_neg.
  replace (-(z*a0)) with (z*(-a0)); [idtac|ring].
  replace (-(z0*a0)) with (z0*(-a0)); [idtac|ring].
  apply Zmult_gt_0_lt_compat_r; omega.
  auto.
  destruct a; destruct b; destruct H; inversion_clear H; 
  repeat (case Zcomp0_dec; intro); auto;
  try (apply False_ind; omega).
  subst.
  intros H; unfold mult'; destruct a; destruct b; destruct H; inversion_clear H;
  repeat rewrite Zmult_0_l;
  repeat rewrite Zmult_0_r; 
  try (destruct Zcomp0_dec; try (apply False_ind; omega); auto);
  auto.
Qed.
Hint Resolve mult'_neg1 mult'_pos1.

Lemma mult'_l_0 : forall x, mult' x (z 0) = z 0.
Proof.
  unfold mult'; intros.
  destruct x;
  repeat rewrite Zmult_0_r; 
  repeat (case Zcomp0_dec; intro); 
  try (apply False_ind; omega); auto.
Qed.
Lemma mult'_r_0 : forall x, mult' (z 0) x = z 0.
Proof.
  unfold mult'; intros.
  destruct x;
  repeat rewrite Zmult_0_l; 
  repeat (case Zcomp0_dec; intro); 
  try (apply False_ind; omega); auto.
Qed.

Lemma mult_min'_monotone'1 : forall a b c d a',
  a' <=' a -> a <=' b -> c <=' d ->
  (mult_min' a' b c d) <=' (mult_min' a b c d).
Proof.
  intros; unfold mult_min' at 2 4; 
  repeat apply case_min'; intros; auto.
  destruct c as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' b infty_neg); auto with zarith.
  destruct c as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' b infty_neg); auto with zarith.
  destruct d as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' b infty_neg); auto with zarith.
  destruct d as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' b infty_neg); auto with zarith.
Qed.

Lemma mult_min'_monotone'2 : forall a b c d b',
  a <=' b -> b <=' b' -> c <=' d ->
  (mult_min' a b' c d) <=' (mult_min' a b c d).
Proof.
  intros; unfold mult_min' at 2 4; 
  repeat apply case_min'; intros; auto.
  destruct d as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' infty_neg); auto with zarith.
  destruct d as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' infty_neg); auto with zarith.
  destruct c as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' infty_neg); auto with zarith.
  destruct c as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' a infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' infty_neg); auto with zarith.
Qed.

Lemma min'_comm : forall x y, min' x y = min' y x.
Proof.
  unfold min'; intros.
  destruct x ;destruct y; auto.
  repeat (apply case_Zmin); intros; auto.
Qed.

Lemma mult'_comm : forall x y, mult' x y = mult' y x.
Proof.
  unfold mult'; intros; destruct x; destruct y;
  try (case Zcomp0_dec; intros); auto.
  rewrite Zmult_comm; auto.
Qed.

Lemma mult_min'_comm : forall a b c d, mult_min' a b c d = mult_min' c d a b.
Proof.
  intros; unfold mult_min'.
  repeat rewrite (mult'_comm c).
  repeat rewrite (mult'_comm d).
  rewrite (min'_comm (mult' a d)); auto.  
Qed.

Lemma mult_min'_monotone' : forall a b c d a' b' c' d',
  a' <=' a -> a <=' b -> b <=' b' -> 
  c' <=' c -> c <=' d -> d <=' d' ->
  (mult_min' a' b' c' d') <=' (mult_min' a b c d).
Proof.
  generalize order_Z'_eq_trans; intros.
  intros.
  apply order_Z'_eq_trans with (mult_min' a b' c' d').
  apply mult_min'_monotone'1; eauto.
  apply order_Z'_eq_trans with (mult_min' a b c' d').
  apply mult_min'_monotone'2; eauto.
  repeat rewrite (mult_min'_comm a).
  apply order_Z'_eq_trans with (mult_min' c d' a b).
  apply mult_min'_monotone'1; eauto.
  apply mult_min'_monotone'2; eauto.
Qed.

Lemma le_max'1 : forall a b c d, mult' a c <=' mult_max' a b c d.
Proof.
  unfold mult_max'; intros.
  apply order_Z'_eq_trans with (max' (mult' a c) (mult' b d));
  apply le_max'_l.
Qed.
Lemma le_max'2 : forall a b c d, mult' a d <=' mult_max' a b c d.
Proof.
  unfold mult_max'; intros.
  apply order_Z'_eq_trans with (max' (mult' b c) (mult' a d));
  apply le_max'_l || apply le_max'_r.
Qed.
Lemma le_max'3 : forall a b c d, mult' b c <=' mult_max' a b c d.
Proof.
  unfold mult_max'; intros.
  apply order_Z'_eq_trans with (max' (mult' b c) (mult' a d));
  apply le_max'_l || apply le_max'_r.
Qed.
Lemma le_max'4 : forall a b c d, mult' b d <=' mult_max' a b c d.
Proof.
  unfold mult_max'; intros.
  apply order_Z'_eq_trans with (max' (mult' a c) (mult' b d));
  apply le_max'_l || apply le_max'_r.
Qed.
Hint Resolve le_max'1 le_max'2 le_max'3 le_max'4.

Lemma mult_max'_monotone'1 : forall a b c d a',
  a' <=' a -> a <=' b -> c <=' d ->
  (mult_max' a b c d) <=' (mult_max' a' b c d).
Proof.
  intros; unfold mult_max' at 1 3; 
  repeat apply case_max'; intros; auto.
  destruct c as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' a' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' infty_neg); auto with zarith.
  destruct c as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' a' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' infty_neg); auto with zarith.
  destruct d as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' a' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' infty_neg); auto with zarith.
  destruct d as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' a' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' a' infty_neg); auto with zarith.
Qed.

Lemma mult_max'_monotone'2 : forall a b c d b',
  a <=' b -> b <=' b' -> c <=' d ->
  (mult_max' a b c d) <=' (mult_max' a b' c d).
Proof.
  intros; unfold mult_max' at 1 3; 
  repeat apply case_max'; intros; auto.
  destruct d as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' a (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' a infty_neg); auto with zarith.
  destruct d as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' a (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' a infty_neg); auto with zarith.
  destruct c as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' a (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' a infty_neg); auto with zarith.
  destruct c as [v| | ].
  destruct (Zcomp0_dec v) as [H5|H5|H5].
  apply order_Z'_eq_trans with (mult' a (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' (z v)); auto with zarith.
  apply order_Z'_eq_trans with (mult' b' infty_pos); auto with zarith.
  apply order_Z'_eq_trans with (mult' a infty_neg); auto with zarith.
Qed.

Lemma max'_comm : forall x y, max' x y = max' y x.
Proof.
  unfold max'; intros.
  destruct x ;destruct y; auto.
  repeat (apply case_Zmax); intros; auto.
Qed.

Lemma mult_max'_comm : forall a b c d, mult_max' a b c d = mult_max' c d a b.
Proof.
  intros; unfold mult_max'.
  repeat rewrite (mult'_comm c).
  repeat rewrite (mult'_comm d).
  rewrite (max'_comm (mult' a d)); auto.  
Qed.

Lemma mult_max'_monotone' : forall a b c d a' b' c' d',
  a' <=' a -> a <=' b -> b <=' b' -> 
  c' <=' c -> c <=' d -> d <=' d' ->
  (mult_max' a b c d) <=' (mult_max' a' b' c' d').
Proof.
  generalize order_Z'_eq_trans; intros.
  intros.
  apply order_Z'_eq_trans with (mult_max' a' b c d).
  apply mult_max'_monotone'1; eauto.
  apply order_Z'_eq_trans with (mult_max' a' b' c d).
  apply mult_max'_monotone'2; eauto.
  repeat rewrite (mult_max'_comm a').
  apply order_Z'_eq_trans with (mult_max' c' d a' b').
  apply mult_max'_monotone'1; eauto.
  apply mult_max'_monotone'2; eauto.
Qed.

Lemma max'_n_n : forall x, max' x x = x.
Proof.
  intros; apply case_max'; auto.
Qed.

Lemma Zmax_Zmax_1 : forall x y,
  Zmax (Zmax y x) x = Zmax y x.
Proof.
  intros; repeat (apply case_Zmax); auto with zarith.
Qed.

Lemma Zmax_Zmax_2 : forall x y,
  Zmax (Zmax x y) x = Zmax x y.
Proof.
  intros; repeat (apply case_Zmax); auto with zarith.
Qed.

Lemma Zmax_l : forall x y, y <= x -> Zmax x y = x.
Proof.
  intros; apply case_Zmax; auto with zarith.
Qed.

Lemma Zmax_r : forall x y, x <= y -> Zmax x y = y.
Proof.
  intros; apply case_Zmax; auto with zarith.
Qed.



Lemma mult_max'_monotone : forall a b c d x y,
  ZgeI x a -> ZleI x b -> 
  ZgeI y c -> ZleI y d -> ZleI (x*y) (mult_max' a b c d).
Proof.
  unfold mult_max'; intros.
  destruct a as [a| | ]; inversion_clear H;
  destruct b as [b| | ]; inversion_clear H0;
  destruct c as [c| | ]; inversion_clear H1;
  destruct d as [d| | ]; inversion_clear H2; unfold mult';
  try (unfold max'; constructor;
       destruct (Mult_interval_correct_min_max a b c d x y); intuition)
  ||
  (repeat (case Zcomp0_dec; intros); subst; 
  try (apply False_ind; omega) || constructor; 
  try (assert (x=0); [omega|subst]);
  try (assert (y=0); [omega|subst]);
  repeat rewrite Zmult_0_l;
  repeat rewrite Zmult_0_r;
  repeat rewrite max'_n_n; unfold max'; constructor;
  try rewrite Zmax_Zmax_1;
  try rewrite Zmax_Zmax_2; 
  try (match goal with
         [|- ?x * ?y <= 0] => auto with sign zarith 
       | [|- _ * _ <= _ * _] => auto with sign zarith end);
  auto).
  destruct (Zpos_or_not c).
  rewrite Zmax_r; auto with sign zarith.
  rewrite Zmax_l; auto with sign zarith.
  destruct (Zpos_or_not y).
  apply Zle_trans with 0; auto with sign zarith.
  auto with sign zarith.
  destruct (Zpos_or_not y).
  apply Zle_trans with 0; auto with sign zarith.
  rewrite Zmax_l; auto with sign zarith.
  destruct (Zpos_or_not y).
  rewrite Zmax_r; auto with sign zarith.
  apply Zle_trans with 0; auto with sign zarith.
  destruct (Zpos_or_not y).
  rewrite Zmax_l; auto with sign zarith.
  destruct (Zpos_or_not d).
  rewrite Zmax_l; auto with sign zarith.
  rewrite Zmax_r; auto with sign zarith.
  destruct (Zpos_or_not a).
  rewrite Zmax_r; auto with sign zarith.
  rewrite Zmax_l; auto with sign zarith.
  destruct (Zpos_or_not x).
  apply Zle_trans with 0; auto with sign zarith.
  auto with sign zarith.
  destruct (Zpos_or_not a).
  rewrite Zmax_r; auto with sign zarith.
  rewrite Zmax_l; auto with sign zarith.
  destruct (Zpos_or_not x).
  apply Zle_trans with 0; auto with sign zarith.
  auto with sign zarith.
  destruct (Zpos_or_not b).
  rewrite Zmax_r; auto with sign zarith.
  destruct (Zpos_or_not x).
  auto with sign zarith.
  apply Zle_trans with 0; auto with sign zarith.
  rewrite Zmax_l; auto with sign zarith.
  destruct (Zpos_or_not b).
  rewrite Zmax_l; auto with sign zarith.
  destruct (Zpos_or_not x).
  auto with sign zarith.
  apply Zle_trans with 0; auto with sign zarith.
  rewrite Zmax_r; auto with sign zarith.
Qed.

Definition sem_op_add (N1 N2:Lat.t) : Lat.t := 
  match N1,N2 with
    Some i1,Some i2 => 
     Some (ITV (add' (min i1) (min i2))
               (add' (max i1) (max i2))
               (add'_monotone _ _ _ _ (min_order_max i1) (min_order_max i2)))
  | _,_ => None
  end.

Definition sem_op_sub (N1 N2:Lat.t) : Lat.t := 
  match N1,N2 with
  | Some i1,Some i2 => 
     Some (ITV (sub' (min i1) (max i2))
               (sub' (max i1) (min i2))
               (sub'_monotone _ _ _ _ (min_order_max i1) (min_order_max i2)))
  | _,_ => None
  end.

Definition sem_op_mult (N1 N2:Lat.t) : Lat.t := 
  match N1,N2 with
  | Some i1,Some i2 => 
     Some (ITV (mult_min' (min i1) (max i1) (min i2) (max i2))
               (mult_max' (min i1) (max i1) (min i2) (max i2))
               (mult'_monotone _ _ _ _))
  | _,_ => None
  end.

Lemma sem_op_add_correct : forall N1 N2 n1 n2 n,
  gamma N1 n1 -> gamma N2 n2 -> 
  n = n1 + n2 ->
  gamma (sem_op_add N1 N2) n.
Proof.
  intros N1 N2 n1 n2 n H; inversion_clear H.
  intros H2; inversion_clear H2.
  intros H4; inversion_clear H4; 
  unfold sem_op_add; constructor; simpl.
  apply add'_monotone2; auto.
  apply add'_monotone1; auto.
Qed.

Lemma sem_op_sub_correct : forall N1 N2 n1 n2 n,
  gamma N1 n1 -> gamma N2 n2 -> 
  n = n1 - n2 ->
  gamma (sem_op_sub N1 N2) n.
Proof.
  intros N1 N2 n1 n2 n H; inversion_clear H.
  intros H2; inversion_clear H2.
  intros H4; inversion_clear H4; 
  unfold sem_op_sub; constructor; simpl.
  apply sub'_monotone2; auto.
  apply sub'_monotone1; auto.
Qed.

Lemma sem_op_mult_correct : forall N1 N2 n1 n2 n,
  gamma N1 n1 -> gamma N2 n2 -> 
  n = n1 * n2 ->
  gamma (sem_op_mult N1 N2) n.
Proof.
  intros N1 N2 n1 n2 n H; inversion_clear H.
  intros H2; inversion_clear H2.
  intros H4; inversion_clear H4; 
  unfold sem_op_mult; constructor; simpl.
  apply mult_min'_monotone; auto.
  apply mult_max'_monotone; auto.
Qed.

Lemma backsem_add1 : forall (x y:Z) (a f:Z'),
 ZgeI (x+y) a -> ZleI y f ->
 ZgeI x (sub' a f).
Proof.
  intros.
  inversion_clear H0 in H; unfold sub'.
  inversion_clear H; constructor.
  inversion_clear H; constructor.
  omega.
Qed.

Lemma backsem_add2 : forall (x y:Z) (b e:Z'),
 ZleI (x+y) b -> ZgeI y e -> 
 ZleI x (sub' b e).
Proof.
  intros.
  inversion_clear H0 in H; unfold sub'.
  inversion_clear H; constructor.
  inversion_clear H; constructor.
  omega.
Qed.

Lemma not_ZgeI_ZleI_2 : forall a b x,
  ZgeI x b ->
  ZleI x a ->
  ~ a <' b.
Proof.
  red; intros.
  inversion_clear H in H0 H1; inversion_clear H1 in H0; inversion_clear H0.
  omega.
Qed.

Definition backsem_add_int (a b c d e f:Z') : Lat.t :=
  let max := min' d (sub' b e) in
  let min := max' c (sub' a f) in
    match order_Z'_dec max min with
      left _ => None
    | right h => Some (ITV min max (not_order_Z' _ _ h))
   end.


Definition backsem_add (N N1 N2:Lat.t) : Lat.t * Lat.t :=
      match N,N1,N2 with
        None,_,_ => (None,None)
      | _,None,_ => (None,None)
      | _,_,None => (None,None)
      | Some (ITV a b h1),Some (ITV c d h2),Some (ITV e f h3) => 
          (backsem_add_int a b c d e f,backsem_add_int a b e f c d)
      end.

Lemma backsem_add_int_correct : forall (x y:Z) (a b c d e f:Z'),
 ZgeI (x+y) a -> ZleI (x+y) b ->
 ZgeI x c -> ZleI x d ->
 ZgeI y e -> ZleI y f ->
 gamma (backsem_add_int a b c d e f) x.
Proof.
  intros; unfold backsem_add_int.
  assert (ZgeI x (max' c (sub' a f))).
  apply case_max'; intro; auto.
  apply backsem_add1 with y; auto.
  assert (ZleI x (min' d (sub' b e))).
  apply case_min'; intro; auto.
  apply backsem_add2 with y; auto.
  destruct order_Z'_dec.
  elim (not_ZgeI_ZleI_2 _ _ _ H5 H6); auto.
  constructor; simpl; auto.
Qed.

Lemma backsem_add_correct : forall N N1 N2 n1 n2 n N1' N2',
  gamma N1 n1 -> gamma N2 n2 ->
  n = n1 + n2 -> gamma N n ->
  backsem_add N N1 N2 = (N1',N2') ->
  gamma N1' n1 /\ gamma N2' n2.
Proof.
  unfold backsem_add; intros until 4.
  inversion_clear H2; inversion_clear H; inversion_clear H0.
  unfold backsem_add; intros.
  destruct i; destruct i0; destruct i1; simpl in *.
  injection H0; intros; subst; clear H0; auto.
  split.  
  apply backsem_add_int_correct with n2; auto.
  apply backsem_add_int_correct with n1; auto.
  rewrite Zplus_comm; auto.
  rewrite Zplus_comm; auto.
Qed.

Lemma backsem_sub1 : forall (x y:Z) (a e:Z'),
 ZgeI (x-y) a -> ZgeI y e ->
 ZgeI x (add' a e).
Proof.
  intros.
  inversion_clear H0 in H; unfold add'.
  inversion_clear H; constructor.
  inversion_clear H; constructor.
  omega.
Qed.

Lemma backsem_sub2 : forall (x y:Z) (b f:Z'),
 ZleI (x-y) b -> ZleI y f -> 
 ZleI x (add' b f).
Proof.
  intros.
  inversion_clear H0 in H; unfold add'.
  inversion_clear H; constructor.
  inversion_clear H; constructor.
  omega.
Qed.

Definition backsem_sub_int (a b c d e f:Z') : Lat.t :=
  let max := min' d (add' b f) in
  let min := max' c (add' a e) in
    match order_Z'_dec max min with
      left _ => None
    | right h => Some (ITV min max (not_order_Z' _ _ h))
   end.

Lemma backsem_sub3 : forall (x y:Z) (a e:Z'),
 ZgeI (x-y) a -> ZleI x e ->
 ZleI y (sub' e a).
Proof.
  intros.
  inversion_clear H0 in H; unfold sub'.
  inversion_clear H; constructor.
  inversion_clear H; constructor.
  omega.
Qed.

Lemma backsem_sub4 : forall (x y:Z) (b f:Z'),
 ZleI (x-y) b -> ZgeI x f -> 
 ZgeI y (sub' f b).
Proof.
  intros.
  inversion_clear H0 in H; unfold sub'.
  inversion_clear H; constructor.
  inversion_clear H; constructor.
  omega.
Qed.

Definition backsem_sub_int' (a b c d e f:Z') : Lat.t :=
  let max := min' f (sub' d a) in
  let min := max' e (sub' c b) in
    match order_Z'_dec max min with
      left _ => None
    | right h => Some (ITV min max (not_order_Z' _ _ h))
   end.

Definition backsem_sub (N N1 N2:Lat.t) : Lat.t * Lat.t :=
      match N,N1,N2 with
        None,_,_ => (None,None)
      | _,None,_ => (None,None)
      | _,_,None => (None,None)
      | Some (ITV a b h1),Some (ITV c d h2),Some (ITV e f h3) => 
          (backsem_sub_int a b c d e f,
           backsem_sub_int' a b  c d e f)
      end.

Lemma backsem_sub_int_correct : forall (x y:Z) (a b c d e f:Z'),
 ZgeI (x-y) a -> ZleI (x-y) b ->
 ZgeI x c -> ZleI x d ->
 ZgeI y e -> ZleI y f ->
 gamma (backsem_sub_int a b c d e f) x.
Proof.
  intros; unfold backsem_sub_int.
  assert (ZgeI x (max' c (add' a e))).
  apply case_max'; intro; auto.
  apply backsem_sub1 with y; auto.
  assert (ZleI x (min' d (add' b f))).
  apply case_min'; intro; auto.
  apply backsem_sub2 with y; auto.
  destruct order_Z'_dec.
  elim (not_ZgeI_ZleI_2 _ _ _ H5 H6); auto.
  constructor; simpl; auto.
Qed.

Lemma backsem_sub_int'_correct : forall (x y:Z) (a b c d e f:Z'),
 ZgeI (x-y) a -> ZleI (x-y) b ->
 ZgeI x c -> ZleI x d ->
 ZgeI y e -> ZleI y f ->
 gamma (backsem_sub_int' a b c d e f) y.
Proof.
  intros; unfold backsem_sub_int'.
  assert (ZgeI y (max' e (sub' c b))).
  apply case_max'; intro; auto.
  apply backsem_sub4 with x; auto.
  assert (ZleI y (min' f (sub' d a))).
  apply case_min'; intro; auto.
  apply backsem_sub3 with x; auto.
  destruct order_Z'_dec.
  elim (not_ZgeI_ZleI_2 _ _ _ H5 H6); auto.
  constructor; simpl; auto.
Qed.

Lemma backsem_sub_correct : forall N N1 N2 n1 n2 n N1' N2',
  gamma N1 n1 -> gamma N2 n2 ->
  n = n1 - n2 -> gamma N n ->
  backsem_sub N N1 N2 = (N1',N2') ->
  gamma N1' n1 /\ gamma N2' n2.
Proof.
  unfold backsem_sub; intros until 4.
  inversion_clear H2; inversion_clear H; inversion_clear H0.
  unfold backsem_sub; intros.
  destruct i; destruct i0; destruct i1; simpl in *.
  injection H0; intros; subst; clear H0; auto.
  split.  
  apply backsem_sub_int_correct with n2; auto.
  apply backsem_sub_int'_correct with n1; auto.
Qed.

Definition backsem_mult (N N1 N2:Lat.t) : Lat.t * Lat.t :=
      match N,N1,N2 with
        None,_,_ => (None,None)
      | _,None,_ => (None,None)
      | _,_,None => (None,None)
      | Some (ITV a b h1),Some (ITV c d h2),Some (ITV e f h3) =>  
          (Some (ITV c d h2),Some (ITV e f h3))
      end.

Lemma backsem_mult_correct : forall N N1 N2 n1 n2 n N1' N2',
  gamma N1 n1 -> gamma N2 n2 ->
  n = n1 * n2 -> gamma N n ->
  backsem_mult N N1 N2 = (N1',N2') ->
  gamma N1' n1 /\ gamma N2' n2.
Proof.
  unfold backsem_mult; intros until 4.
  inversion_clear H2; inversion_clear H; inversion_clear H0.
  destruct i; destruct i0; destruct i1; simpl in *.
  intros H0; injection H0; intros; subst; clear H0; auto.
  split; constructor; simpl; auto.  
Qed.

Definition const (n:Z) : Lat.t := 
  Some (ITV (z n) (z n) (order_Z'_eq_refl _)).

Lemma const_correct : forall n, gamma (const n) n.
Proof.
  unfold const; intros; constructor; simpl; constructor; omega.
Qed.

Lemma const_complete : forall n N,
  gamma N n -> Lat.order (const n) N.
Proof.
  unfold const; intros.
  inversion_clear H; simpl; constructor.
  destruct i; simpl in *.
  inversion_clear H0 in o; inversion_clear H1 in o; constructor; 
  simpl; try (repeat constructor; fail) ||
             (apply order_Z'_eq_Zle; auto; fail).
Qed.

Lemma top_correct : forall n, gamma Lat.top n.
Proof.
  unfold Lat.top; constructor; simpl; constructor.
Qed.

Lemma bottom_empty : forall n, ~ gamma Lat.bottom n.
Proof.
  red; intros.
  inversion_clear H.
Qed.

End IntervalOp.

