Require Export List.

Set Implicit Arguments.

Section termination_criteria.

 Variable t : Set.
 Variable eq order : t -> t -> Prop.
 Variable bottom : t.
 Variable meet : t -> t -> t.

  Definition widen_rel (widen:t->t->t) : t*t -> t*t -> Prop := 
    fun x y =>
    eq  (snd x) (widen (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Record widening_operator : Set := WidOp {
    widen :> t -> t -> t;
    widen_bound1 : forall x y : t, order x (widen x y);
    widen_bound2 : forall x y : t, order y (widen x y);
    widen_eq1 : forall x y z: t, 
      eq x y -> eq (widen x z) (widen y z);
    widen_eq2 : forall x y z: t, 
      eq x y -> eq (widen z x) (widen z y);
    widen_bottom1 : forall x : t, eq x (widen x bottom);
    widen_bottom2 : forall x : t, eq x (widen bottom x);
    widen_acc_property : forall x:t, Acc (widen_rel widen) (x,x)
  }.

  Definition narrow_rel (narrow:t->t->t) : (t*t) -> (t*t) -> Prop := 
    fun x y =>
    eq (snd x) (narrow (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Record narrowing_operator : Set := NarOp {
    narrow :> t -> t -> t;
    narrow_bound1 : forall x y, order (narrow x y) x;
    narrow_bound2 : forall x y : t, order (meet x y) (narrow x y);
    narrow_eq1 : forall x y z: t, 
      eq x y -> eq (narrow x z) (narrow y z);
    narrow_eq2 : forall x y z: t, 
      eq x y -> eq (narrow z x) (narrow z y);
    narrow_acc_property : forall x:t, Acc (narrow_rel narrow) (x,x)
  }.

End termination_criteria.

Unset Implicit Arguments.

Module Type Lattice.

  Parameter t : Set.

  Parameter eq : t -> t -> Prop.
  Parameter eq_refl : forall x : t, eq x x.
  Parameter eq_sym : forall x y : t, eq x y -> eq y x.
  Parameter eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Parameter eq_dec : forall x y : t, {eq x y}+{~ eq x y}.

  Parameter order : t -> t -> Prop.
  Parameter order_refl : forall x y : t, eq x y -> order x y.
  Parameter order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Parameter order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Parameter order_dec : forall x y : t, {order x y}+{~ order x y}.

  Parameter join  : t -> t -> t.
  Parameter join_bound1 : forall x y : t, order x (join x y).
  Parameter join_bound2 : forall x y : t, order y (join x y).
  Parameter join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z.

  Parameter meet  : t -> t -> t.
  Parameter meet_bound1 : forall x y : t, order (meet x y) x.
  Parameter meet_bound2 : forall x y : t, order (meet x y) y.
  Parameter meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y).

  Parameter bottom : t.
  Parameter bottom_is_bottom : forall x : t, order bottom x.

  Parameter widening : widening_operator eq order bottom.
  Parameter narrowing : narrowing_operator eq order meet.

  Hint Immediate eq_sym.
  Hint Resolve eq_refl eq_trans 
               order_refl order_antisym order_trans
               join_bound1 join_bound2 join_least_upper_bound
               meet_bound1 meet_bound2 meet_greatest_lower_bound
               bottom_is_bottom.

End Lattice.


Section lattice_prop.

  Variable t : Set.

  Variable eq : t -> t -> Prop.
  Variable eq_refl : forall x : t, eq x x.
  Variable eq_sym : forall x y : t, eq x y -> eq y x.
  Variable eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.

  Variable order0 : t -> t -> Prop.
  Variable order_refl : forall x y : t, eq x y -> order0 x y.
  Variable order_antisym : forall x y : t, order0 x y -> order0 y x -> eq x y.
  Variable order_trans : forall x y z : t, order0 x y -> order0 y z -> order0 x z.

  Variable join  : t -> t -> t.
  Variable join_bound1 : forall x y : t, order0 x (join x y).
  Variable join_bound2 : forall x y : t, order0 y (join x y).
  Variable join_least_upper_bound : forall x y z : t, order0 x z -> order0 y z -> order0 (join x y) z.

  Variable meet  : t -> t -> t.
  Variable meet_bound1 : forall x y : t, order0 (meet x y) x.
  Variable meet_bound2 : forall x y : t, order0 (meet x y) y.
  Variable meet_greatest_lower_bound : forall x y z : t, order0 z x -> order0 z y -> order0 z (meet x y).

  Variable bottom : t.
  Variable bottom_is_bottom : forall x : t, order0 bottom x.

Lemma meet_monotone : forall x1 x2 y1 y2,
  order0 x1 x2 -> order0 y1 y2 -> order0 (meet x1 y1) (meet x2 y2).
Proof.
  eauto.
Qed.

Lemma join_monotone : forall x1 x2 y1 y2,
  order0 x1 x2 -> order0 y1 y2 -> order0 (join x1 y1) (join x2 y2).
Proof.
  eauto.
Qed.

Lemma order_eq_order0 :
 forall x y z : t, order0 x y -> eq y z -> order0 x z.
Proof.
  eauto.
Qed.

Lemma eq_order_order0 :
 forall x y z : t, eq x y -> order0 y z -> order0 x z.
Proof.
  intros x y z H H1; apply order_trans with y; auto.
Qed.

Lemma meet_sym : forall x y :t, eq (meet x y) (meet y x).
  auto.
Qed.

Lemma join_sym : forall x y :t, eq (join x y) (join y x).
  auto.
Qed.

Lemma meet_assoc : forall x y z, eq (meet x (meet y z)) (meet (meet x y) z).
Proof.
  intros; apply order_antisym; apply meet_greatest_lower_bound.
  apply meet_greatest_lower_bound; auto.
  apply order_trans with (meet y z); auto. 
  apply order_trans with (meet y z); auto. 
  apply order_trans with (meet x y); auto. 
  apply meet_greatest_lower_bound; auto.
  apply order_trans with (meet x y); auto. 
Qed.

Lemma join_assoc : forall x y z, eq (join x (join y z)) (join (join x y) z).
Proof.
  intros; apply order_antisym; apply join_least_upper_bound.
  apply order_trans with (join x y); auto.
  apply join_least_upper_bound; auto.
  apply order_trans with (join x y); auto.
  apply join_least_upper_bound; auto.
  apply order_trans with (join y z); auto. 
  apply order_trans with (join y z); auto. 
Qed.

Lemma meet_eq1 :
  forall x y z, eq x y -> eq (meet x z) (meet y z).
Proof.
  intros; apply order_antisym.
  apply meet_greatest_lower_bound; auto.
  apply order_eq_order0 with x; auto.
  apply meet_greatest_lower_bound; auto.
  apply order_eq_order0 with y; auto.
Qed.

Lemma join_eq1 :
  forall x y z, eq x y -> eq (join x z) (join y z).
Proof.
  intros; apply order_antisym.
  apply join_least_upper_bound; auto.
  apply eq_order_order0 with y; auto.
  apply join_least_upper_bound; auto.
  apply eq_order_order0 with x; auto.
Qed.

Lemma meet_eq2 :
  forall x y z, eq y z -> eq (meet x z) (meet x y).
Proof.
  intros; apply order_antisym.
  apply meet_greatest_lower_bound; auto.
  apply order_eq_order0 with z; auto.
  apply meet_greatest_lower_bound; auto.
  apply order_eq_order0 with y; auto.
Qed.

Lemma join_eq2 :
  forall x y z : t, eq y z -> eq (join x z) (join x y).
Proof.
  intros; apply order_antisym.
  apply join_least_upper_bound; auto.
  apply eq_order_order0 with y; auto.
  apply join_least_upper_bound; auto.
  apply eq_order_order0 with z; auto.
Qed.

Lemma meet_bottom1 : forall x : t, eq bottom (meet bottom x).
Proof.
  intros; apply order_antisym; auto.
Qed.

Lemma join_bottom1 : forall x : t, eq x (join bottom x).
Proof.
  intros; apply order_antisym; auto.
Qed.

Lemma meet_bottom2 : forall x : t, eq bottom (meet x bottom).
Proof.
  intros; apply order_antisym; auto.
Qed.

Lemma join_bottom2 : forall x : t, eq x (join x bottom).
Proof.
  intros; apply order_antisym; auto.
Qed.

Fixpoint join_list (l : list t) : t :=
  match l with
  | nil => bottom
  | a :: q => join a (join_list q)
  end.

Lemma join_list_greater :
 forall (l : list t) (x : t), In x l -> order0 x (join_list l).
Proof.
  induction l; simpl in |- *; intuition.
  subst; auto.
  apply order_trans with (join_list l); auto.
Qed.

Lemma join_list_least_greater :
 forall (l : list t) (z : t),
 (forall x : t, In x l -> order0 x z) -> order0 (join_list l) z.
Proof.
  induction l; simpl in |- *; intuition.
Qed.

Inductive list_order0 : list t -> list t -> Prop :=
  | list_order_nil : list_order0 nil nil
  | list_order_cons :
      forall (x1 x2 : t) (l1 l2 : list t),
      order0 x1 x2 -> list_order0 l1 l2 -> list_order0 (x1 :: l1) (x2 :: l2).

Lemma list_order_join_list :
 forall l1 l2 : list t,
 list_order0 l1 l2 -> order0 (join_list l1) (join_list l2).
Proof.
  induction 1; simpl in |- *; auto.
  apply join_least_upper_bound.
  apply order_trans with x2; auto.
  apply order_trans with (join_list l2); auto.
Qed.

Hint Resolve join_bottom1 join_bottom2 join_list_greater
  join_list_least_greater.


Require Import Inclusion.
Require Import Inverse_Image.


  Lemma ascending_chain_condition_to_widening_operator :
    well_founded (fun x y => ~ eq y x /\ order0 y x) -> 
    widening_operator eq order0 bottom .
  Proof.
    intros wf.
    exists join; intros; auto.
    apply join_eq1; auto.
    apply join_eq2; auto.
    set (R:=(fun x y : t => ~ eq y x /\ order0 y x)) in *.
    set (mes := fun p:t*t => snd p).
    apply Inclusion.Acc_incl with
      (fun p1 p2 => R (mes p1) (mes p2)).
    intros (x1,y1) (x2,y2); simpl; unfold R; intros.
    destruct H; simpl in *; intuition eauto.
    apply Acc_inverse_image with (f:=mes); auto.
  Defined.

  Lemma descending_chain_condition_to_narrowing_operator :
    well_founded (fun x y => ~ eq y x /\ order0 x y) -> 
    narrowing_operator eq order0 meet.
  Proof.
    intros wf.
    exists meet; intros; auto.
    apply meet_eq1; auto.
    apply meet_eq2; auto.
    set (R:=(fun x y : t => ~ eq y x /\ order0 x y)) in *.
    set (mes := fun p:t*t => snd p).
    apply Acc_incl with
      (fun p1 p2 => R (mes p1) (mes p2)).
    intros (x1,y1) (x2,y2); simpl; unfold R; intros.
    destruct H; simpl in *; split; auto.
    apply order_trans with (meet y2 x1); auto.
    apply Acc_inverse_image with (f:=mes); auto.
  Qed.


(* iter *)

 Fixpoint iter (f:t->t) (n : nat) {struct n} :  t ->  t :=
   match n with
   | O => fun x :  t => x
   | S p => fun x :  t => f (iter f p x)
   end.

 Lemma iter_assoc : forall f n x, f (iter f n x) = iter f n (f x).
 Proof.
   induction n; simpl in |- *; intros.
   reflexivity.
   rewrite IHn; reflexivity.
 Qed.

  Definition monotone (f :  t ->  t) :=
    forall x y :  t, order0 x y -> order0 (f x) (f y).

 Lemma iter_monotone : forall f, monotone f ->
   forall n : nat, monotone (iter f n).
 Proof.
   induction n; simpl in |- *; unfold monotone in |- *; intros; auto.
 Qed.

(* iter_list *)

 Fixpoint iter_listf (l : list ( t ->  t)) (x:t) {struct l} : t :=
   match l with
   | nil => x
   | f :: q => iter_listf q (join x (f x))
   end.

 Lemma iter_listf_is_monotone :
  forall l : list ( t ->  t),
  (forall f :  t ->  t, In f l -> monotone f) -> monotone (iter_listf l).
 Proof.
   unfold monotone in |- *; induction l; simpl in |- *; intros.
   auto.
   apply IHl; auto.
   apply join_monotone; auto.
 Qed.

 Lemma f_in_listf_prop1 :
  forall (l : list ( t ->  t)) (x :  t), order0 x (iter_listf l x).
 Proof.
   induction l; simpl in |- *; intros; auto.
   apply order_trans with (join x (a x)); auto.
 Qed.

 Lemma f_in_listf_prop2 :
  forall (l : list ( t ->  t)) (f :  t ->  t),
  In f l -> monotone f -> forall x :  t, order0 (f x) (iter_listf l x).
 Proof.
   induction l; simpl in |- *; intros; auto.
   elim H.
   decompose [or] H.
   subst.
   apply order_trans with (join x (f x)); auto.
   apply f_in_listf_prop1.
   apply order_trans with (f (join x (a x))); auto.
 Qed.

 Lemma f_in_listf_prop3 :
  forall l : list ( t ->  t),
   (forall f :  t ->  t, In f l -> monotone f) ->
  forall x :  t,
   (forall f :  t ->  t, In f l -> order0 (f x) x) ->
  order0 (iter_listf l x) x.
 Proof.
   induction l; simpl in |- *; intros; auto.
   apply order_trans with (join x (a x)); auto.
   apply IHl; auto.
   intros.
   apply order_trans with (f x).
   apply H; auto.
   apply order_trans with x; auto.
 Qed.

  Lemma narrow_x_x: forall n x,
    eq (@narrow _ eq order0 meet n x x) x.
  Proof.
    intros; apply order_antisym.
    apply narrow_bound1.
    apply order_trans with (meet x x); auto.
    apply narrow_bound2.
  Qed.


End lattice_prop.


