Require Export solve.
Require Export wf_prop.

Module option.

  Section A.

  Variable A : Set.

  Inductive eq (P : A -> A -> Prop) : option A -> option A -> Prop :=
    eq_none : eq P None None
  | eq_some : forall x y : A, P x y -> eq P (Some x) (Some y).

  Inductive order_top (P : A -> A -> Prop) : A -> option A -> Prop :=
    order_none_top : forall x : A, order_top P x None
  | order_some_top : forall x y : A, P x y -> order_top P x (Some y).
  Inductive order_bot (P : A -> A -> Prop) : option A -> A -> Prop :=
    order_none_bot : forall x : A, order_bot P None x
  | order_some_bot : forall x y : A, P x y -> order_bot P (Some x) y.

  End A.

  Implicit Arguments eq [A].
  Implicit Arguments order_top [A].
  Implicit Arguments order_bot [A].

End option.

Set Implicit Arguments.

Section termination_criteria.

 Variable t : Set.
 Variable eq order : t -> t -> Prop.
 Variable bottom : t.
 Variable meet : t -> t -> option t.

  Definition partial_widen_rel (widen:t->t->option t) : (t*t) -> (t*t) -> Prop := 
    fun x y =>
    option.eq eq (Some (snd x)) (widen (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Record partial_widening_operator : Set := WidOp {
    widen :> t -> t -> option t;
    widen_bound1 : forall x y w : t,
      widen x y = Some w ->  order x w;
    widen_bound2 : forall x y w : t, 
      widen x y = Some w ->  order y w;
    widen_eq1 : forall x y z: t, 
      eq x y -> option.eq eq (widen x z) (widen y z);
    widen_eq2 : forall x y z: t, 
      eq x y -> option.eq eq (widen z x) (widen z y);
    widen_acc_property : forall x:t, Acc (partial_widen_rel widen) (x,x)
  }.

  Definition partial_narrow_rel (narrow:t->t->option t) : (t*t) -> (t*t) -> Prop := 
    fun x y =>
    option.eq eq (Some (snd x)) (narrow (snd y) (fst x)) /\ 
    ~ eq (snd y) (snd x).

  Record partial_narrowing_operator : Set := NarOp {
    narrow :> t -> t -> option t;
    narrow_bound1 : forall x y n, 
      narrow x y = Some n -> order n x;
    narrow_bound2 : forall x y m : t,
      meet x y = Some m -> exists n,
      narrow x y = Some n /\ order m n;
    narrow_eq1 : forall x y z: t, 
      eq x y -> option.eq eq (narrow x z) (narrow y z);
    narrow_eq2 : forall x y z: t, 
      eq x y -> option.eq eq (narrow z x) (narrow z y);
    narrow_acc_property : forall x:t, Acc (partial_narrow_rel narrow) (x,x)
  }.

End termination_criteria.

Unset Implicit Arguments.

Module Type PartialLattice.

  Parameter t : Set.

  Parameter eq : t -> t -> Prop.
  Parameter eq_refl : forall x : t, eq x x.
  Parameter eq_sym : forall x y : t, eq x y -> eq y x.
  Parameter eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Parameter eq_dec : forall x y : t, {eq x y}+{~ eq x y}.

  Parameter order : t -> t -> Prop.
  Parameter order_refl : forall x y : t, eq x y -> order x y.
  Parameter order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Parameter order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Parameter order_dec : forall x y : t, {order x y}+{~ order x y}.

  Parameter join  : t -> t -> option t.
  Parameter join_bound1 : forall x y j : t, join x y = Some j ->  order x j.
  Parameter join_bound2 : forall x y j : t, join x y = Some j ->  order y j.
  Parameter join_least_upper_bound : forall x y z : t, 
    order x z -> order y z ->
    exists j, join x y = Some j /\ order j z.

  Parameter meet  : t -> t -> option t.
  Parameter meet_bound1 : forall x y m : t, meet x y = Some m -> order m x.
  Parameter meet_bound2 : forall x y m : t, meet x y = Some m -> order m y.
  Parameter meet_greatest_lower_bound : forall x y z : t, 
    order z x -> order z y -> 
    exists m, meet x y = Some m /\ order z m.

  Parameter widening :     partial_widening_operator eq order.
  Parameter narrowing :     partial_narrowing_operator eq order meet.


  Hint Immediate eq_sym.
  Hint Resolve eq_refl eq_trans 
               order_refl order_antisym order_trans
               join_bound1 join_bound2 join_least_upper_bound
               meet_bound1 meet_bound2 meet_greatest_lower_bound.

End PartialLattice.

Module PartialLattice_of_Lattice (L:Lattice) <: PartialLattice
  with Definition t := L.t
  with Definition eq := L.eq
  with Definition order := L.order
  with Definition join := fun x y => Some (L.join x y)
  with Definition meet := fun x y => Some (L.meet x y).

  Definition t : Set := L.t.

  Definition eq : t -> t -> Prop := L.eq.
  Definition eq_refl : forall x : t, eq x x := L.eq_refl.
  Definition eq_sym : forall x y : t, eq x y -> eq y x := L.eq_sym.
  Definition eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z := L.eq_trans.
  Definition eq_dec : forall x y : t, {eq x y}+{~ eq x y} := L.eq_dec.

  Definition order : t -> t -> Prop := L.order.
  Definition order_refl : forall x y : t, eq x y -> order x y := L.order_refl.
  Definition order_antisym : forall x y : t, order x y -> order y x -> eq x y := L.order_antisym.
  Definition order_trans : forall x y z : t, order x y -> order y z -> order x z := L.order_trans.
  Definition order_dec : forall x y : t, {order x y}+{~ order x y} := L.order_dec.

  Definition join  : t -> t -> option t := fun x y => Some (L.join x y).
  Lemma join_bound1 : forall x y j : t, join x y = Some j ->  order x j.
  Proof.
    unfold join; intros.
    inversion_clear H; apply L.join_bound1.
  Qed.
  Lemma join_bound2 : forall x y j : t, join x y = Some j ->  order y j.
  Proof.
    unfold join; intros.
    inversion_clear H; apply L.join_bound2.
  Qed.
  Lemma join_least_upper_bound : forall x y z : t, 
    order x z -> order y z ->
    exists j, join x y = Some j /\ order j z.
  Proof.
    intros.
    exists (L.join x y); split; auto.
    apply L.join_least_upper_bound; auto.
  Qed.

  Definition meet  : t -> t -> option t := fun x y => Some (L.meet x y).
  Lemma meet_bound1 : forall x y m : t, meet x y = Some m -> order m x.
  Proof.
    unfold meet; intros.
    inversion_clear H; apply L.meet_bound1.
  Qed.
  Lemma meet_bound2 : forall x y m : t, meet x y = Some m -> order m y.
  Proof.
    unfold meet; intros.
    inversion_clear H; apply L.meet_bound2.
  Qed.
  Lemma meet_greatest_lower_bound : forall x y z : t, 
    order z x -> order z y -> 
    exists m, meet x y = Some m /\ order z m.
  Proof.
    intros.
    exists (L.meet x y); split; auto.
    apply L.meet_greatest_lower_bound; auto.
  Qed.


  Definition widening :
    partial_widening_operator eq order.
  Proof.
    destruct L.widening; unfold order, eq.
    exists (fun x y => Some (widen0 x y)).
    intros; inversion_clear H; auto.
    intros; inversion_clear H; auto.
    intros; constructor; auto.
    intros; constructor; auto.
    intros x.
    apply (Acc_ind (A:=(L.t*L.t)%type)
      (R:=(widen_rel L.eq widen0))); intros.
    clear H; constructor; intros (y1,y2) (H1,H2); simpl in *.
    inversion_clear H1.
    apply H0.
    split; auto.
    auto.
  Defined.

  Definition narrowing :
    partial_narrowing_operator eq order meet.
  Proof.
    destruct L.narrowing; unfold order, eq, meet.
    exists (fun x y => Some (narrow0 x y)).
    intros; inversion_clear H; auto.
    intros; inversion_clear H; eauto.
    intros; constructor; auto.
    intros; constructor; auto.
    intros x.
    apply (Acc_ind (A:=(L.t*L.t)%type)
      (R:=(narrow_rel L.eq narrow0))); intros.
    clear H; constructor; intros (y1,y2) (H1,H2); simpl in *.
    inversion_clear H1.
    apply H0.
    split; auto.
    auto.
  Defined.

End PartialLattice_of_Lattice.

Module lift_top_bot.

 Section A.
 Variable A : Set.
 Variable R : A -> A -> Prop.

  Inductive t : Set :=
    top : t
  | inj : A -> t
  | bot : t.

  Inductive eq : t -> t -> Prop :=
    eq_top : eq top top
  | eq_inj : forall a1 a2, R a1 a2 -> eq (inj a1) (inj a2)
  | eq_bot : eq bot bot.

  Inductive order : t -> t -> Prop :=
    order_top : forall x, order x top
  | order_inj : forall a1 a2, R a1 a2 -> order (inj a1) (inj a2)
  | order_bot : forall x, order bot x.

 End A.
Implicit Arguments eq [A].
Implicit Arguments order [A].
Implicit Arguments top [A].
Implicit Arguments bot [A].
Implicit Arguments inj [A].

End lift_top_bot.



Module Lattice_of_PartialLattice (L:PartialLattice) <: Lattice 
  with Definition t := lift_top_bot.t L.t
  with Definition eq := lift_top_bot.eq L.eq
  with Definition order := lift_top_bot.order L.order
  with Definition join := fun x y =>
    match x,y with
    | lift_top_bot.bot, _ => y
    | _, lift_top_bot.bot => x
    | lift_top_bot.inj a1,lift_top_bot.inj a2 => 
        match L.join a1 a2 with
          None => lift_top_bot.top
        | Some a => lift_top_bot.inj a
       end
    | _, _ => lift_top_bot.top
   end
  with Definition meet := fun x y =>
    match x,y with
    | lift_top_bot.top, _ => y
    | _, lift_top_bot.top => x
    | lift_top_bot.inj a1,lift_top_bot.inj a2 => 
        match L.meet a1 a2 with
          None => lift_top_bot.bot
        | Some a => lift_top_bot.inj a
       end
    | _, _ => lift_top_bot.bot
   end.

  Definition t := lift_top_bot.t L.t.
  
  Definition eq := lift_top_bot.eq L.eq.

  Lemma eq_refl : forall x : t, eq x x.
  Proof.
    destruct x; constructor; auto.
  Qed.

  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof.
    intros.
    inversion_clear H; constructor; auto.
  Qed.

  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof.
    intros.
    inversion_clear H in H0; inversion_clear H0; constructor; eauto.
  Qed.

  Lemma eq_dec : forall x y : t, {eq x y}+{~ eq x y}.
  Proof.
    destruct x; destruct y; try (left;constructor;fail) || (right;intros H;inversion_clear H; fail).
    destruct (L.eq_dec t0 t1).
    left; constructor; auto.
    right; intros H; elim n; inversion_clear H; auto.
  Qed.

  Definition order := lift_top_bot.order L.order.

  Lemma order_refl : forall x y : t, eq x y -> order x y.
  Proof.
    intros.
    inversion_clear H; constructor; auto.
  Qed.

  Lemma order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Proof.
    intros.
    inversion_clear H in H0; inversion_clear H0; constructor; eauto.
  Qed.

  Lemma order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Proof.
    intros.
    inversion_clear H in H0; inversion_clear H0; constructor; eauto.
  Qed.

  Lemma order_dec : forall x y : t, {order x y}+{~ order x y}.
  Proof.
    destruct x; destruct y; try (left;constructor;fail) || (right;intros H;inversion_clear H; fail).
    destruct (L.order_dec t0 t1).
    left; constructor; auto.
    right; intros H; elim n; inversion_clear H; auto.
  Qed.

  Definition join := fun x y =>
    match x,y with
    | lift_top_bot.bot, _ => y
    | _, lift_top_bot.bot => x
    | lift_top_bot.inj a1,lift_top_bot.inj a2 => 
        match L.join a1 a2 with
          None => lift_top_bot.top
        | Some a => lift_top_bot.inj a
       end
    | _, _ => lift_top_bot.top
   end.
  
  Lemma join_bound1 : forall x y : t, order x (join x y).
  Proof.
    unfold order; destruct x; destruct y; simpl; try constructor; auto.
    generalize (L.join_bound1 t0 t1).
    destruct L.join; constructor; auto.
  Qed.

  Lemma join_bound2 : forall x y : t, order y (join x y). 
  Proof.
    unfold order; destruct x; destruct y; simpl; try constructor; auto.
    generalize (L.join_bound2 t0 t1).
    destruct L.join; constructor; auto.
  Qed.

  Lemma join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z.
  Proof.
    unfold order; intros x y z H; inversion_clear H;
    intros H1; inversion_clear H1; simpl; try constructor; auto.
    destruct (L.join_least_upper_bound a1 a0 _ H0 H) as [j (H1,H2)].
    rewrite H1; constructor; auto.
  Qed.

  Definition meet := fun x y =>
    match x,y with
    | lift_top_bot.top, _ => y
    | _, lift_top_bot.top => x
    | lift_top_bot.inj a1,lift_top_bot.inj a2 => 
        match L.meet a1 a2 with
          None => lift_top_bot.bot
        | Some a => lift_top_bot.inj a
       end
    | _, _ => lift_top_bot.bot
   end.

  Lemma meet_bound1 : forall x y : t, order (meet x y) x.
  Proof.
    unfold order; destruct x; destruct y; simpl; try constructor; auto.
    generalize (L.meet_bound1 t0 t1).
    destruct L.meet; constructor; auto.
  Qed.

  Lemma meet_bound2 : forall x y : t, order (meet x y) y. 
  Proof.
    unfold order; destruct x; destruct y; simpl; try constructor; auto.
    generalize (L.meet_bound2 t0 t1).
    destruct L.meet; constructor; auto.
  Qed.

  Lemma meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y).
  Proof.
    unfold order; intros x y z H; inversion_clear H;
    intros H1; inversion_clear H1; simpl; try constructor; auto.
    destruct (L.meet_greatest_lower_bound _ _ _ H0 H) as [j (H1,H2)].
    rewrite H1; constructor; auto.
  Qed.

  Definition bottom : t := lift_top_bot.bot.

  Lemma bottom_is_bottom : forall x : t, order bottom x.
  Proof.
    unfold order, bottom; constructor.
  Qed.

  Lemma ascending_chain_condition :
    well_founded (fun x y => ~ L.eq y x /\ L.order y x) ->
    well_founded (fun x y => ~ eq y x /\ order y x).
  Proof.
    intros wf x.
    assert (Acc (fun x0 y : lift_top_bot.t L.t => ~ eq y x0 /\ order y x0) 
                lift_top_bot.top).
    constructor; intros y (H1,H2).
    inversion_clear H2 in H1; elim H1; constructor.
    assert (forall a, Acc (fun x0 y : lift_top_bot.t L.t => ~ eq y x0 /\ order y x0) 
                (lift_top_bot.inj a)).
    intros a; induction (wf a) as [a _ Hind].
    constructor; intros y (H1,H2).
    inversion_clear H2 in H1; auto.
    apply Hind; split; auto.
    intro; elim H1; constructor; auto.
    constructor; intros y (H1,H2).
    inversion_clear H2 in H1; auto.
    destruct y; auto.
    elim H1; constructor.
  Qed.


  Lemma widening :
    widening_operator eq order bottom.
  Proof.
    set (w:=L.widening).
    exists (fun x y =>
    match x,y with
    | lift_top_bot.bot, _ => y
    | _, lift_top_bot.bot => x
    | lift_top_bot.inj a1,lift_top_bot.inj a2 => 
        match widen w a1 a2 with
          None => lift_top_bot.top
        | Some a => lift_top_bot.inj a
       end
    | _, _ => lift_top_bot.top
   end).
   destruct x; destruct y; try constructor; auto.
   generalize (widen_bound1 w t0 t1).
   destruct (widen w t0 t1); constructor; auto.
   destruct x; destruct y; try constructor; auto.
   generalize (widen_bound2 w t0 t1).
   destruct (widen w t0 t1); constructor; auto.
   intros x y z H; inversion_clear H; destruct z; try constructor; auto.
   generalize (widen_eq1 w a1 a2 t0 H0); intros.
   inversion_clear H; constructor; auto.
   intros x y z H; inversion_clear H; destruct z; try constructor; auto.
   generalize (widen_eq2 w a1 a2 t0 H0); intros.
   inversion_clear H; constructor; auto.
   unfold bottom; destruct x; try constructor; auto.
   unfold bottom; destruct x; try constructor; auto.
   assert (   Acc
     (widen_rel eq 
        (fun x0 y : lift_top_bot.t L.t =>
         match x0 with
         | lift_top_bot.top =>
             match y with
             | lift_top_bot.top => lift_top_bot.top (A:=L.t)
             | lift_top_bot.inj _ => lift_top_bot.top (A:=L.t)
             | lift_top_bot.bot => x0
             end
         | lift_top_bot.inj a1 =>
             match y with
             | lift_top_bot.top => lift_top_bot.top (A:=L.t)
             | lift_top_bot.inj a2 =>
                 match widen w a1 a2 with
                 | Some a => lift_top_bot.inj a
                 | None => lift_top_bot.top (A:=L.t)
                 end
             | lift_top_bot.bot => x0
             end
         | lift_top_bot.bot => y
         end)) (lift_top_bot.top,lift_top_bot.top)).
   constructor; intros (y1,y2) (H1,H2); simpl in *.
   destruct y1; elim H2; apply eq_sym; auto.
   assert (forall a,   Acc
     (widen_rel eq 
        (fun x0 y : lift_top_bot.t L.t =>
         match x0 with
         | lift_top_bot.top =>
             match y with
             | lift_top_bot.top => lift_top_bot.top (A:=L.t)
             | lift_top_bot.inj _ => lift_top_bot.top (A:=L.t)
             | lift_top_bot.bot => x0
             end
         | lift_top_bot.inj a1 =>
             match y with
             | lift_top_bot.top => lift_top_bot.top (A:=L.t)
             | lift_top_bot.inj a2 =>
                 match widen w a1 a2 with
                 | Some a => lift_top_bot.inj a
                 | None => lift_top_bot.top (A:=L.t)
                 end
             | lift_top_bot.bot => x0
             end
         | lift_top_bot.bot => y
         end)) (lift_top_bot.inj a,lift_top_bot.top)).
   intros a.
   constructor; intros (y1,y2) (H1,H2); simpl in *.
   destruct y1; elim H2; apply eq_sym; auto.
   assert (forall a,   Acc
     (widen_rel eq 
        (fun x0 y : lift_top_bot.t L.t =>
         match x0 with
         | lift_top_bot.top =>
             match y with
             | lift_top_bot.top => lift_top_bot.top (A:=L.t)
             | lift_top_bot.inj _ => lift_top_bot.top (A:=L.t)
             | lift_top_bot.bot => x0
             end
         | lift_top_bot.inj a1 =>
             match y with
             | lift_top_bot.top => lift_top_bot.top (A:=L.t)
             | lift_top_bot.inj a2 =>
                 match widen w a1 a2 with
                 | Some a => lift_top_bot.inj a
                 | None => lift_top_bot.top (A:=L.t)
                 end
             | lift_top_bot.bot => x0
             end
         | lift_top_bot.bot => y
         end)) (lift_top_bot.inj (fst (a,a)),lift_top_bot.inj (snd (a,a)))).
   intros a; induction (widen_acc_property w a) as [(a1,a2) _ Hind].
   constructor; intros (y1,y2) (H1,H2); simpl in *.
   destruct y1.
   inversion_clear H1; auto.
   CaseEq (widen w a2 t0); intros; rewrite H3 in H1.
   inversion_clear H1 in H2.
   apply (Hind (t0,a0)); repeat split; simpl; auto.
   rewrite H3; constructor; auto.
   intro; elim H2; constructor; auto.
   inversion_clear H1; auto. 
   elim H2; apply eq_sym; auto.
   destruct x; auto.
   apply (H1 t0).
   (* bot,bot *)
   constructor; intros (y1,y2) (H4,H2); simpl in *.
   inversion_clear H4 in H2; auto.
   apply acc_eq with (eq:=fun p1 p2:t*t => eq (snd p1) (snd p2) /\ (fst p1)=(fst p2))
     (x:=(lift_top_bot.inj a2, lift_top_bot.inj a2)); simpl; auto.
   clear H H0 H1 y1 y2 a1 a2 H2 H3.
   intros (x1,x2) (y1,y2) (z1,z2) (H1,H2) (H4,H5); simpl in *; subst.
   repeat split; simpl; auto.
   inversion_clear H4 in H1 H2; auto.
   destruct x1; auto.
   generalize (widen_eq1 w a1 a2 t0 H); intros.
   inversion_clear H0 in H1; auto.
   apply eq_trans with (1:=H1); constructor; auto.
   apply eq_trans with (1:=H1); constructor; auto. 
   intro; elim H2; apply eq_trans with z2; auto.
   apply eq_sym; auto.
   repeat constructor; auto.
   elim H2; constructor.
  Defined.

  Lemma narrowing :
    narrowing_operator eq order meet.
  Proof.
    set (n:=L.narrowing).
    exists (fun x y =>
    match x,y with
    | lift_top_bot.top, _ => y
    | _, lift_top_bot.top => x
    | lift_top_bot.inj a1,lift_top_bot.inj a2 => 
        match narrow n a1 a2 with
          None => lift_top_bot.bot
        | Some a => lift_top_bot.inj a
       end
    | _, _ => lift_top_bot.bot
   end).
   intros.
   destruct x; try constructor.
   destruct y; try constructor; auto.
   generalize (narrow_bound1 n t0 t1).
   destruct (narrow n t0 t1); constructor; auto.
   destruct y; constructor.
   intros.
   destruct x; destruct y; simpl; try constructor; auto.
   generalize (narrow_bound2 n t0 t1).
   destruct (L.meet t0 t1); try constructor.
   intros H; destruct (H t2) as [nx (H1,H2)]; auto.
   rewrite H1; constructor; auto.
   intros x y z H; inversion_clear H; destruct z; try constructor; auto.
   generalize (narrow_eq1 n a1 a2 t0 H0); intros.
   inversion_clear H; constructor; auto.
   intros x y z H; inversion_clear H; destruct z; try constructor; auto.
   generalize (narrow_eq2 n a1 a2 t0 H0); intros.
   inversion_clear H; constructor; auto.
   assert (   Acc
     (narrow_rel eq 
        (fun x0 y : lift_top_bot.t L.t =>
         match x0 with
         | lift_top_bot.top => y
         | lift_top_bot.inj a1 =>
             match y with
             | lift_top_bot.top => x0
             | lift_top_bot.inj a2 =>
                 match narrow n a1 a2 with
                 | Some a => lift_top_bot.inj a
                 | None => lift_top_bot.bot (A:=L.t)
                 end
             | lift_top_bot.bot => lift_top_bot.bot (A:=L.t)
             end
         | lift_top_bot.bot =>
             match y with
             | lift_top_bot.top => x0
             | lift_top_bot.inj _ => lift_top_bot.bot (A:=L.t)
             | lift_top_bot.bot => lift_top_bot.bot (A:=L.t)
             end
         end)) (lift_top_bot.bot,lift_top_bot.bot)).
   constructor; intros (y1,y2) (H1,H2); simpl in *.
   destruct y1; elim H2; apply eq_sym; auto.
   assert (forall a,   Acc
     (narrow_rel eq 
        (fun x0 y : lift_top_bot.t L.t =>
         match x0 with
         | lift_top_bot.top => y
         | lift_top_bot.inj a1 =>
             match y with
             | lift_top_bot.top => x0
             | lift_top_bot.inj a2 =>
                 match narrow n a1 a2 with
                 | Some a => lift_top_bot.inj a
                 | None => lift_top_bot.bot (A:=L.t)
                 end
             | lift_top_bot.bot => lift_top_bot.bot (A:=L.t)
             end
         | lift_top_bot.bot =>
             match y with
             | lift_top_bot.top => x0
             | lift_top_bot.inj _ => lift_top_bot.bot (A:=L.t)
             | lift_top_bot.bot => lift_top_bot.bot (A:=L.t)
             end
         end)) (lift_top_bot.inj a,lift_top_bot.bot)).
   intros a.
   constructor; intros (y1,y2) (H1,H2); simpl in *.
   destruct y1; elim H2; apply eq_sym; auto.
   assert (forall a,   Acc
     (narrow_rel eq 
        (fun x0 y : lift_top_bot.t L.t =>
         match x0 with
         | lift_top_bot.top => y
         | lift_top_bot.inj a1 =>
             match y with
             | lift_top_bot.top => x0
             | lift_top_bot.inj a2 =>
                 match narrow n a1 a2 with
                 | Some a => lift_top_bot.inj a
                 | None => lift_top_bot.bot (A:=L.t)
                 end
             | lift_top_bot.bot => lift_top_bot.bot (A:=L.t)
             end
         | lift_top_bot.bot =>
             match y with
             | lift_top_bot.top => x0
             | lift_top_bot.inj _ => lift_top_bot.bot (A:=L.t)
             | lift_top_bot.bot => lift_top_bot.bot (A:=L.t)
             end
         end)) (lift_top_bot.inj (fst (a,a)),lift_top_bot.inj (snd (a,a)))).
   intros a; induction (narrow_acc_property n a) as [(a1,a2) _ Hind].
   constructor; intros (y1,y2) (H1,H2); simpl in *.
   destruct y1.
   elim H2; apply eq_sym; auto.
   CaseEq (narrow n a2 t0); intros; rewrite H3 in H1.
   inversion_clear H1 in H2.
   apply (Hind (t0,a0)); repeat split; simpl; auto.
   rewrite H3; constructor; auto.
   intro; elim H2; constructor; auto.
   inversion_clear H1; auto.
   inversion_clear H1; auto.
   destruct x; auto.
   (* top,top *)
   constructor; intros (y1,y2) (H4,H2); simpl in *.
   inversion_clear H4 in H1 H2; auto.
   elim H2; constructor; auto.
   apply acc_eq with (eq:=fun p1 p2:t*t => eq (snd p1) (snd p2) /\ (fst p1)=(fst p2))
     (x:=(lift_top_bot.inj a2, lift_top_bot.inj a2)); simpl; auto.
   clear H H0 H1 y1 y2 H2 a1 a2  H3.
   intros (x1,x2) (y1,y2) (z1,z2) (H1,H2) (H4,H5); simpl in *; subst.
   repeat split; simpl; auto.
   inversion_clear H4 in H1 H2; auto.
   destruct x1; auto.
   apply eq_trans with (1:=H1); constructor; auto.
   generalize (narrow_eq1 n a1 a2 t0 H); intros.
   inversion_clear H0 in H1; auto.
   apply eq_trans with (1:=H1); constructor; auto. 
   intro; elim H2; apply eq_trans with z2; auto.
   apply eq_sym; auto.
   repeat constructor; auto.
   apply (H1 t0); auto.
  Defined.

End Lattice_of_PartialLattice.


Module Type UnPartialLattice.

  Declare Module L : PartialLattice.

  Parameter t : Set.

  Parameter eq : t -> t -> Prop.
  Parameter eq_dec : forall x y : t, {eq x y}+{~ eq x y}.

  Parameter order : t -> t -> Prop.
  Parameter order_eq1 : forall x y : t, eq x y -> order x y /\ order y x.
  Parameter order_eq2 : forall x y : t, order x y -> order y x -> eq x y.
  Parameter order_dec : forall x y : t, {order x y}+{~ order x y}.

  Parameter join  : t -> t -> t.
  Parameter meet  : t -> t -> t.

  Parameter widen : t -> t -> t.
  Parameter narrow : t -> t -> t.

  Parameter i : t -> L.t.
  Parameter i_order_preserve : forall x y, order x y -> L.order (i x) (i y).
  Parameter inv_i_order_preserve : forall x y, L.order (i x) (i y) -> order x y.

  Parameter join_commut : forall x y,
    option.eq L.eq (Some (i (join x y))) (L.join (i x) (i y)).
  Parameter meet_commut : forall x y,
    option.eq L.eq (Some (i (meet x y))) (L.meet (i x) (i y)).
  Parameter widen_commut : forall x y, 
    option.eq L.eq (Some (i (widen x y))) (L.widening (i x) (i y)).
  Parameter narrow_commut : forall x y, 
    option.eq L.eq (Some (i (narrow x y))) (L.narrowing (i x) (i y)).

  Parameter bottom : t.
  Parameter bottom_is_bottom : forall x : t, order bottom x.

  Parameter widen_x_bot : forall x : t, eq x (widen x bottom).
  Parameter widen_bot_x : forall x : t, eq x (widen bottom x).

  Hint Resolve order_eq1 order_eq2 i_order_preserve inv_i_order_preserve
    join_commut meet_commut widen_commut narrow_commut widen_bot_x widen_x_bot 
    bottom_is_bottom.

End UnPartialLattice.

Module Lattice_of_UnPartialLattice (U:UnPartialLattice) <: Lattice
  with Definition t := U.t
  with Definition eq := U.eq
  with Definition order := U.order
  with Definition join := U.join
  with Definition meet := U.meet
  with Definition bottom := U.bottom.

  Definition t := U.t.
  Definition eq := U.eq.
  Definition order := U.order.
  Definition join := U.join.
  Definition meet := U.meet.
  Definition bottom := U.bottom.
  
  Lemma eq_refl : forall x : t, eq x x.
  Proof.
    unfold eq; auto.
  Qed.

  Lemma eq_sym : forall x y : t, eq x y -> eq y x.
  Proof.
    unfold eq; intros.
    destruct (U.order_eq1 _ _ H); auto.
  Qed.

  Lemma eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof.
    unfold eq; intros.
    destruct (U.order_eq1 _ _ H).
    destruct (U.order_eq1 _ _ H0).
    apply U.order_eq2; eauto.
  Qed.

  Definition eq_dec : forall x y : t, {eq x y}+{~ eq x y} := U.eq_dec.

  Lemma order_refl : forall x y : t, eq x y -> order x y.
  Proof.
    unfold eq, order; intros.
    destruct (U.order_eq1 _ _ H); auto.    
  Qed.

  Lemma order_antisym : forall x y : t, order x y -> order y x -> eq x y.
  Proof.
    unfold eq, order; intros.
    apply U.order_eq2; auto.
  Qed.

  Lemma order_trans : forall x y z : t, order x y -> order y z -> order x z.
  Proof.
    unfold order; eauto.
  Qed.

  Definition order_dec : forall x y : t, {order x y}+{~ order x y} := U.order_dec.

  Lemma i_injective : forall x y, U.L.eq (U.i x) (U.i y) -> eq x y.
  Proof.
    intros; apply order_antisym; apply U.inv_i_order_preserve; auto.
  Qed.

  Lemma i_preserve_eq : forall x y, eq x y -> U.L.eq (U.i x) (U.i y).
  Proof.
    unfold eq; intros.
    destruct (U.order_eq1 _ _ H); auto.    
  Qed.

  Lemma join_bound1 : forall x y : t, order x (join x y).
  Proof.
    intros.
    apply U.inv_i_order_preserve.
    generalize (U.L.join_bound1 (U.i x) (U.i y)); intros. 
    generalize (U.join_commut x y); intros.
    inversion_clear H0 in H.
    apply U.L.order_trans with y0; auto.
  Qed.
 
  Lemma join_bound2 : forall x y : t, order y (join x y).
  Proof.
    intros.
    apply U.inv_i_order_preserve.
    generalize (U.L.join_bound2 (U.i x) (U.i y)); intros. 
    generalize (U.join_commut x y); intros.
    inversion_clear H0 in H.
    apply U.L.order_trans with y0; auto.
  Qed.
 
  Lemma join_least_upper_bound : forall x y z : t, order x z -> order y z -> order (join x y) z.
  Proof.
    intros.
    destruct (U.L.join_least_upper_bound (U.i x) (U.i y) (U.i z)) as [w (H1, H2)]; auto.
    generalize (U.join_commut x y); rewrite H1; intros.
    inversion_clear H3.
    apply U.inv_i_order_preserve.
    eauto.
  Qed.

  Lemma meet_bound1 : forall x y : t, order (meet x y) x.
  Proof.
    intros.
    apply U.inv_i_order_preserve.
    generalize (U.L.meet_bound1 (U.i x) (U.i y)); intros. 
    generalize (U.meet_commut x y); intros.
    inversion_clear H0 in H.
    apply U.L.order_trans with y0; auto.
  Qed.

  Lemma meet_bound2 : forall x y : t, order (meet x y) y. 
  Proof.
    intros.
    apply U.inv_i_order_preserve.
    generalize (U.L.meet_bound2 (U.i x) (U.i y)); intros. 
    generalize (U.meet_commut x y); intros.
    inversion_clear H0 in H.
    apply U.L.order_trans with y0; auto.
  Qed.

  Lemma meet_greatest_lower_bound : forall x y z : t, order z x -> order z y -> order z (meet x y).
  Proof.
    intros.
    destruct (U.L.meet_greatest_lower_bound (U.i x) (U.i y) (U.i z)) as [w (H1, H2)]; auto.
    generalize (U.meet_commut x y); rewrite H1; intros.
    inversion_clear H3.
    apply U.inv_i_order_preserve.
    eauto.
  Qed.

  Definition bottom_is_bottom : forall x : t, order bottom x := U.bottom_is_bottom.

  Lemma ascending_chain_condition :
      well_founded (fun x y => ~ U.L.eq y x /\ U.L.order y x) ->
      well_founded (fun x y => ~ eq y x /\ order y x).
  Proof.
    intros.
    apply wf_mes with (mes:=U.i) (2:=H).
    intuition.
    elim H1; apply i_injective; auto.
  Qed.

  Definition widening : widening_operator eq order bottom.
  Proof.
    intros.
    exists U.widen.
    intros.
    apply U.inv_i_order_preserve.
    generalize (widen_bound1 U.L.widening (U.i x) (U.i y)); intros. 
    generalize (U.widen_commut x y); intros.
    inversion_clear H0 in H.
    apply U.L.order_trans with y0; auto.
    intros.
    apply U.inv_i_order_preserve.
    generalize (widen_bound2 U.L.widening (U.i x) (U.i y)); intros. 
    generalize (U.widen_commut x y); intros.
    inversion_clear H0 in H.
    apply U.L.order_trans with y0; auto.
    intros.
    apply i_injective.
    generalize (widen_eq1 U.L.widening (U.i x) (U.i y) (U.i z) (i_preserve_eq _ _ H)); intros. 
    generalize (U.widen_commut x z); intros.
    inversion_clear H1 in H0.
    generalize (U.widen_commut y z); intros.
    inversion_clear H1 in H0.
    inversion_clear H0; eauto.
    intros.
    apply i_injective.
    generalize (widen_eq2 U.L.widening (U.i x) (U.i y) (U.i z) (i_preserve_eq _ _ H)); intros. 
    generalize (U.widen_commut z x); intros.
    inversion_clear H1 in H0.
    generalize (U.widen_commut z y); intros.
    inversion_clear H1 in H0.
    inversion_clear H0; eauto.
    intros; apply U.widen_x_bot.
    intros; apply U.widen_bot_x.
    intros.
    apply acc_mes with (mes:=fun x => (U.i (fst x),U.i (snd x)))
                       (ltB:=partial_widen_rel U.L.eq (widen U.L.widening)).
    destruct a1; destruct a2; intros; simpl.
    destruct H as [HH HH2]; simpl in *; repeat split; simpl; auto.
    generalize (U.widen_commut t3 t0); intros.
    inversion_clear H.
    constructor.
    apply U.L.eq_trans with (2:=H0).
    apply i_preserve_eq; auto.
    intro; elim HH2; apply i_injective; auto.
    simpl; apply (widen_acc_property U.L.widening).
  Defined.

  Definition narrowing : narrowing_operator eq order meet.
  Proof.
    intros.
    exists U.narrow.
    intros.
    apply U.inv_i_order_preserve.
    generalize (narrow_bound1 U.L.narrowing (U.i x) (U.i y)); intros. 
    generalize (U.narrow_commut x y); intros.
    inversion_clear H0 in H.
    apply U.L.order_trans with y0; auto.
    intros.
    apply U.inv_i_order_preserve.
    generalize (narrow_bound2 U.L.narrowing (U.i x) (U.i y)); intros. 
    generalize (U.meet_commut x y); intros.
    inversion_clear H0 in H.
    destruct (H y0); auto.
    destruct H0; auto.
    generalize (U.narrow_commut x y); rewrite H0; intros.
    inversion_clear H3; eauto.
    intros.
    apply i_injective.
    generalize (narrow_eq1 U.L.narrowing (U.i x) (U.i y) (U.i z) (i_preserve_eq _ _ H)); intros. 
    generalize (U.narrow_commut x z); intros.
    inversion_clear H1 in H0.
    generalize (U.narrow_commut y z); intros.
    inversion_clear H1 in H0.
    inversion_clear H0; eauto.
    intros.
    apply i_injective.
    generalize (narrow_eq2 U.L.narrowing (U.i x) (U.i y) (U.i z) (i_preserve_eq _ _ H)); intros. 
    generalize (U.narrow_commut z x); intros.
    inversion_clear H1 in H0.
    generalize (U.narrow_commut z y); intros.
    inversion_clear H1 in H0.
    inversion_clear H0; eauto.
    intros.
    apply acc_mes with (mes:=fun x => (U.i (fst x),U.i (snd x)))
                       (ltB:=partial_narrow_rel U.L.eq (narrow U.L.narrowing)).
    destruct a1; destruct a2; intros; simpl.
    destruct H as [HH HH1]; simpl in *; repeat split; simpl; auto.
    generalize (U.narrow_commut t3 t0); intros.
    inversion_clear H.
    constructor.
    apply U.L.eq_trans with (2:=H0).
    apply i_preserve_eq; auto.
    intro; elim HH1; apply i_injective; auto.
    simpl; apply (narrow_acc_property U.L.narrowing).
  Defined.

End Lattice_of_UnPartialLattice.

(* 
 Local Variables: 
 coq-prog-name: "coqtop -I ../chap4 -I ../chap6  -emacs"
 End:
*)