let rec pos_of_int n =
  assert (n>=1);
  if n=1 then Coq.XH
  else if n land 1 = 1 then Coq.XI (pos_of_int (n lsr 1))
  else Coq.XO (pos_of_int (n lsr 1))

let rec nat_of_int n =
  assert (n>=0);
  if n=0 then Coq.O else Coq.S (nat_of_int (n-1))

let rec int_of_pos = function
  | Coq.XH -> 1
  | Coq.XO p -> 2*(int_of_pos p)
  | Coq.XI p -> 2*(int_of_pos p)+1

let int_of_z = function
  | Coq.Zpos p -> int_of_pos p
  | Coq.Z0 -> 0
  | Coq.Zneg p -> - (int_of_pos p)

let bool_of_sumbool = function
  | Coq.Left -> true
  | Coq.Right -> false

let bool_of_bool = function
  | Coq.True -> true
  | Coq.False -> false

let instruction pp = function
  | Coq.Nop -> "Nop"
  | Coq.Push z -> Printf.sprintf "Push %d" (int_of_z z)
  | Coq.Pop -> "Pop"
  | Coq.Dup -> "Dup"
  | Coq.Load x -> Printf.sprintf "Load %s" (pp x)
  | Coq.Store x -> Printf.sprintf "Store %s" (pp x)
  | Coq.Binop Coq.Add -> "Binop Add"
  | Coq.Binop Coq.Sub -> "Binop Sub"
  | Coq.Binop Coq.Mult -> "Binop Mult"
  | Coq.Newarray -> "Newarray"
  | Coq.Arraylength -> "Arraylength"
  | Coq.Arrayload -> "Arrayload"
  | Coq.Arraystore -> "Arraystore"
  | Coq.Input -> "Iinput"
  | Coq.If (Coq.Eq0,pc) -> Printf.sprintf "If Eq %d" (int_of_pos pc)
  | Coq.If (Coq.Ne,pc) -> Printf.sprintf "If Ne %d" (int_of_pos pc)
  | Coq.If (Coq.Gt0,pc) -> Printf.sprintf "If Gt0 %d" (int_of_pos pc)
  | Coq.If (Coq.Ge,pc) -> Printf.sprintf "If Ge %d" (int_of_pos pc)
  | Coq.Goto pc -> Printf.sprintf "Goto %d" (int_of_pos pc)

let rec expr0 = function
  | Coq.Const z -> (string_of_int (int_of_z z))
  | Coq.Var x -> Printf.sprintf "x%d" (int_of_pos x)
  | Coq.Bin (Coq.Add,e1,e2) -> Printf.sprintf "Add(%s,%s)" (expr0 e1) (expr0 e2)
  | Coq.Bin (Coq.Sub,e1,e2) -> Printf.sprintf "Sub(%s,%s)" (expr0 e1) (expr0 e2)
  | Coq.Bin (Coq.Mult,e1,e2) -> Printf.sprintf "Mult(%s,%s)" (expr0 e1) (expr0 e2)

let expr = function
  | Coq.Top -> "Top"
  | Coq.Base e -> expr0 e
  | Coq.Bot -> "Bot"

let z' = function
  | Coq.Z z -> Printf.sprintf "%3d" (int_of_z z)
  | Coq.Infty_pos -> "+oo"
  | Coq.Infty_neg -> "-oo"

let pad s n =
  if String.length s < n then
    s^(String.make (n-(String.length s)) ' ')
  else s


let sign_res p vars res =

  let sign = function
    | Coq.SignLattice.Coq_bot -> "BOT"
    | Coq.SignLattice.Coq_tops -> "TOP"
    | Coq.SignLattice.Coq_pos -> "POS"
    | Coq.SignLattice.Coq_neg -> "NEG"
    | Coq.SignLattice.Coq_zero -> "ZERO"
    | Coq.SignLattice.Coq_pos0 -> "POS0"
    | Coq.SignLattice.Coq_neg0 -> "NEG0"
    | Coq.SignLattice.Coq_not0 -> "NOT0" in

  let sign i = pad (sign i) 10 in
    
  let stack s =
    let rec aux = function
	Coq.Nil -> ""
      | Coq.Cons (e,q) -> Printf.sprintf "%s%s::" (aux q) (expr e)
    in
      match s with
	  Coq.Nil -> "[]"
	| Coq.Cons (e,q) -> Printf.sprintf "%s%s" (aux q) (expr e) in

  let stack s = pad (stack s) 30 in

  let rec locvar vars l =
    match vars with
	[] -> ""
      | (x,xs)::q ->
	  Printf.sprintf "%s:%s %s"
	    xs
	    (sign (Coq.SignAnalysis.LocalVarLattice.get l x))
	    (locvar q l) in

  let mem vars (Coq.Pair (s, l)) =
    match s with
      | Coq.Top -> "Top"
      | Coq.Base s -> Printf.sprintf "%s  %s" (stack s) (locvar vars l)
      | Coq.Bot -> "Bot" in

  let rec prog p vars res =
    let pp x = List.assoc x vars in
    match p with
      | Coq.Nil -> ""
      | Coq.Cons (Coq.Pair (pc,ins),Coq.Nil) -> 
	  Printf.sprintf "              // %s\n%2d: %s\n              // %s\n"
	    (mem vars (Coq.SignAnalysis.GlobalLattice.get res pc))
	    (int_of_pos pc)
	    (instruction pp ins)
	    (mem vars (Coq.SignAnalysis.GlobalLattice.get res (Coq.next pc)))
      | Coq.Cons (Coq.Pair (pc,ins),q) -> 
	  Printf.sprintf "              // %s\n%2d: %s\n%s"
	    (mem vars (Coq.SignAnalysis.GlobalLattice.get res pc))
	    (int_of_pos pc)
	    (instruction pp ins)
	    (prog q vars res) in
    prog p vars res

let interval_res p vars res =

  let interval = function
    | Coq.None -> "Bot"
    | Coq.Some (Coq.ITV (a, b)) -> Printf.sprintf "[%s,%s]" (z' a) (z' b) in

  let interval i = pad (interval i) 10 in
    
  let stack s =
    let rec aux = function
	Coq.Nil -> ""
      | Coq.Cons (e,q) -> Printf.sprintf "%s%s::" (aux q) (expr e)
    in
      match s with
	  Coq.Nil -> "[]"
	| Coq.Cons (e,q) -> Printf.sprintf "%s%s" (aux q) (expr e) in

  let stack s = pad (stack s) 30 in

  let rec locvar vars l =
    match vars with
	[] -> ""
      | (x,xs)::q ->
	  Printf.sprintf "%s:%s %s"
	    xs
	    (interval (Coq.IntervalAnalysis.LocalVarLattice.get l x))
	    (locvar q l) in

  let mem vars (Coq.Pair (s, l)) =
    match s with
      | Coq.Top -> "Top"
      | Coq.Base s -> Printf.sprintf "%s  %s" (stack s) (locvar vars l)
      | Coq.Bot -> "Bot" in

  let rec prog p vars res =
    let pp x = List.assoc x vars in
    match p with
      | Coq.Nil -> ""
      | Coq.Cons (Coq.Pair (pc,ins),Coq.Nil) -> 
	  Printf.sprintf "              // %s\n%2d: %s\n              // %s\n"
	    (mem vars (Coq.IntervalAnalysis.GlobalLattice.get res pc))
	    (int_of_pos pc)
	    (instruction pp ins)
	    (mem vars (Coq.IntervalAnalysis.GlobalLattice.get res (Coq.next pc)))
      | Coq.Cons (Coq.Pair (pc,ins),q) -> 
	  Printf.sprintf "              // %s\n%2d: %s\n%s"
	    (mem vars (Coq.IntervalAnalysis.GlobalLattice.get res pc))
	    (int_of_pos pc)
	    (instruction pp ins)
	    (prog q vars res) in
    prog p vars res



