Require Export Library.
Require Export List.
Require Export ZArith.
Open Scope Z_scope.

Definition word : Set := Word.
Definition eq_word_dec : forall w1 w2 : word, {w1=w2}+{w1<>w2} :=
  eq_word_dec.

Definition label := word.
Definition var := word.
Definition integer := Z.
Definition pc := Word.

Inductive binop := Add | Sub | Mult.
Inductive cmp := Eq | Ne | Gt | Ge.

Inductive instruction := 
  | Nop
  | Push (n:integer)
  | Pop 
  | Dup
  | Load (x:var)
  | Store (x:var)
  | Binop (op:binop)
  | Newarray
  | Arraylength
  | Arrayload
  | Arraystore
  | Input
  | If (c:cmp) (jump:pc)
  | Goto (jump:pc).

Definition program := list (pc * instruction).

Definition location := Word.
Inductive val : Set :=
  | Num (i:integer)
  | Ref (l:location).

Inductive var_val : Set := Undef | Def (v:val).
Definition locvar : Set := var -> var_val.
Definition opstack : Set := list val.

Record array : Set := {
  array_length : integer;
  array_values : forall i, 0 <= i < array_length -> integer
}.

Inductive heap_val : Set :=
  No | Array (a:array).
Definition heap := location -> heap_val.

Inductive state : Set := 
| St (pc:pc) (s:opstack) (l:locvar) (h:heap)
| Error.

Inductive subst : locvar -> var -> val -> locvar -> Prop :=
  subst_def : forall (env env' : locvar) x n,
    env' x = Def n ->
    (forall y, x <> y -> env y = env' y) ->
    subst env x n env'.
Inductive array_subst : array -> integer -> integer -> array -> Prop :=
  array_subst_def : forall a a' i n,
    forall h:0 <= i < array_length a',
    array_length a' = array_length a ->
    array_values a' i h = n ->
    (forall j,
       i<>j -> 
       forall h:0 <= j < array_length a,
       forall h':0 <= j < array_length a',
      array_values a' j h' = array_values a j h) ->
    array_subst a i n a'.
Inductive heap_subst : heap -> location -> heap_val -> heap -> Prop :=
  heap_subst_def : forall h h' loc hv,
    h' loc = hv ->
    (forall loc', loc<>loc' -> h' loc' = h loc') ->
    heap_subst h loc hv h'.
Inductive heap_update_array : heap -> location -> integer -> integer -> heap -> Prop :=
  heap_update_array_def : forall h loc i n h' a a',
     h loc = Array a ->
     array_subst a i n a' ->
     heap_subst h loc (Array a') h' ->
     heap_update_array h loc i n h'.

Inductive create_new_array : heap -> integer -> heap -> location -> Prop :=
 create_new_array_def : forall h n h' loc a,
   h loc = No ->
   (forall loc', loc'<>loc -> h loc' = h' loc') ->
   h' loc = Array a ->
   array_length a = n ->
   (forall i (h:0<=i<array_length a), array_values a i h = 0) ->
   create_new_array h n h' loc.

Definition binop_sem (binop:binop) : integer -> integer -> integer :=
  match binop with
    | Add => Zplus
    | Sub => Zminus
    | Mult => Zmult
  end.

Definition cmp_sem (cmp:cmp) : integer -> integer -> Prop :=
  match cmp with
    | Eq => fun x y => x=y
    | Ne => fun x y => x<>y
    | Gt => fun x y => x>y
    | Ge => fun x y => x>=y
  end.

Fixpoint instr_at (l:list (pc*instruction)) (n:pc) : option instruction :=
  match l with
    | nil => None
    | (i,ins)::q => if eq_word_dec i n then Some ins else instr_at q n
  end.

Definition next : pc -> pc := succ_word.

Reserved Notation "s1 -[ p ]-> s2" (at level 70, no associativity).
Inductive step (p:program) : state -> state -> Prop :=
 | step_nop : forall pc h s l,
     instr_at p pc = Some Nop ->
     (St pc s l h) -[p]-> (St (next pc) s l h)
 | step_push : forall pc h s l n,
     instr_at p pc = Some (Push n) ->
     (St pc s l h) -[p]-> (St (next pc) (Num n :: s) l h)
 | step_pop : forall pc h s l v,
     instr_at p pc = Some Pop ->
     (St pc (v::s) l h) -[p]-> (St (next pc) s l h)
 | step_dup : forall pc h s l v,
     instr_at p pc = Some Dup ->
     (St pc (v::s) l h) -[p]-> (St (next pc) (v::v::s) l h)
 | step_iload : forall pc h s l x v,
     instr_at p pc = Some (Load x) ->
     l x = Def v ->
     (St pc s l h) -[p]-> (St (next pc) (v::s) l h)
 | step_istore : forall pc h s l x v l',
     instr_at p pc = Some (Store x) ->
     subst l x v l' ->
     (St pc (v :: s) l h) -[p]-> (St (next pc) s l' h)
 | step_binop : forall pc h s l v1 v2 b,
     instr_at p pc = Some (Binop b) ->
     (St pc (Num v2:: Num v1::s) l h) 
         -[p]-> (St (next pc) (Num (binop_sem b v1 v2)::s) l h)
 | step_newarray_1 : forall pc h h' s l n loc,
     instr_at p pc = Some Newarray ->
     create_new_array h n h' loc ->
     (St pc (Num n :: s) l h) -[p]-> (St (next pc) (Ref loc :: s) l h')
 | step_newarray_2 : forall pc h s l n,
     instr_at p pc = Some Newarray ->
     n<0 ->
     (St pc (Num n :: s) l h) -[p]-> Error
 | step_arraylength : forall pc h a s l loc,
     instr_at p pc = Some Arraylength ->
     h loc = Array a ->
     (St pc (Ref loc :: s) l h)
       -[p]-> (St (next pc) (Num a.(array_length) :: s) l h)
 | step_iaload_1 : forall pc h a  i s l loc,
     instr_at p pc = Some Arrayload ->
     h loc = Array a ->
     forall hi:0 <= i < array_length a,
     (St pc (Num i :: Ref loc :: s) l h)
       -[p]-> (St (next pc) (Num (array_values a i hi) :: s) l h)
 | step_iaload_2 : forall pc h a  i s l loc,
     instr_at p pc = Some Arrayload ->
     h loc = Array a ->
     ~ 0 <= i < array_length a ->
     (St pc (Num i::Ref loc :: s) l h) -[p]-> Error
 | step_iastore_1 : forall pc h i s l loc n h',
     instr_at p pc = Some Arraystore ->
     heap_update_array h loc i n h' ->
     (St pc (Num n :: Num i :: Ref loc :: s) l h)
        -[p]-> (St (next pc) s l h')
 | step_iastore_2 : forall pc h i s l loc n a,
     instr_at p pc = Some Arraystore ->
     h loc = Array a ->
     ~ 0 <= i < array_length a ->
     (St pc (Num n :: Num i :: Ref loc :: s) l h) -[p]-> Error
 | step_goto : forall pc jump h s l,
     instr_at p pc = Some (Goto jump) ->
     (St pc s l h) -[p]-> (St jump s l h)
 | step_if_1 : forall pc h s n1 n2 l cmp jump,
     instr_at p pc = Some (If cmp jump) ->
     cmp_sem cmp n1 n2 ->
     (St pc (Num n2 :: Num n1 :: s) l h) -[p]-> (St jump s l h)
 | step_if_2 : forall pc h s n1 n2 l cmp jump,
     instr_at p pc = Some (If cmp jump) ->
     ~ cmp_sem cmp n1 n2 ->
     (St pc (Num n2 :: Num n1 :: s) l h) -[p]-> (St (next pc) s l h)
 | step_input : forall pc h s l n,
     instr_at p pc = Some Input ->
     (St pc s l h) -[p]-> (St (next pc) (Num n :: s) l h)
where "s1 -[ p ]-> s2" := (step p s1 s2).

Inductive ReachableStates (p:program) : state -> Prop :=
   reach_init : 
   ReachableStates p (St Word_1 nil (fun _ => Undef) (fun _ => No))
 | reach_next : forall st1 st2,
     ReachableStates p st1 ->
     st1 -[p]-> st2 ->
     ReachableStates p st2.

Definition Safe : state -> Prop :=
  fun st => st<>Error.



(* 
*** Local Variables: ***
*** coq-prog-name: "coqtop" ***
*** coq-prog-args: ("-emacs-U" "-I" "library") ***
*** End: ***
 *)